CREATE TABLE languages (
  code varchar(5) CONSTRAINT languages_pk PRIMARY KEY,
  name varchar(20) NOT NULL
);

CREATE TABLE users (
  id serial CONSTRAINT users_pk PRIMARY KEY,
  username varchar(50) NOT NULL,
  name varchar(100) DEFAULT NULL,
  institution varchar(100) DEFAULT NULL,
  email varchar(100) NOT NULL,
  password varchar(32) NOT NULL,
  admin boolean DEFAULT False,
  confidence integer DEFAULT 100,
  register_date date NOT NULL,
  last_login timestamp DEFAULT NULL,
  activated boolean DEFAULT False,
  settings text DEFAULT NULL
);
CREATE UNIQUE INDEX users_username_index ON users (username);
CREATE UNIQUE INDEX users_email_index ON users (email);

CREATE TABLE videos (
  id serial CONSTRAINT videos_pk PRIMARY KEY,
  external_id varchar(200) NOT NULL,
  duration integer NOT NULL,
  language varchar(5) REFERENCES languages(code) NOT NULL,
  owner_id integer REFERENCES users(id) NOT NULL,
  hash varchar(64) NOT NULL,
  public boolean DEFAULT False
);
CREATE INDEX videos_external_id_index ON videos (external_id);
CREATE UNIQUE INDEX videos_hash_index ON videos (hash);

CREATE TABLE videos_metadata (
  id integer REFERENCES videos(id) CONSTRAINT videos_metadata_pk PRIMARY KEY,
  path varchar(200) DEFAULT NULL,
  title varchar(400) DEFAULT NULL,
  topic varchar(200) DEFAULT NULL,
  keywords varchar(200) DEFAULT NULL,
  date date DEFAULT NULL,
  removed boolean DEFAULT False
);



CREATE TABLE caption_types (
  id integer CONSTRAINT captions_types_pk PRIMARY KEY,
  description varchar(100) NOT NULL
);

CREATE TYPE t_subtitle_type AS ENUM ('tx', 'tl');
CREATE TABLE captions (
  id serial CONSTRAINT captions_pk PRIMARY KEY,
  video_id integer REFERENCES videos(id) NOT NULL,
  language varchar(5) DEFAULT NULL,
  type integer REFERENCES caption_types(id) NOT NULL,
  locked bool DEFAULT False,
  sub_type t_subtitle_type NOT NULL,
  sup_ratio float DEFAULT 0,
  sup_wer float DEFAULT NULL,
  sup_cer float DEFAULT NULL,
  sup_bleu float DEFAULT NULL,
  sup_ter float DEFAULT NULL,
  rev_time integer DEFAULT 0,
  rev_rtf float DEFAULT 0,
  from_language varchar(5) DEFAULT NULL,
  system_id integer DEFAULT NULL,
  system_time integer DEFAULT NULL,
  system_rtf float DEFAULT NULL
);
CREATE INDEX captions_video_id_index ON captions (video_id);

CREATE TYPE t_audiotrack_vtype AS ENUM ('rec', 'tts');
CREATE TYPE t_audiotrack_atype AS ENUM ('cl', 'db');
CREATE TYPE t_audiotrack_gender AS ENUM ('m', 'f');
CREATE TABLE audiotracks (
  id serial CONSTRAINT audiotracks_pk PRIMARY KEY,
  language varchar(5) NOT NULL,
  voice_type t_audiotrack_vtype NOT NULL,
  audio_type t_audiotrack_atype DEFAULT NULL,
  voice_gender t_audiotrack_gender DEFAULT NULL,
  description varchar(100) DEFAULT NULL,
  video_id integer REFERENCES videos(id) NOT NULL,
  caption_id integer REFERENCES captions(id) DEFAULT NULL,
  system_id integer DEFAULT NULL
);

CREATE TABLE file_types (
  id integer CONSTRAINT file_types_pk PRIMARY KEY,
  name varchar(20) NOT NULL,
  description varchar(100) NOT NULL
);

CREATE TYPE t_file_formats_type AS ENUM ('video', 'audio', 'text', 'image', 'package');
CREATE TABLE file_formats (
  name varchar(4) CONSTRAINT file_formats_pk PRIMARY KEY,
  description varchar(100) DEFAULT NULL,
  type t_file_formats_type NOT NULL
);

CREATE TABLE object_types (
  id integer CONSTRAINT object_types_pk PRIMARY KEY,
  table_name varchar(50) NOT NULL
);

CREATE TABLE files (
  id serial CONSTRAINT files_pk PRIMARY KEY,
  object_id integer NOT NULL,
  object_type_id integer REFERENCES object_types(id) NOT NULL,
  file_type_id integer REFERENCES file_types(id) NOT NULL,
  file_format varchar(4) REFERENCES file_formats(name) NOT NULL,
  file_path varchar(500) NOT NULL,
  is_url boolean NOT NULL
);

CREATE INDEX files_object_id_index ON files (object_id);

CREATE TABLE speakers (
  id serial CONSTRAINT speakers_pk PRIMARY KEY,
  external_id varchar(200) NOT NULL,
  name varchar(100) NOT NULL,
  gender varchar(1) DEFAULT NULL,
  native_language varchar(5) REFERENCES languages(code) DEFAULT NULL,
  age integer DEFAULT NULL,
  email varchar(100) DEFAULT NULL,
  phone varchar(30) DEFAULT NULL,
  send_notifications boolean DEFAULT True,
  owner_id integer REFERENCES users(id) NOT NULL,
  hash varchar(64) NOT NULL
);
CREATE INDEX speakers_external_id_index ON speakers (external_id);
CREATE UNIQUE INDEX speakers_hash_index ON speakers (hash);

CREATE TABLE speakers_videos (
  video_id integer REFERENCES videos(id),
  speaker_id integer REFERENCES speakers(id),
  CONSTRAINT speakers_videos_pk PRIMARY KEY(video_id, speaker_id)
);
CREATE INDEX speakers_videos_video_id_index ON speakers_videos (video_id);
CREATE INDEX speakers_videos_speaker_id_index ON speakers_videos (speaker_id);

CREATE TABLE machines (
  id integer CONSTRAINT machines_pk PRIMARY KEY,
  hostname varchar(100) NOT NULL,
  ip varchar(40) DEFAULT NULL
);

CREATE TABLE mount_points (
  name varchar(50),
  machine_id integer REFERENCES machines(id),
  path varchar(300) NOT NULL,
  CONSTRAINT paths_pk PRIMARY KEY(name,machine_id)
);


CREATE TYPE t_object_type AS ENUM ('videos', 'documents', 'websites');
CREATE TABLE uploads (
  id varchar(50) CONSTRAINT uploads_pk PRIMARY KEY,
  op_code integer DEFAULT NULL,
  external_id varchar(200) DEFAULT NULL,
  language varchar(5) DEFAULT NULL,
  user_id integer DEFAULT NULL,
  session_id integer DEFAULT NULL,
  email varchar(100) DEFAULT NULL,
  timestamp timestamp NOT NULL,
  status_id integer NOT NULL,
  info varchar(200),
  error_code integer DEFAULT NULL,
  event_code varchar(50) DEFAULT NULL,
  cluster_job_ids integer[],
  last_status_update timestamp NOT NULL,
  consumed_quota integer DEFAULT NULL, 
  hash varchar(64) DEFAULT NULL,
  manifest text DEFAULT NULL,
  object_type t_object_type NOT NULL,
  control_data text DEFAULT NULL,
  cancel_request boolean DEFAULT False,
  debug boolean DEFAULT False,
  api_version integer NOT NULL,
  eta timestamp DEFAULT NULL
);
CREATE INDEX uploads_external_id_index ON uploads (external_id);

CREATE TABLE users_videos (
  user_id integer REFERENCES users(id),
  video_id integer REFERENCES videos(id),
  CONSTRAINT users_videos_pk PRIMARY KEY(user_id, video_id)
);

CREATE TABLE groups (
  id serial CONSTRAINT groups_pk PRIMARY KEY,
  groupname varchar(50) NOT NULL,
  description varchar(300), 
  admin_id integer REFERENCES users(id) NOT NULL
);

CREATE TABLE users_groups (
  user_id integer REFERENCES users(id),
  group_id integer REFERENCES groups(id),
  CONSTRAINT users_groups_pk PRIMARY KEY(user_id, group_id)
);

CREATE TABLE groups_videos (
  group_id integer REFERENCES groups(id),
  video_id integer REFERENCES videos(id),
  CONSTRAINT groups_videos_pk PRIMARY KEY(group_id, video_id)
);

CREATE TABLE users_quota (
  user_id integer REFERENCES users(id),
  quota_time integer DEFAULT NULL,
  quota_videos integer DEFAULT NULL,
  accum_time integer DEFAULT 0,
  accum_videos integer DEFAULT 0,
  quota_documents integer DEFAULT NULL,
  quota_sentences integer DEFAULT NULL,
  quota_words integer DEFAULT NULL,
  accum_documents integer DEFAULT 0,
  accum_sentences integer DEFAULT 0,
  accum_words integer DEFAULT 0
);

CREATE TYPE system_type AS ENUM ('asr', 'mt', 'tts', 'align');
CREATE TYPE t_systems_voice_gender AS ENUM ('m', 'f');
CREATE TABLE systems (
  id serial CONSTRAINT systems_pk PRIMARY KEY,
  lang1 varchar(5) NOT NULL,
  lang2 varchar(5) DEFAULT NULL,
  type system_type NOT NULL,
  name varchar(100) NOT NULL,
  description varchar(100) DEFAULT NULL,
  module_path varchar(500) NOT NULL,
  public boolean default True,
  as_default boolean default True,
  enabled boolean default True,
  system_dir varchar(500) DEFAULT NULL,
  tts_voice_gender t_systems_voice_gender DEFAULT NULL,
  features_tadapt boolean NOT NULL,
  rtf float default NULL,
  const_time float default NULL,
  eval_value float default NULL
);

CREATE TABLE users_systems (
  user_id integer REFERENCES users(id),
  system_id integer REFERENCES systems(id),
  as_default boolean DEFAULT True,
  CONSTRAINT users_systems_pk PRIMARY KEY(user_id, system_id)
);

CREATE TYPE t_mod_session_revised_via AS ENUM ('mark_revised', 'accept', 'reject');
CREATE TYPE t_mod_session_author_type AS ENUM ('human', 'computer');
CREATE TYPE t_mod_session_session_type AS ENUM ('player', 'transcriber');
CREATE TABLE mod_sessions (
  id serial CONSTRAINT mod_sessions_pk PRIMARY KEY,
  object_type t_object_type NOT NULL,
  session_type t_mod_session_session_type NOT NULL,
  object_id integer NOT NULL,
  user_id integer REFERENCES users(id) NOT NULL,
  author_id varchar(200) NOT NULL,
  author_conf integer NOT NULL,
  author_name varchar(300) DEFAULT NULL,
  author_type t_mod_session_author_type DEFAULT 'human',
  resumes_session_id integer REFERENCES mod_sessions(id) DEFAULT NULL,
  resumed_by_session_id integer REFERENCES mod_sessions (id) DEFAULT NULL,
  started_at timestamp NOT NULL,
  last_update timestamp DEFAULT NULL,
  last_alive timestamp DEFAULT NULL,
  ended_at timestamp DEFAULT NULL,
  ended_by_id varchar(200) DEFAULT NULL,
  ended_by_name varchar(300) DEFAULT NULL,
  ended_by_type t_mod_session_author_type DEFAULT NULL,
  requires_revision boolean DEFAULT NULL,
  revised boolean DEFAULT False,
  revised_at timestamp DEFAULT NULL,
  revised_by_id varchar(200) DEFAULT NULL,
  revised_by_name varchar(300) DEFAULT NULL,
  revised_via t_mod_session_revised_via DEFAULT NULL,
  revised_in_session_id integer REFERENCES mod_sessions (id) DEFAULT NULL,
  requires_postprocess boolean DEFAULT True,
  postprocessed boolean DEFAULT False,
  evaluated boolean DEFAULT False,
  alignment_completed boolean DEFAULT False, 
  aligned boolean DEFAULT False,
  video_alignments_id varchar(50) DEFAULT NULL,
  do_regenerate boolean DEFAULT True,
  mods text DEFAULT NULL,
  mods_stats text DEFAULT NULL
);
CREATE INDEX mod_sessions_object_id ON mod_sessions (object_id);
CREATE INDEX mod_sessions_author_id ON mod_sessions (author_id);

CREATE TABLE mods (
  id serial CONSTRAINT mods_pk PRIMARY KEY,
  mod_session_id integer REFERENCES mod_sessions(id) NOT NULL,
  timestamp timestamp NOT NULL,
  language varchar(5) REFERENCES languages(code) NOT NULL,
  on_session boolean DEFAULT True,
  mod text NOT NULL,
  mod_stats text DEFAULT NULL
);
CREATE INDEX mods_mod_session_id ON mods (mod_session_id);

CREATE TABLE video_alignments (
  id varchar(50) CONSTRAINT video_alignments_pk PRIMARY KEY,
  video_id integer NOT NULL,
  timestamp timestamp NOT NULL,
  eta timestamp DEFAULT NULL,
  completed_on timestamp DEFAULT NULL,
  status_id integer NOT NULL,
  info varchar(200),
  error_code integer DEFAULT NULL,
  input_data json NOT NULL,
  resegment boolean NOT NULL,
  control_data json DEFAULT NULL,
  output_data json DEFAULT NULL,
  committed boolean DEFAULT False,
  cancelled boolean DEFAULT False,
  debug boolean DEFAULT False
);

/* DOCUMENTS */

CREATE TABLE documents (
  id serial CONSTRAINT documents_pk PRIMARY KEY,
  external_id varchar(200) NOT NULL,
  language varchar(2) REFERENCES languages(code) NOT NULL,
  sentences integer NOT NULL,
  words integer NOT NULL,
  chars integer NOT NULL,
  format varchar(4) REFERENCES file_formats(name) NOT NULL,
  title varchar(400) DEFAULT NULL,
  date date DEFAULT NULL,
  owner_id integer REFERENCES users(id) NOT NULL,
  hash varchar(64) NOT NULL
);
CREATE INDEX documents_external_id_index ON documents (external_id);
CREATE UNIQUE INDEX documents_hash_index ON documents (hash);

CREATE TYPE t_translation_status AS ENUM ('not_translated', 'fully_automatic', 'partially_human', 'fully_human');
CREATE TABLE translations (
  id serial CONSTRAINT translations_pk PRIMARY KEY,
  document_id integer REFERENCES documents(id) NOT NULL,
  language varchar(2) REFERENCES languages(code) NOT NULL,
  status t_translation_status NOT NULL,
  locked bool DEFAULT False,
  sup_ratio float DEFAULT 0,
  sup_bleu float DEFAULT NULL,
  sup_ter float DEFAULT NULL,
  from_language varchar(5) DEFAULT NULL,
  from_translation_id integer DEFAULT NULL,
  system_id integer DEFAULT NULL
);
CREATE INDEX translations_document_id_index ON translations (document_id);

CREATE TABLE users_documents (
  user_id integer,
  document_id integer REFERENCES documents(id),
  CONSTRAINT users_documents_pk PRIMARY KEY(user_id, document_id)
);

CREATE TABLE groups_documents (
  group_id integer REFERENCES groups(id),
  document_id integer REFERENCES documents(id),
  CONSTRAINT groups_documents_pk PRIMARY KEY(group_id, document_id)
);

/* WEB SITES */

CREATE TABLE websites (
  id serial CONSTRAINT websites_pk PRIMARY KEY,
  key varchar(32) NOT NULL,
  name varchar(200) NOT NULL,
  language varchar(2) REFERENCES languages(code) NOT NULL,
  url_pattern varchar(300) NOT NULL,
  active boolean default True,
  owner_id integer REFERENCES users(id) NOT NULL,
  options json DEFAULT NULL
);
CREATE INDEX websites_key_index ON websites (key);

CREATE TABLE source_texts (
  id serial CONSTRAINT texts_pk PRIMARY KEY,
  website_id integer REFERENCES websites(id),
  hash varchar(36) NOT NULL,
  added_on timestamp NOT NULL,
  last_access timestamp NOT NULL,
  text text NOT NULL
);
CREATE INDEX source_texts_website_id_index ON source_texts (website_id);
CREATE INDEX source_texts_hash_index ON source_texts (hash);
CREATE UNIQUE INDEX source_texts_website_hash ON source_texts (website_id, hash);

CREATE TYPE t_translated_texts_status AS ENUM ('automatic', 'revised');
CREATE TABLE translated_texts (
  id serial CONSTRAINT translated_texts_pk PRIMARY KEY,
  source_text_id integer REFERENCES source_texts(id),
  language varchar(2) REFERENCES languages(code) NOT NULL,
  status t_translated_texts_status NOT NULL,
  last_update timestamp NOT NULL,
  text text DEFAULT NULL
);
CREATE INDEX translated_texts_source_text_id_index ON translated_texts (source_text_id);
CREATE UNIQUE INDEX translated_texts_source_text_lang ON translated_texts (source_text_id, language);

CREATE TABLE users_quota_orders (
  id serial CONSTRAINT users_quota_orders_pk PRIMARY KEY,
  user_id integer REFERENCES users(id) NOT NULL,
  order_number varchar(32) NOT NULL,
  currency varchar(5) NOT NULL,
  amount integer NOT NULL,
  quota_time integer NOT NULL,
  quota_words integer NOT NULL,
  timestamp timestamp DEFAULT CURRENT_TIMESTAMP,
  confirmed_at timestamp DEFAULT NULL,
  error varchar(255) DEFAULT NULL
);
