import sys

from datetime import datetime, timedelta
from cStringIO import StringIO
import traceback
from ConfigParser import ConfigParser
from imp import load_source

from storage import Storage
from scheduler import Scheduler
from modules.base.mailer import TextMailer, AlignMailer
from modules.base.urldownloader import UrlDownloader

class ISCoreException(Exception):
    """Generic Exception"""
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg

class ISCore():

    OBJ_SPEECH = 'videos'
    OBJ_TEXT = 'documents'
    OBJ_WEB = 'websites'
    OBJ_VALIGN = 'valign'

    def __init__(self, config_file, object_type, db=None, verbose=False, debug=False, commit_errors=True, script_dir=None):
        self._verbose = verbose
        if self._verbose: self._log("----------------- INGEST INIT ------------------")
        if object_type not in (self.OBJ_SPEECH, self.OBJ_TEXT, self.OBJ_WEB, self.OBJ_WEB, self.OBJ_VALIGN):
            raise ISCoreException("Bad Object Type")
        self._object_type = object_type
        self._debug = debug
        self._commit_errors = commit_errors
        self._cfg = {}
        self._stg = None
        self._jbs = None
        self._sys = None
        self._mailer = None
        self._parse_config(config_file, script_dir)
        if db != None: 
           self._cfg['STO']['DB_NAME'] = db
        self._create_storage()
        self._create_scheduler()
        self._create_mailer()
 
    def _create_storage(self):
        stg_ot = Storage.OBJ_SPEECH
        if self._object_type == self.OBJ_TEXT:
            stg_ot = Storage.OBJ_TEXT
        elif self._object_type == self.OBJ_WEB:
            stg_ot = Storage.OBJ_WEB
        elif self._object_type == self.OBJ_VALIGN:
            stg_ot = Storage.OBJ_VALIGN
        self._stg = Storage(stg_ot, self._cfg['GEN']['HOSTNAME'], self._cfg['GEN']['TL_USER'], self._cfg['GEN']['TL_GROUP'], self._cfg['STO']['DB_NAME'], self._cfg['STO']['DB_USER'], self._cfg['STO']['DB_HOST'], self._cfg['STO']['DB_PASS'])
  
    def _create_scheduler(self):
        self._jbs = Scheduler(self._cfg['SCH']['LOCALHOST'], self._cfg['SCH']['STATUS_CMD'], self._cfg['SCH']['SUBMIT_SCR'], self._cfg['SCH']['SUBMIT_OPTS'])
 
    def _create_mailer(self):
        if self._cfg['MAI']['ENABLED']:
            if self._cfg['MAI']['SRC_ADDRESS'] != "" and self._cfg['MAI']['SMTP_SERVER']  != "":
                if self._object_type == self.OBJ_TEXT:
                    try:
                        self._mailer = TextMailer(self._cfg['MAI']['SRC_ADDRESS'], self._cfg['MAI']['SMTP_SERVER'], self._cfg['MAI']['AAS'], self._cfg['MAI']['AAE'], self._cfg['MAI']['AAF'])
                    except:
                        self._print_excpt(send_mail=False)
                        self._mailer = None
                elif self._object_type == self.OBJ_VALIGN:
                    try:
                        self._mailer = AlignMailer(self._cfg['MAI']['SRC_ADDRESS'], self._cfg['MAI']['SMTP_SERVER'], self._cfg['MAI']['AAE'])
                    except:
                        self._print_excpt(send_mail=False)
                        self._mailer = None
        else:
            self._mailer = None

    def _load_systems_align(self):    
        res = self._stg.get_systems()
        if res['code'] != 0:
            raise ISCoreException(res['info'])
        sys_list = res['data']
        self._sys = { 'AL':{} }
        self._sys['AL']['_ids'] = {}
        for s in sys_list:
            if s['type'] == 'align':
                sl = s['lang1']
                module_fp = "%s/%s" %(self._stg.systems_mp, s['file_path'])
                module_id = "al_%s__sysid_%d" % (sl, s['id'])
                mod = load_source(module_id, module_fp)
                if "align" not in dir(mod):
                    raise ISCoreException("align() function not found in '%s' ALIGN module at %s" % (sl, module_fp))
                if "ID" not in dir(mod):
                    raise ISCoreException("ID attribute not found in '%s' ALIGN module at %s" % (sl, module_fp))
                dsys = { 'id': s['id'], 'public': s['public'], 'allowed_users': s['user_list'], 'module': mod, 'rtf': s['rtf']}
                if sl not in self._sys['AL'].keys(): self._sys['AL'][sl] = { 'default':None, 'dsys':{} }
                self._sys['AL'][sl]['dsys'][s['id']] = dsys
                self._sys['AL']['_ids'][s['id']] = sl
                if s['as_default']: 
                    if self._sys['AL'][sl]['default'] == None:
                        self._sys['AL'][sl]['default'] = s['id']
                    else:
                        self._warn("[INIT] Already exists a default '%s' ALIGN System. Ignoring." % (sl))
                if self._debug: self._log("[INIT]  --> '%s' ALIGN module succesfully loaded" % sl)
        # Check defaults
        for p in self._sys['AL'].keys():
            if p != '_ids':
                if self._sys['AL'][p]['default'] == None:
                    self._sys['AL'][p]['default'] = self._sys['AL'][p]['dsys'].keys()[0]
                    self._warn("[INIT] Auto-defining default '%s' ALIGN Module: ID %d" % (p, self._sys['AL'][p]['default']))
 

    def _load_systems_mt(self):    
        res = self._stg.get_systems()
        if res['code'] != 0:
            raise ISCoreException(res['info'])
        sys_list = res['data']
        self._sys = { 'MT':{} }
        self._sys['MT']['_ids'] = {}
        for s in sys_list:
            if s['type'] == 'mt':
                sl = s['lang1']
                tl = s['lang2']
                pair = "%s-%s" % (sl, tl)
                module_fp = "%s/%s" %(self._stg.systems_mp, s['file_path'])
                module_id = "mt_%s__sysid_%d" % (pair, s['id'])
                mod = load_source(module_id, module_fp)
                if "translate" not in dir(mod):
                    raise ISCoreException("translate() function not found in '%s' MT module at %s" % (pair, module_fp))
                if "ID" not in dir(mod):
                    raise ISCoreException("ID attribute not found in '%s' MT module at %s" % (pair, module_fp))
                dsys = { 'id': s['id'], 'public': s['public'], 'allowed_users': s['user_list'], 'module': mod}
                if pair not in self._sys['MT'].keys(): self._sys['MT'][pair] = { 'default':None, 'dsys':{} }
                self._sys['MT'][pair]['dsys'][s['id']] = dsys
                self._sys['MT']['_ids'][s['id']] = pair
                if s['as_default']: 
                    if self._sys['MT'][pair]['default'] == None:
                        self._sys['MT'][pair]['default'] = s['id']
                    else:
                        self._warn("[INIT] Already exists a default '%s' MT System. Ignoring." % (pair))
                if self._debug: self._log("[INIT]  --> '%s' MT module succesfully loaded" % pair)
        # Check defaults
        for p in self._sys['MT'].keys():
            if p != '_ids':
                if self._sys['MT'][p]['default'] == None:
                    self._sys['MT'][p]['default'] = self._sys['MT'][p]['dsys'].keys()[0]
                    self._warn("[INIT] Auto-defining default '%s' MT Module: ID %d" % (p, self._sys['MT'][p]['default']))
    
    def _parse_config(self, config_file, script_dir): 
        # Create object
        config = ConfigParser(allow_no_value=True)
        with open(config_file, 'r') as cfg_fd:
            config.readfp(cfg_fd)
        # General
        self._cfg['GEN'] = {}
        self._cfg['GEN']['HOSTNAME'] = config.get('general', 'hostname')
        self._cfg['GEN']['TL_USER'] = config.get('general', 'tl_user')
        self._cfg['GEN']['TL_GROUP'] = config.get('general', 'tl_group')
        self._cfg['GEN']['RM_FIN_UP_DAYS'] = int(config.get('general', 'rm_finished_up_days'))
        self._cfg['GEN']['RM_ERR_UP_DAYS'] = int(config.get('general', 'rm_error_up_days'))
        self._cfg['GEN']['LOCAL_REPO'] = config.getboolean('general', 'local_repository')
        # Authentication
        self._cfg['AUT'] = {}
        self._cfg['AUT']['ENABLED'] = True
        # Storage
        self._cfg['STO'] = {}
        self._cfg['STO']['DB_NAME'] = config.get('storage', 'db_name')
        self._cfg['STO']['DB_USER'] = config.get('storage', 'db_user')
        self._cfg['STO']['DB_PASS'] = config.get('storage', 'db_passwd')
        if self._cfg['STO']['DB_PASS'] != None:
            if self._cfg['STO']['DB_PASS'].strip() == "": self._cfg['STO']['DB_PASS'] = None
        self._cfg['STO']['DB_HOST'] = config.get('storage', 'db_host')
        # Scheduler
        self._cfg['SCH'] = {}
        self._cfg['SCH']['LOCALHOST'] = config.get('scheduler', 'localhost')
        self._cfg['SCH']['STATUS_CMD'] = config.get('scheduler', 'status_cmd')
        self._cfg['SCH']['SUBMIT_SCR'] = config.get('scheduler', 'submit_scr')
        self._cfg['SCH']['SUBMIT_OPTS'] = config.get('scheduler', 'submit_opts')
        self._cfg['SCH']['JOB_NAME_PREFIX'] = config.get('scheduler', 'job_name_prefix') + ".%s" % (self._object_type)
        # SESSIONS
        self._cfg['SES'] = {}
        self._cfg['SES']['ENABLED'] = config.getboolean('sessions', 'enabled')
        self._cfg['SES']['AID'] = config.get('sessions', 'author_id')
        self._cfg['SES']['ANM'] = config.get('sessions', 'author_name')
        self._cfg['SES']['ACF'] = config.getint('sessions', 'author_conf')
        # Mailing
        self._cfg['MAI'] = {}
        try:
            self._cfg['MAI']['ENABLED'] = config.getboolean('mailing', 'enabled')
            self._cfg['MAI']['SRC_ADDRESS'] = config.get('mailing', 'from_address')
            self._cfg['MAI']['SMTP_SERVER'] = config.get('mailing', 'smtp_server')
        except:
            warn("Mailing configuration variables are not defined. Mailing module will be disabled.")
            self._cfg['MAI']['ENABLED'] = False
        self._cfg['MAI']['SCSM'] = config.getboolean('mailing', 'send_client_started_mail')
        self._cfg['MAI']['SCEM'] = config.getboolean('mailing', 'send_client_error_mail')
        self._cfg['MAI']['SCFM'] = config.getboolean('mailing', 'send_client_finished_mail')
        self._cfg['MAI']['SASM'] = config.getboolean('mailing', 'send_admin_started_mail')
        try:
            self._cfg['MAI']['AAS'] = config.get('mailing', 'admin_address_started').replace(' ', '').split(',')
        except:
            self._cfg['MAI']['AAS'] = None
        self._cfg['MAI']['SAEM'] = config.getboolean('mailing', 'send_admin_error_mail')
        try:
            self._cfg['MAI']['AAE'] = config.get('mailing', 'admin_address_error').replace(' ', '').split(',')
        except:
            self._cfg['MAI']['AAE'] = None
        self._cfg['MAI']['SAFM'] = config.getboolean('mailing', 'send_admin_finished_mail')
        try:
            self._cfg['MAI']['AAF'] = config.get('mailing', 'admin_address_finished').replace(' ', '').split(',')
        except:
            self._cfg['MAI']['AAF'] = None
        # Document Format Parsers
        self._doc_parsers = {}
        for x in config.items('document_format_parsers'):
            form = x[0]
            path = x[1]
            if path[0] != '/':
                path = script_dir + '/' + path
            mod = load_source(form, path)
            if "parse_document" not in dir(mod):
                raise ISCoreException("parse_document() function not found in '%s' document parser module at %s" % (form, path))
            self._doc_parsers[form] = mod


    def _log(self, msg):
        if self._verbose:
            sys.stderr.write("[LL] (%s) %s\n" %(self._now(), msg))
    
    
    def _error(self, msg, exit_code=1):
        sys.stderr.write("\n[EE] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" %self._now())
        sys.stderr.write("[EE] (%s) |                         FATAL ERROR!!                     |\n" %self._now())
        sys.stderr.write("[EE] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" %self._now())
        sys.stderr.write("[EE] (%s) %s\n" %(self._now(), msg))
        sys.stderr.write("[EE] (%s) ------------------------- ERROR CODE %d ---------------------\n" %(self._now(), exit_code))
        self._print_excpt() 
        sys.exit(exit_code)
    
    
    def _bug(self, msg=None, exit=False):
        sys.stderr.write("\n[BB] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" %self._now())
        sys.stderr.write("[BB] (%s) |                     UNCAUGHT EXCEPTION!!                  |\n" %self._now())
        sys.stderr.write("[BB] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" %self._now())
        if msg != None:
            sys.stderr.write("[BB] (%s) %s\n" % (self._now(), msg))
        self._print_excpt()
        if exit:
             sys.exit(1)
    
    
    def _print_excpt(self, string='', send_mail=True):
        out = StringIO()
        traceback.print_exc(file=out)
        excpt_str = out.getvalue().strip()
        if excpt_str == "None" and string == "":
            return
        sys.stderr.write("\n[TT] (%s) ++++++++++++++++++++++++ ERROR TRACE ++++++++++++++++++++++++\n" %self._now())
        if string != "":
            sys.stderr.write("%s\n" % str(string))
        if excpt_str != "None":
            sys.stderr.write("%s\n" % str(excpt_str))
        sys.stderr.write("[TT] (%s) +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n" %self._now())
        sys.stderr.write("\n")
        if self._mailer != None and send_mail:
            try:
                self._mailer.send_mail_ingest_bug(excpt_str, extra_info=string)
            except:
                out = StringIO()
                traceback.print_exc(file=out)
                excpt_str = out.getvalue().strip()
                sys.stderr.write("%s\n" % str(excpt_str))
    
    def _warn(self, msg):
        sys.stderr.write("[WW] (%s) %s\n" %(self._now(), msg))
    
    
    def _now(self):
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    
    def check_running_processes(self, rjobs, jobs):
        for jid in jobs:
            try:
                jst = rjobs[jid]
            except KeyError:
                continue
            if jst == Scheduler.ST_RUNNING:
                return {'code':0, 'data':Scheduler.ST_RUNNING}
            elif jst == Scheduler.ST_QUEUED:
                return {'code':0, 'data':Scheduler.ST_QUEUED}
            elif jst == Scheduler.ST_ERROR:
                return {'code':0, 'data':Scheduler.ST_ERROR}
        return {'code':0, 'data':Scheduler.ST_FINISHED}
    
    def do_infer_tlpaths(self, plist, slist):
        return self.infer_tlpaths(plist, [], slist, True, True)
    
    def infer_tlpaths(self, plist, clist, slist, lb, la):
        if len(plist) == 0: 
            if len(clist) > 0:
                slist.append(clist)
            return
        if len(clist) == 0:
            clist.insert(0, plist.pop())
        # Search input
        if lb:
            slang = clist[0].split("-")[0]
            i = 0; fb = False
            while i <len(plist) and not(fb):
                tlang = plist[i].split("-")[1]
                if tlang == slang:
                    fb = True
                i += 1
            if fb:
                clist.insert(0, plist.pop(i-1))
        else:
            fb = False
        # Search output
        if la:
            tlang = clist[-1].split("-")[1]
            i = 0; fa = False
            while i <len(plist) and not(fa):
                slang = plist[i].split("-")[0]
                if slang == tlang:
                    fa = True
                i += 1
            if fa:
                clist.append(plist.pop(i-1))
        else:
            fa = False
        if not(fb) and not(fa):
            slist.append(clist)
            clist = []
            fb = True; fa = True
        return self.infer_tlpaths(plist, clist, slist, fb, fa)
    
    def get_user_custom_default_systems(self, user_id):
        res = self._stg.get_user_default_systems(user_id)
        if res['code'] != 0:
            return res
        lsys = res['data']
        ret = { 'mt':{} }
        for s in lsys:
            if s['type'] == "mt":
                lang = "%s-%s" % (s["lang1"], s["lang2"])
                ret["mt"][lang] = s["id"]
        return {'code':0, 'data':ret}
   
    def mtsort(self, keys):
        if keys == []:
            return []
        skeys=[]
        skeys.append(keys.pop())
        self._mtsort(keys,[],skeys)
        return skeys
    
    def _mtsort(self, keys, akeys, skeys):
        if len(keys) == 0:
            if len(akeys) > 0:
                skeys.append(akeys.pop())
                return self._mtsort(akeys,[],skeys)
        else:
            p = keys.pop()
            tsrc = p.split('-')[0]
            ttrg = p.split('-')[1]
            i=0; f=False
            while i < len(skeys) and not(f):
                src = skeys[i].split('-')[0]
                trg = skeys[i].split('-')[1]
                if trg == tsrc:
                    skeys.insert(i+1,p)
                    f=True
                elif src == ttrg:
                    skeys.insert(i,p)
                    f=True
                i+=1
            if f:
                return self._mtsort(keys,akeys,skeys)
            else:
                akeys.append(p)
                return self._mtsort(keys,akeys,skeys)
     
    def get_user_info(self, user_id):
        res = self._stg.get_user(user_id)
        if res['code'] != 0:
            return res
        user_is_admin = res['data'][0]['is_admin']
        username = res['data'][0]['username']
        realusername = res['data'][0]['name']
        usermail = res['data'][0]['email']
        d = {'id':user_id, 'is_admin':user_is_admin, 'name':username, 'real_name':realusername, 'email':usermail}
        return {'code':0, 'data':d}

