# -*- coding: utf-8 -*-
#
#  Copyright 2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

#---------- Module Imports -------------#

# Standard Libraries
import sys
import os
import traceback
import re
import json
import base64
import unicodedata
from cStringIO import StringIO
from imp import load_source
from optparse import OptionParser
from ConfigParser import ConfigParser
from ast import literal_eval
from datetime import datetime, timedelta
import hashlib

# Custom module imports
from ISCore import ISCore, ISCoreException
from storage import Storage
from scheduler import Scheduler
from modules.base.media import download_media

class VideoAlignment():

    STCODE_PENDING = 0
    STCODE_PROCESSING = 1
    STCODE_FINISHED = 2
    STCODE_CANCELLED = 40
    STCODE_ERROR = 100

    def __init__(self, valign_id, video_id, timestamp, status_id, input_data, resegment, 
                      eta=None, completed_on=None, info=None, control_data=None, 
                      output_data=None, error_code=None, cancelled=False, debug=False):
        self.id = valign_id
        self.video_id = video_id
        self.timestamp = timestamp
        self.eta = eta
        self.status_id = status_id
        self.info = info
        self.error_code = error_code
        self.input_data = input_data
        self.resegment = resegment
        self.output_data = output_data
        self.cancelled = cancelled
        self.debug = debug
        self.completed_on = completed_on
        if control_data == None:
            self._init_control()
        else:
            self.control_data = control_data
 
    def __repr__(self):
        return "([VideoAlignment Object] <id = %s> <status = %d> <info = %s>)" % (self.id, self.status, self.info)

    def _init_control(self):
        self.control_data = { 'jbs':{'all':[], 'dep':{'align':[]}, 'prefix':None}, 'files':{}, 'meta':{} }

class ISAlignCore(ISCore):

    CONST_IS_TIME = 60 # 1 min 

    def __init__(self, config_file, db=None, verbose=False, debug=False, commit_errors=True, script_dir=None):
        self._video_alignments = None
        ISCore.__init__(self, config_file, ISCore.OBJ_VALIGN, db, verbose, debug, commit_errors, script_dir)
        self._load_systems_align() 
        self._get_video_alignments() 
        self._videos_being_processed = []
 
    def _get_video_alignments(self): 
        self._video_alignments = { 'failed':[], 'finished':[], 'running':[], 'new':[] }
        res = self._stg.get_video_alignments() 
        if res['code'] != 0:
            raise ISCoreException(res['info'])
        for r in res['data']:
            va = VideoAlignment(r['id'], r['video_id'], r['timestamp'], r['status_id'], 
                                r['input_data'], r['resegment'], r['eta'], r['completed_on'], r['info'], 
                                r['control_data'], r['output_data'], r['error_code'], 
                                r['cancelled'], r['debug']) 
            if self._debug and not(va.debug): # do not process regular upload under debug mode
                #self._log("[Debug mode] Skipping regular upload ID '%s'" % up.id)
                continue
            elif not(self._debug) and va.debug: # do not process debug uploads under normal mode 
                #self._log("Skipping debug upload ID '%s' " % up.id)
                continue
            if va.status_id == va.STCODE_ERROR:
                self._video_alignments['failed'].append(va)
            elif va.status_id == va.STCODE_FINISHED:
                self._video_alignments['finished'].append(va)
            elif va.status_id == va.STCODE_PROCESSING:
                self._video_alignments['running'].append(va)
            elif va.status_id == va.STCODE_PENDING:
                self._video_alignments['new'].append(va)
            elif va.status_id == va.STCODE_CANCELLED:
                self._video_alignments['finished'].append(va)
            else:
                raise ISCoreException("Unknown status code %d for Alignment ID %s" % (va.status_id, va.id))

    def _generate_text_file(self, va):
        data = ""
        for s in va.input_data:
            data = "%s%.3f %.3f %s\n" % (data, float(s['b']), float(s['e']), s['t'])
        res = self._stg.save_align_text_file(va.id, data)
        if res['code'] != 0:
            return res
        text_file_fp = res['data']
        va.control_data['files']['text'] = text_file_fp
        return {'code':0}

    def _launch_alignment(self, va):
        vid_lang = va.control_data['meta']['language']
        media_file = va.control_data['files']['media']
        text_file = va.control_data['files']['text']
        tmp_dir = self._stg.tmp_align_dir(va.id)
        job_name = "%s.align" % va.control_data['jbs']['prefix']
        hold_job_ids = va.control_data['jbs']['dep']['align']
        system_id = va.control_data['system']
        al = self._sys['AL'][vid_lang]['dsys'][system_id]['module']
        # Call ALIGN Module
        res = al.align(self._jbs, media_file, text_file, tmp_dir, job_name=job_name, hold_job_ids=hold_job_ids, resegment=va.resegment)
        if res['code'] != 0:
            return res
        jobs = res['jids']
        va.control_data['jbs']['all'] += jobs
        json_file = res['json_fp']
        va.control_data['files']['out_json'] = json_file
        # Compute ETA
        eta = None
        sys_rtf = self._sys['AL'][vid_lang]['dsys'][system_id]['rtf']
        if sys_rtf != None:
            ac = 0
            for s in va.input_data:
                ac += (float(s['e']) - float(s['b']))
            eta =  ac * sys_rtf
        self._log("Launched alignment processes with IDs: %s" % str(jobs))
        return {'code':0, 'eta':eta}
    
    def _cancel_alignment(self, va):
        res = self._jbs.delete_jobs(va.control_data['jbs']['all'])
        if res['code'] != 0:
            self._warn(res['info'])
            self._print_excpt()
        return {'code':0} 

    def _check_input_align_json_format(self, jd):
        if not(isinstance(jd, list)):
            return {'code':1100, 'info':'JSON data must be a list'}
        for e in jd:
            if not(isinstance(e, dict)):
                return {'code':1101, 'info':'JSON data must be a list of dicts'}
            if 't' not in e.keys():
                return {'code':1102, 'info':'"t" key not found in JSON data'}
            if 'b' not in e.keys():
                return {'code':1102, 'info':'"b" key not found in JSON data'}
            try:
                x = float(e['b'])
            except:
                return {'code':1102, 'info':'"b" key must be a float value'}
            if 'e' not in e.keys():
                return {'code':1102, 'info':'"e" key not found in JSON data'}
            try:
                x = float(e['e'])
            except:
                return {'code':1102, 'info':'"e" key must be a float value'}
        return {'code':0}

    def _check_output_align_json_format(self, jd):
        if not(isinstance(jd, list)):
            return {'code':1100, 'info':'JSON data must be a list'}
        for e in jd:
            if not(isinstance(e, dict)):
                return {'code':1101, 'info':'JSON data must be a list of dicts'}
            if 'b' not in e.keys():
                return {'code':1102, 'info':'"b" key not found in JSON data'}
            try:
                x = float(e['b'])
            except:
                return {'code':1102, 'info':'"b" key must be a float value'}
            if 'e' not in e.keys():
                return {'code':1102, 'info':'"e" key not found in JSON data'}
            try:
                x = float(e['e'])
            except:
                return {'code':1102, 'info':'"e" key must be a float value'}
            if 'wl' not in e.keys():
                return {'code':1102, 'info':'"wl" key not found in JSON data'}
            if not(isinstance(e['wl'], list)):
                return {'code':1102, 'info':'"wl" key must be a list'}
            for w in e['wl']:
                if not(isinstance(w, dict)):
                    return {'code':1103, 'info':'"wl" key must be a list of dicts'}
                if 'b' not in w.keys():
                    return {'code':1103, 'info':'"b" key not found in JSON data'}
                try:
                    x = float(w['b'])
                except:
                    return {'code':1103, 'info':'"b" key must be a float value'}
                if 'e' not in w.keys():
                    return {'code':1103, 'info':'"e" key not found in JSON data'}
                try:
                    x = float(w['e'])
                except:
                    return {'code':1103, 'info':'"e" key must be a float value'}
                if 'w' not in w.keys():
                    return {'code':1103, 'info':'"w" key not found on word list'}
        return {'code':0}

    def ingest_update(self, va):
        res = self._stg.update_video_alignment(va.id, va.status_id, va.info, va.control_data, 
                                               error_code=va.error_code, output=va.output_data,
                                               completed_on=va.completed_on, eta=va.eta) 
        if res['code'] != 0:
            self._error("Cannot update alignment.", exit_code=res['code'])
        self._log("[Status Update] %s" %(va.info))
     
   
    def ingest_error(self, va, email_admin=True, email_user=False):
        self._log("[Ingest Error %d] %s" %(va.error_code, va.info.encode('utf-8')))
        self._print_excpt(send_mail=False)
        if self._debug:
            self._log("Control JSON:")
            sys.stderr.write("\n%s\n" % json.dumps(va.control_data, indent=2, encoding='utf-8'))
        if self._commit_errors:
            # Update alignment status
            res = self._stg.update_video_alignment(va.id, va.status_id, va.info, va.control_data, 
                                                   error_code=va.error_code, completed_on=va.completed_on) 
            if res['code'] != 0:
                self._error("Cannot update alignment.", exit_code=res['code'])
            #self._log("[Status Update] %s" %(va.info))
            # Send Mail
            if self._mailer != None:
                if email_admin and self._cfg['MAI']['SAEM'] and self._cfg['MAI']['AAE'] != None:
                    res = self._mailer.send_mail_admin_error(va.id, va.timestamp, va.error_code, va.info)
                    if res['code'] != 0:
                        self._warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                        self._print_excpt()
        
    
    ##---------- Upload processing functions -------------#
    
    def process_failed_alignments(self):
        for va in self._video_alignments['failed']:
            try:
                td = (datetime.now() - va.completed_on).days
                if td >= self._cfg['GEN']['RM_ERR_UP_DAYS']:
                    self._log("[*] Deleting Alignment ID %s ..." % va.id)
                    res = self._stg.delete_video_alignment(va.id)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    self._log("Alignment ID %s succesfully deleted." % va.id)
            except:
                self._bug("Alignment ID: %s" % va.id)
                continue
    
    
    def process_finished_alignments(self):
        for va in self._video_alignments['finished']:
            try:
                td = (datetime.now() - va.completed_on).days
                if td >= self._cfg['GEN']['RM_FIN_UP_DAYS']:
                    self._log("[*] Deleting Alignment ID %s ..." % va.id)
                    res = self._stg.delete_video_alignment(va.id)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    self._log("Alignment ID %s succesfully deleted." % va.id)
            except:
                self._bug("Upload ID: %s" % va.id)
                continue
    
    def process_running_alignments(self):
        # Get Running jobs
        res = self._jbs.get_running_jobs()
        if res['code'] != 0:
            self._error(res['info'], exit_code=res['code'])
        rjobs = res['data']
        # For each alignment
        for va in self._video_alignments['running']:
            try:
                self._log("[*] Checking running Alignment ID %s ..." % va.id)
                if va.cancelled:
                    self._log("Oh nooo! Cancel request!")
                    res = self._cancel_alignment(va)
                    if res['code'] != 0:
                        self._bug("Cannot cancel alignment")
                        continue
                    va.completed_on = datetime.now()
                    va.info = "Alignment successfully cancelled."
                    va.status_id = VideoAlignment.STCODE_CANCELLED 
                    self.ingest_update(va) 
                    continue
                res = self.check_running_processes(rjobs, va.control_data['jbs']['all']) 
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                job_status = res['data']
                if job_status == Scheduler.ST_QUEUED:
                    self._log("Processes queued.")
                    va.info = "Processes queued."
                    self.ingest_update(va)
                    self._videos_being_processed.append(va.video_id)
                elif job_status == Scheduler.ST_RUNNING:
                    self._log("Processes running.")
                    va.info = "Processes running."
                    self.ingest_update(va)
                    self._videos_being_processed.append(va.video_id)
                elif job_status == Scheduler.ST_ERROR:
                    if va.status_id == VideoAlignment.STCODE_PROCESSING:
                        va.completed_on = datetime.now()
                        va.info = "Alignment Error."
                        va.status_id = VideoAlignment.STCODE_ERROR
                        va.error_code = 2001
                        self.ingest_error(va)
                        continue
                    else:
                        self._bug("Unexpected alignment status %d" % (job_status), exit=True)
                elif job_status == Scheduler.ST_FINISHED:
                    if va.status_id == VideoAlignment.STCODE_PROCESSING:
                        self._log("Alignment processes finished.")
                        # Load JSON 
                        res = self._stg.load_output_json_file(va.control_data['files']['out_json'])
                        if res['code'] != 0:
                            va.completed_on = datetime.now()
                            va.info = res['info']
                            va.error_code = res['code']
                            va.status_id = VideoAlignment.STCODE_ERROR
                            self.ingest_error(va)
                            continue
                        output_json = res['data']
                        # Check JSON format
                        res = self._check_output_align_json_format(output_json)
                        if res['code'] != 0:
                            va.completed_on = datetime.now()
                            va.info = "Invalid output JSON data: %s" % res['info']
                            va.error_code = res['code']
                            va.status_id = VideoAlignment.STCODE_ERROR
                            self.ingest_error(va)
                            continue
                        # Save data
                        va.output_data = output_json
                        va.completed_on = datetime.now()
                        va.info = "Alignment successfully performed."
                        va.status_id = VideoAlignment.STCODE_FINISHED 
                        self.ingest_update(va)
                    else:
                        self._bug("Unexpected alignment status %d" % (job_status), exit=True)
            except:
                self._bug("Alignment ID: %s" % va.id)
                continue

                
    def process_new_alignments(self):
        for va in self._video_alignments['new']:
            try:
                self._log("[*] Processing new alignment ID %s ..." % va.id)
                if self._debug:
                    self._log("Input data:")
                    sys.stderr.write("\n%s\n" % json.dumps(va.input_data, indent=2, encoding='utf-8'))
                if va.cancelled:
                    self._log("Oh nooo! Cancel request!")
                    va.completed_on = datetime.now()
                    va.info = "Alignment successfully cancelled."
                    va.status_id = VideoAlignment.STCODE_CANCELLED
                    self.ingest_update(va)
                    continue
                if va.video_id in self._videos_being_processed:
                    va.completed_on = datetime.now()
                    va.info = "There is an alignment process running for the requested video."
                    va.status_id = VideoAlignment.STCODE_CANCELLED
                    self.ingest_update(va)
                    continue
                # Create TMP dir
                res = self._stg.create_tmp_dirs(va.id)
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                # Create Outputs dir
                res = self._stg.create_outputs_dir(va.id)
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                # Get lecture info
                res = self._stg.get_video_by_id(va.video_id)
                if res['code'] != 0:
                    va.completed_on = datetime.now()
                    va.info = res['info']
                    va.error_code = res['code']
                    va.status_id = VideoAlignment.STCODE_ERROR
                    self.ingest_error(va)
                    continue
                va.control_data['meta'] = res['data']
                vid_hash = va.control_data['meta']['hash']
                vid_lang = va.control_data['meta']['language']
                self._log("Aligment for video: [ID: %d, EXTID: '%s', OWNER: %d]" % 
                           (va.control_data['meta']['id'], va.control_data['meta']['external_id'],
                            va.control_data['meta']['owner_id']))
                # Check alignment system
                if vid_lang not in self._sys['AL']:
                    va.completed_on = datetime.now()
                    va.info = "No '%s' alignment systems available." % vid_lang
                    va.status_id = VideoAlignment.STCODE_CANCELLED
                    self.ingest_update(va)
                    continue
                va.control_data['system'] = self._sys['AL'][vid_lang]['default']
                # Check JSON format
                res = self._check_input_align_json_format(va.input_data)
                if res['code'] != 0:
                    va.completed_on = datetime.now()
                    va.info = "Invalid input JSON data: %s" % res['info']
                    va.error_code = res['code']
                    va.status_id = VideoAlignment.STCODE_ERROR
                    self.ingest_error(va)
                    continue
                va.control_data['jbs']['prefix'] = "%s.%s.%s.%s" % (self._cfg['SCH']['JOB_NAME_PREFIX'], self._cfg['STO']['DB_NAME'], va.id, vid_lang)
                # Generate text file
                res = self._generate_text_file(va)
                if res['code'] != 0:
                    va.completed_on = datetime.now()
                    va.info = res['info']
                    va.error_code = res['code']
                    va.status_id = VideoAlignment.STCODE_ERROR
                    self.ingest_error(va)
                    continue
                # Get media file
                res = self._stg.get_media_file_abspath(vid_hash) #FIXME for future
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                media_file = res['data']
                va.control_data['files']['media'] = media_file
                # Download media file if necessary
                if res['is_url']:
                    tmp_dir = self._stg.tmp_align_dir(va.id)
                    job_name = "%s.download_media" % va.control_data['jbs']['prefix']
                    media_dest = "%s/media.%s" % (tmp_dir, res['ext'])
                    ret = download_media(self._jbs, media_file, media_dest, tmp_dir, job_name)
                    if ret['code'] != 0:
                        va.completed_on = datetime.now()
                        va.info = res['info']
                        va.error_code = res['code']
                        va.status_id = VideoAlignment.STCODE_ERROR
                        self.ingest_error(va)
                        continue
                    self._log("Found a media URL. I've submitted the download of the media file, JIDs %s" % (str(ret['jids'])))
                    va.control_data['files']['media'] = media_dest
                    va.control_data['jbs']['dep']['align'] += ret['jids']
                    va.control_data['jbs']['all'] += ret['jids']
                # launch alignment
                res = self._launch_alignment(va) 
                if res['code'] != 0:
                    va.completed_on = datetime.now()
                    va.info = res['info']
                    va.error_code = res['code']
                    va.status_id = VideoAlignment.STCODE_ERROR
                    self.ingest_error(va)
                    continue
                etas = res['eta']
                if etas != None:
                    etas += self.CONST_IS_TIME
                    eta = datetime.now() + timedelta(seconds=etas)
                    va.eta = eta
                    self._log("Estimated time to complete: %s" % (eta))
                va.info = "Alignment in progress."
                va.status_id = VideoAlignment.STCODE_PROCESSING
                self.ingest_update(va)
                # Append VideoAlignment object to "running", 
                #self._video_alignments['running'].append(va)
                self._videos_being_processed.append(va.video_id)
                if self._debug:
                    self._log("Control JSON:")
                    sys.stderr.write("\n%s\n" % json.dumps(va.control_data, indent=2, encoding='utf-8'))

            except:
                self._bug("Alignment ID: %s" % va.id)
                continue
    


if __name__ == "__main__":

    DEFAULT_CFG_FP = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))
    DEFAULT_CFG_FN = "config.ini"
    SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

    # Define Option Parser
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option("-v", "--verbose", dest="verb", action="store_true", default=False, help="Verbose power on!")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="Debug mode")
    parser.add_option("-e", "--disable-errors", dest="errors", action="store_false", default=True, help="Disable ingest errors")
    parser.add_option("-C", "--config-file", dest="config_file", type="string", default=DEFAULT_CFG_FP, help="Configuration file. Default: %s" % DEFAULT_CFG_FN)
    parser.add_option("-D", "--database", dest="db_name", type="string", default=None, help="Database with which to work. Default: specified in config file")
    (opts, args) = parser.parse_args()


    core = ISAlignCore(opts.config_file, verbose=opts.verb, debug=opts.debug, commit_errors=opts.errors, script_dir=SCRIPT_DIR)

    try:
        # Remove finished alignments
        if opts.verb: core._log("----------- Checking failed alignments ------------")
        core.process_failed_alignments()
    
        # Remove finished alignments
        if opts.verb: core._log("---------- Checking finished alignments -----------")
        core.process_finished_alignments()
    
        # Check running alignments
        if opts.verb: core._log("----------- Checking running alignments -----------")
        core.process_running_alignments()
    
        # Check new alignments
        if opts.verb: core._log("------------ Checking new alignments --------------")
        core.process_new_alignments()
    
        if opts.verb: core._log("------------------- EXIT 0 ---------------------")
    except: 
        out = StringIO()
        traceback.print_exc(file=out)
        exception_str = out.getvalue().strip()
        print exception_str
        if core._mailer != None:
            core._mailer.send_mail_ingest_bug(exception_str)

