# -*- coding: utf-8 -*-
#
#  Copyright 2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

#---------- Module Imports -------------#

# Standard Libraries
import sys
import os
import traceback
import re
import json
import base64
import unicodedata
from cStringIO import StringIO
from imp import load_source
from optparse import OptionParser
from ConfigParser import ConfigParser
from ast import literal_eval
from datetime import datetime
import hashlib

# Custom module imports
from ISCore import ISCore, ISCoreException
from storage import Storage
from scheduler import Scheduler
from modules.base.mailer import TextMailer
from modules.base.urldownloader import UrlDownloader
from modules.base.format_adapters.dtlx.libdtlx import *
from modules.base.format_adapters.dtlx.dtlx_tools import dtlx2txt, txt2dtlx, create_empty_dtlx, compute_dtlx_src_stats, compute_dtlx_tl_stats

class TextUpload():

    OP_CODE_NEW = 0
    OP_CODE_UPDATE = 1
    OP_CODE_DELETE = 2
    OP_CODE_CANCEL = 3
    
    DELETE_MODE_SOFT = "soft"
    DELETE_MODE_HARD = "hard"
    
    STCODE_INGESTED = 0
    STCODE_MP_PROC = 1
    STCODE_TL_PROC = 2
    STCODE_NEW_FINISHED = 3
    STCODE_UPD_FINISHED = 20
    STCODE_DEL_FINISHED = 30
    STCODE_CANCELLED = 40
    STCODE_UNK_ERROR = 100
    STCODE_MP_ERROR = 101
    STCODE_TL_ERROR = 102
    STCODE_DB_ERROR = 103
    STCODE_UPD_ERROR = 200
    STCODE_DEL_ERROR = 300
    STCODE_CANCEL_ERROR = 400
    ERR_STCODES = [ STCODE_UNK_ERROR, STCODE_MP_ERROR, STCODE_TL_ERROR, 
                    STCODE_DB_ERROR, STCODE_UPD_ERROR, STCODE_DEL_ERROR, 
                    STCODE_CANCEL_ERROR ]
    RUN_STCODES = [ STCODE_MP_PROC, STCODE_TL_PROC ]
    FIN_STCODES = [ STCODE_NEW_FINISHED, STCODE_UPD_FINISHED, STCODE_DEL_FINISHED, STCODE_CANCELLED ]

    def __init__(self, up_id, op_code, ts, last_ts, manifest, user_id, status=None, info=None, control_data=None, error_code=None, cancel_request=False, debug=False):
        self.id = up_id
        self.op_code = op_code
        self.status = status
        self.prev_status = status
        self.info = info
        self.error_code = error_code
        self.ts = ts 
        self.last_ts = last_ts 
        self.user_id = user_id
        self.cancel = cancel_request
        self.debug = debug
        self.manifest = manifest
        if control_data == None:
            self._init_control()
        else:
            self.control = control_data
 
    def __repr__(self):
        return "([TextSpeech Object] <id = %s> <op_code = %d> <status = %d> <info = %s>)" % (self.id, self.op_code, self.status, self.info)

    def _init_control(self):
        self.control = { 'jobs':[], 'ids':{}, 'user':{}, 'systems':{}, 'quota':{'d':0, 's':0, 'w':0, 'c':0} }
        for d in self.manifest['documents']:
            obj_id = d['external_id']
            self.control['ids'][obj_id] = { 'existing_tl':{'tl':{}, 'indexes':{'by_src':{}, 'by_trg':{}, 'by_pair':{}}}, 
                                            'input':{}, 'outputs':{}, 
                                            'stats':{'s':0, 'w':0, 'c':0} }

class ISTextCore(ISCore):

    def __init__(self, config_file, db=None, verbose=False, debug=False, commit_errors=True, script_dir=None):
        self._doc_parsers = None
        self._uploads = None
        self._usr_docs = None
        ISCore.__init__(self, config_file, ISCore.OBJ_TEXT, db, verbose, debug, commit_errors, script_dir)
        self._udl = UrlDownloader()
        self._load_systems_mt()
        self._get_uploads()
        self._get_documents()
 
    def _get_uploads(self): 
        self._uploads = { 'failed':[], 'finished':[], 'running':[], 'new':[] }
        res = self._stg.get_uploads()
        if res['code'] != 0:
            raise ISCoreException(res['info'])
        for r in res['data']:
            manif = json.loads(base64.b64decode(r['manifest']), encoding='utf-8')
            control = None
            if r['control_data'] != None:
                 control = json.loads(base64.b64decode(r['control_data']), encoding='utf-8')
            up = TextUpload(r['id'], r['op_code'], r['ts'], r['last_ts'], manif, r['user_id'], status=r['up_status'], info=r['info'], control_data=control, cancel_request=r['cancel_request'], debug=r['debug'])
            if self._debug and not(up.debug): # do not process regular upload under debug mode
                #self._log("[Debug mode] Skipping regular upload ID '%s'" % up.id)
                continue
            elif not(self._debug) and up.debug: # do not process debug uploads under normal mode 
                #self._log("Skipping debug upload ID '%s' " % up.id)
                continue
            if up.status in up.ERR_STCODES:
                self._uploads['failed'].append(up)
            elif up.status in up.FIN_STCODES:
                self._uploads['finished'].append(up)
            elif up.status in up.RUN_STCODES:
                self._uploads['running'].append(up)
            elif up.status == up.STCODE_INGESTED:
                self._uploads['new'].append(up)
            else:
                raise ISCoreException("Unknown status code %d for Upload ID %s" % (up.status, up.id))
 
    def _get_documents(self):
        res = self._stg.get_documents()
        if res['code'] != 0:
            raise ISCoreException(res['info'])
        ret = res['data']
        self._usr_docs = {}
        for r in ret:
            user_id = r['owner_id']
            obj_id = r['external_id']
            try:
                self._usr_docs[user_id][obj_id] = r
            except:
                self._usr_docs[user_id] = { obj_id:r }

    def ingest_update(self, up):
        res = self._stg.update_upload__new(up.id, up.status, up.info, up.control, up.error_code)
        if res['code'] != 0:
            self._error("Cannot update upload.", exit_code=res['code'])
        self._log("[Status Update] %s" %(up.info))
        # Send Mail
        if self._mailer != None and up.manifest.get('from_webui', False):
            user_email = up.control["user"]["email"] 
            user_name = up.control["user"]["name"]
            if up.op_code == TextUpload.OP_CODE_NEW:
                language = up.manifest["language"]
                documents = up.control["quota"]["d"]
                sentences = up.control["quota"]["s"]
                words = up.control["quota"]["w"]
                if up.status == TextUpload.STCODE_TL_PROC and up.prev_status == TextUpload.STCODE_INGESTED:
                    if user_email != None and user_email != "" and self._cfg['MAI']['SCSM']:
                        res = self._mailer.send_mail_user_processing(up.id, up.ts, language, documents, sentences, words, [user_email])
                        if res['code'] != 0:
                            self._warn("Cannot send user processing mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                            self._print_excpt()
                    if self._cfg['MAI']['SASM'] and self._cfg['MAI']['AAS'] != None:
                        res = self._mailer.send_mail_admin_processing(up.id, up.ts, language, documents, sentences, words, user_name)
                        if res['code'] != 0:
                            self._warn("Cannot send admin processing mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                            self._print_excpt()
                elif up.status == TextUpload.STCODE_NEW_FINISHED:
                    if user_email != None and user_email != "" and self._cfg['MAI']['SCFM']:
                        res = self._mailer.send_mail_user_finished(up.id, up.ts, language, documents, sentences, words, [user_email])
                        if res['code'] != 0:
                            self._warn("Cannot send user finished mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                            self._print_excpt()
                    if self._cfg['MAI']['SAFM'] and self._cfg['MAI']['AAF'] != None:
                        res = self._mailer.send_mail_admin_finished(up.id, up.ts, language, documents, sentences, words, user_name)
                        if res['code'] != 0:
                            self._warn("Cannot send admin finished mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                            self._print_excpt()
     
   
    def ingest_error(self, up, user=True):
        self._log("[Ingest Error %d] %s\n" %(up.error_code, up.info.encode('utf-8')))
        self._print_excpt(send_mail=False)
        if self._debug:
            self._log("Control JSON:")
            sys.stderr.write("\n%s\n" % json.dumps(up.control, indent=4, encoding='utf-8'))
        if self._commit_errors:
            user_id = up.control['user']['id']
            docs = up.control['quota']['d']
            sents = up.control['quota']['s']
            words = up.control['quota']['w']
            # Restore quota
            res = self._stg.restore_user_quota_documents(user_id, docs, sents, words)
            if res['code'] != 0:
                return res
            # Update upload status
            res = self._stg.update_upload__new(up.id, up.status, up.info, up.control, up.error_code)
            if res['code'] != 0:
                self._error("Cannot update upload.", exit_code=res['code'])
            self._log("[Status Update] %s" %(up.info))
            # Send Mail
            if self._mailer != None:
                user_email = up.control["user"]["email"] 
                user_name = up.control["user"]["name"]
                if user and user_email != None and user_email != "" and self._cfg['MAI']['SCEM']:
                    res = self._mailer.send_mail_user_error(up.id, up.ts, up.error_code, up.info, [user_email])
                    if res['code'] != 0:
                        self._warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                        self._print_excpt()
                if self._cfg['MAI']['SAEM'] and self._cfg['MAI']['AAE'] != None:
                    res = self._mailer.send_mail_admin_error(up.id, up.ts, up.error_code, up.info, user_name)
                    if res['code'] != 0:
                        self._warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                        self._print_excpt()
        
    def document_exists(self, user_id, obj_id):
        if user_id not in self._usr_docs.keys():
            return False
        elif obj_id not in self._usr_docs[user_id].keys():
            return False
        return True

    def check_documents(self, up):
        # For each document
        for d in up.manifest['documents']:
            obj_id = d['external_id']
            obj_hash = self._stg.get_hash(obj_id, up.user_id)
            in_dtlx = up.control['ids'][obj_id]['hash'] = obj_hash
            exists = self.document_exists(up.user_id, obj_id)
            if up.op_code == TextUpload.OP_CODE_NEW:
                if exists:
                    return {'code':2224, 'info':'On new operation: Document ID "%s" already exists.' % obj_id}
            elif up.op_code in (TextUpload.OP_CODE_UPDATE, TextUpload.OP_CODE_DELETE):
                if not(exists):
                    return {'code':2225, 'info':'On update/delete operation: Document ID "%s" does not exist.' % obj_id}
                for oup in self._uploads['running']:
                    if obj_id in oup.control['ids'].keys():
                        return {'code':2226, 'info':'On update/delete operation: Document ID "%s" is already being processed under Upload ID "%s".' % (obj_id, oup.id)}
        return {'code':0}
 
    def get_input_documents(self, up):
        language = up.manifest['language']
        # Extract MP ZIP file, if exists
        res = self._stg.extract_media_package(up.id)
        if res['code'] != 0:
            return res
        exists_mpf = res['exists']
        all_sents = 0; all_words = 0; all_chars = 0
        # For each document
        for d in up.manifest['documents']:
            obj_id = d['external_id']
            obj_hash = up.control['ids'][obj_id]['hash']
            # Create outputs subdir
            res = self._stg.create_outputs_subdir(up.id, obj_hash)
            if res['code'] != 0:
                return res
            tdir =  res['data']
            newf = False
            ext = None
            # Document files attached in the MP file
            if 'filename' in d:
                if not(exists_mpf):
                    return {'code': 2222, 'info':'Missing media package file'}
                newf = True
                fn = d['filename']
                ext = d['fileformat']
                res = self._stg.check_mp_file(up.id, fn, d['md5'])
                if res['code'] != 0:
                    return res
                sfp = self._stg.media_package_file_abspath(fn, up.id)
                tfn = "original.%s.%s" % (language, ext)
                tfp = self._stg.get_outputs_subdir_file_abspath(up.id, obj_hash, tfn)
                res = self._stg.copy_file(sfp, tfp)
                if res['code'] != 0:
                    return res
                up.control['ids'][obj_id]['outputs'][language] = {}
                up.control['ids'][obj_id]['outputs'][language]['original'] = { 'fp':tfp, 'format':ext }
                up.control['ids'][obj_id]['input']['original'] = { 'fp':tfp, 'format':ext }
             # URL files
            elif 'url' in d:
                newf = True
                tfn = "original.%s" % (language)
                tfp = self._stg.output_file_abspath(tfn, up.id, obj_hash)
                res = self._udl.download_url(d['url'], tfp, "html")
                if res['code'] != 0:
                    return res
                ext = res['data'][0]
                up.control['ids'][obj_id]['outputs'][language] = {}
                up.control['ids'][obj_id]['outputs'][language]['original'] = { 'fp':tfp, 'format':ext }
                up.control['ids'][obj_id]['input']['original'] = { 'fp':tfp, 'format':ext }
            if ext == None:
                user_id = up.control['user']['id']
                if obj_id in self._usr_docs[user_id]:
                    ext = self._usr_docs[user_id][obj_id]['format']
            if ext != None and ext not in self._doc_parsers.keys():
                return {'code':2227, 'info':'Unsupported document format "%s"' % ext}
            # New document file: parse it to generate internal html and dtlx files
            up.control['ids'][obj_id]['format'] = ext
            d['fileformat'] = ext
            up.control['ids'][obj_id]['new_file'] = newf
            if newf:
                in_doc = up.control['ids'][obj_id]['input']['original']['fp']
                out_html_fp = self._stg.output_file_abspath('parsed.%s.html' % language, up.id, obj_hash)
                out_dtlx_fp = self._stg.output_file_abspath('internal.%s.dtlx' % language, up.id, obj_hash)
                res = self._doc_parsers[ext].parse_document(in_doc, obj_id, language, out_html_fp, out_dtlx_fp)
                if res['code'] != 0:
                    return res
                up.control['ids'][obj_id]['outputs'][language]['parsed'] = { 'fp':out_html_fp }
                up.control['ids'][obj_id]['input']['parsed'] = { 'fp':out_html_fp, 'new':newf }
                up.control['ids'][obj_id]['outputs'][language]['internal'] = { 'fp':out_dtlx_fp }
                up.control['ids'][obj_id]['input']['internal'] = { 'fp':out_dtlx_fp, 'new':newf }
            # Existing document file: get internal html and dtlx files from DB
            else:
                res = self._stg.get_src_document(obj_hash, Storage.FT_DOC_PARSED)
                if res['code'] != 0:
                    return res
                html_fp = res['data']
                up.control['ids'][obj_id]['input']['parsed'] = { 'fp':html_fp, 'new':newf }
                res = self._stg.get_src_document(obj_hash, Storage.FT_DOC_INTERNAL)
                if res['code'] != 0:
                    return res
                dtlx_fp = res['data']
                up.control['ids'][obj_id]['input']['internal'] = { 'fp':dtlx_fp, 'new':newf }
            # Compute document stats
            res = compute_dtlx_src_stats(up.control['ids'][obj_id]['input']['internal']['fp']) 
            if res['code'] != 0:
                return res
            stats = res['data']
            up.control['ids'][obj_id]['stats']['s'] = stats['sentences'];  all_sents += stats['sentences']
            up.control['ids'][obj_id]['stats']['w'] = stats['words'];      all_words += stats['words']
            up.control['ids'][obj_id]['stats']['c'] = stats['characters']; all_chars += stats['characters']
        up.control['quota']['d'] = len(up.manifest['documents'])
        up.control['quota']['s'] = all_sents
        up.control['quota']['w'] = all_words
        up.control['quota']['c'] = all_chars        
        return {'code':0}
               
    def get_existing_translations(self, up):
        for d in up.manifest["documents"]:
            obj_id = d['external_id']
            obj_hash = up.control['ids'][obj_id]['hash']
            res = self._stg.get_document_translations(obj_hash)
            if res['code'] != 0:
                return res
            for t in res['data']:
                tid = t['id']
                src_lang = t['from_language']
                trg_lang = t['language']
                pair = "%s-%s" % (src_lang, trg_lang)
                ld = {'id':t['id'], 'sid':t['system_id'], 'from_lang':t['from_language'], 'dtlx_fp':t['dtlx_fp'], 'status':t['status']}
                up.control['ids'][obj_id]['existing_tl']['tl'][tid] = ld
                up.control['ids'][obj_id]['existing_tl']['indexes']['by_src'][src_lang] = tid
                up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg'][trg_lang] = tid
                up.control['ids'][obj_id]['existing_tl']['indexes']['by_pair'][pair] = tid
        return {'code':0}
    
     
    def set_tl_tasks(self, up):
        regenerate = up.manifest.get('re-generate', False)
        force = up.manifest.get('force', False)
        doc_lang = up.manifest['language']
        user_id = up.control['user']['id']
        user_is_admin = up.control['user']['is_admin']

        res = self.get_user_custom_default_systems(user_id)
        if res['code'] != 0:
            return res
        duser_def_sys = res['data']
        regen_opt = up.manifest.get("re-generate", False)
        force_opt = up.manifest.get("force", False)
        req_langs_opt = up.manifest.get('requested_langs', None)
        use_system_ids = { 'mt':{} }
        if req_langs_opt != None:
            if not(isinstance(req_langs_opt, dict)):
                return {'code':3070, 'info':"Bad format for 'requested_langs' option"}
            for rl in req_langs_opt.keys():
                d = req_langs_opt[rl]
                if not(isinstance(d,dict)):
                    return {'code':3070, 'info':"Bad format for 'requested_langs' option"}
                for k in d.keys():
                    if d[k] == None:
                        d[k] = {}
                if 'tlpath' in d: 
                    pair = "%s-%s" %(doc_lang, rl)
                    lpath = []
                    sl = doc_lang
                    for ds in d['tlpath']:
                        if not(isinstance(ds, dict)):
                            code = 3076
                            info = "Invalid format for 'requested_langs' -> 'tlpath' option"
                            return {'code':code, 'info':info}                           
                        if 'l' not in ds.keys():
                            code = 3076
                            info = "Invalid format for 'requested_langs' -> 'tlpath' option"
                            return {'code':code, 'info':info}
                        ppair = "%s-%s" % (sl, ds['l'])
                        skip = ds.get("skip", False)
                        if ppair not in self._sys['MT'].keys():
                            code = 3077
                            info = "There are not %s MT systems available to indirectly translate media file into %s language" % (ppair, rl)
                            return {'code':code, 'info':info}
                        if ds.get('sid', None) != None:
                            dsys = self._sys['MT'][ppair]['dsys']
                            if ds['sid'] not in dsys:
                                code = 3078
                                info = "MT System ID %s does not exist for translation direction %s" % (str(ds['sid']), ppair)
                                return {'code':code, 'info':info}
                            s = dsys[ds['sid']]
                            if not(s['public']) and not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                                    code = 3079
                                    info = "You are not allowed to explicitly make use of the %s MT System ID %s" % (ppair, str(ds['sid']))
                                    return {'code':code, 'info':info}
                            lpath.append({ 'p':ppair, 'id':ds['sid'], 'skip':skip})
                        else:
                            if ppair in duser_def_sys['mt'].keys():
                                lpath.append({ 'p':ppair, 'id':duser_def_sys['mt'][ppair], 'skip':skip})
                            else:
                                lpath.append({ 'p':ppair, 'id':self._sys['MT'][ppair]['default'], 'skip':skip})
                        sl = ds['l']
                    use_system_ids['mt'][pair] = lpath
                else:
                    clsid = d.get('sid', None)
                    if clsid != None:
                        try:
                            pair = self._sys['MT']['_ids'][clsid]
                        except:
                            code = 3075
                            info = "MT System ID %s does not exist" % (str(clsid))
                            return {'code':code, 'info':info}
                        tl = pair.split("-")[1]
                        if rl != tl:
                            code = 3075
                            info = "MT System ID %s (%s) does not generate translations into the requested language %s" % (str(clsid), pair, rl)
                            return {'code':code, 'info':info}
                    else:
                        pair = "%s-%s" %(doc_lang, rl)
                    if pair not in self._sys['MT'].keys():
                        code = 3073
                        info = "There are not %s MT systems available to directly translate media file into %s language" % (pair, rl)
                        return {'code':code, 'info':info}
                    if clsid != None: # Client has specfied a System ID
                        dsys = self._sys['MT'][pair]['dsys']
                        if clsid not in dsys:
                            code = 3075
                            info = "MT System ID %s does not exist for translation direction %s" % (str(clsid), pair)
                            return {'code':code, 'info':info}
                        s = dsys[clsid]
                        if not(s['public']) and not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                            code = 3074
                            info = "You are not allowed to explicitly make use of the %s MT System ID %s" % (pair, str(clsid))
                            return {'code':code, 'info':info}
                        use_system_ids['mt'][pair] = [ { 'p':pair, 'id':clsid } ]
                    else:
                        if pair in duser_def_sys['mt'].keys():
                            use_system_ids['mt'][pair] = [ { 'p':pair, 'id':duser_def_sys['mt'][pair] } ]
                        else:
                            use_system_ids['mt'][pair] = [ { 'p':pair, 'id':self._sys['MT'][pair]['default'] } ]

        else: # Requested Langs no present. Time to guess user thoughts!
            if  up.op_code == TextUpload.OP_CODE_NEW:
                self._warn("'requested langs' option not present. Won't launch any translation process.")
            elif up.op_code == TextUpload.OP_CODE_UPDATE:
                if not(regenerate):
                    self._warn("'requested langs' and 're-generate' options not present. Won't launch any translation process.")
                else:
                    
                    dp = {}
                    gall_langs = []
                    for d in up.manifest["documents"]:
                        obj_id = d['external_id']
                        all_langs = up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg']
                        gall_langs = list(set(gall_langs + all_langs.keys()))
                        for sublang in all_langs:
                            tid = all_langs[sublang]
                            system_id = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['sid']
                            from_lang = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['from_lang']
                            if system_id in self._sys['MT']['_ids'].keys():
                                pair = self._sys['MT']['_ids'][system_id]
                            elif from_lang != None:
                                pair = "%s-%s" % (from_lang, sublang)
                            else:
                                pair = "%s-%s" % (doc_lang, sublang)
                            if pair not in dp.keys() and pair in self._sys['MT'].keys():
                                if system_id == None or self._sys['MT'][pair]['dsys'].get(system_id, None) == None:
                                    if pair in duser_def_sys['mt'].keys():
                                        system_id = duser_def_sys['mt'][pair]
                                    else:
                                        system_id = self._sys['MT'][pair]['default']
                                dp[pair] = { 'p':pair, 'id':system_id }
                    pathlist = []
                    self.do_infer_tlpaths(dp.keys(), pathlist)
                    for path in pathlist:
                        sl = path[0].split("-")[0]
                        tl = path[-1].split("-")[1]
                        pair = "%s-%s" % (sl, tl)
                        pairpath = []
                        for spair in path:
                            pairpath.append(dp[spair])
                        use_system_ids['mt'][pair] = pairpath
                    

                    # Check if all translations will have the required inputs
    
                    # Gather Ingest subtitle outputs
                    ingest_outputs = []
                    for pair in use_system_ids['mt'].keys():
                        tl = pair.split("-")[1]
                        ingest_outputs.append(tl)
                    all_langs = list(set([doc_lang] + gall_langs + ingest_outputs))
                    # MT
                    for pair in use_system_ids['mt'].keys():
                        sl = pair.split("-")[0]
                        if sl not in all_langs:
                            code = 3070
                            info = "Impossible to launch %s MT system for document ID '%s' due to the lack of input '%s' text." % (pair, obj_id, sl)
                            return {'code':code, 'info':info}
        up.control['systems'] = use_system_ids                
        return {'code':0}
    

    def prepare_translation(self, up, slang, tlang, system_id, tmp_dir):
        doc_lang = up.manifest['language']
        l_in_dtlx = []
        l_in_txt = []
        l_out_dtlx = []
        l_ex_out_dtlx = []
        l_out_txt = []
        is_html = False
        res = self._stg.create_tmp_tl_pair_dir(up.id, slang, tlang)
        if res['code'] != 0: 
            return res
        for d in up.manifest['documents']:
            in_dtlx = None
            ex_dtlx = None
            obj_id = d['external_id']
            obj_hash = up.control['ids'][obj_id]['hash']
            if up.control['ids'][obj_id]['format'] != 'txt': 
                is_html = True
            # Check existing supervised translations
            if tlang in up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg']:
                tid = up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg'][tlang]
                status = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['status']
                if up.op_code == TextUpload.OP_CODE_UPDATE \
                               and not(up.control['ids'][obj_id]['new_file']) \
                               and not(up.manifest.get('force', False)) \
                               and status == Document.STATUS_MANUAL:
                    continue # Do not retranslate if translation is manual and 'force' is disabled.
                from_lang = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['from_lang']
                if from_lang == slang and status in (Document.STATUS_MANUAL, Document.STATUS_MIXED):
                    ex_dtlx = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['dtlx_fp']

            pair = "%s-%s" % (slang, tlang)
            if slang == doc_lang:
                if up.control['ids'][obj_id]['input']['internal']['new']:
                    in_dtlx = up.control['ids'][obj_id]['input']['internal']['fp']
                elif pair in up.control['ids'][obj_id]['existing_tl']['indexes']['by_pair']:
                    tid = up.control['ids'][obj_id]['existing_tl']['indexes']['by_pair'][pair]
                    in_dtlx = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['dtlx_fp']
                else:
                    in_dtlx = up.control['ids'][obj_id]['input']['internal']['fp']
            else:
                if slang in up.control['ids'][obj_id]['outputs']:
                   in_dtlx = up.control['ids'][obj_id]['outputs'][slang]['internal']['fp']
                elif pair in up.control['ids'][obj_id]['existing_tl']['indexes']['by_pair']:
                    tid = up.control['ids'][obj_id]['existing_tl']['indexes']['by_pair'][pair]
                    in_dtlx = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['dtlx_fp']
                elif slang in up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg']:
                   tid = up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg'][slang]
                   in_dtlx = up.control['ids'][obj_id]['existing_tl']['tl'][tid]['dtlx_fp']
                else:
                   return {'code':2228, 'info':"Can't find input text for generating a %s-%s translation for document ID %s" % (slang, tlang, obj_id)}

            l_in_dtlx.append(in_dtlx)
            if in_dtlx == ex_dtlx:
                l_ex_out_dtlx.append(None) # Input = Existing. Nosense!
            else:
                l_ex_out_dtlx.append(ex_dtlx)

            res = self._stg.create_tmp_tl_pair_object_dir(up.id, slang, tlang, obj_hash)
            if res['code'] != 0: 
                return res
            in_txt = self._stg.tmp_tl_pair_object_file_abspath(up.id, slang, tlang, obj_hash, "in.txt")
            l_in_txt.append(in_txt)
            out_txt = self._stg.tmp_tl_pair_object_file_abspath(up.id, slang, tlang, obj_hash, "out.txt")
            l_out_txt.append(out_txt)
            out_dtlx = self._stg.output_file_abspath('internal.tl.%s-%s.dtlx' % (slang, tlang), up.id, obj_hash)
            l_out_dtlx.append(out_dtlx)
            out_html = self._stg.output_file_abspath('parsed.tl.%s-%s.html' % (slang, tlang), up.id, obj_hash)
            fformat = up.control['ids'][obj_id]['format']
            out_original = self._stg.output_file_abspath('original.tl.%s-%s.%s' % (slang, tlang, fformat), up.id, obj_hash)
            up.control['ids'][obj_id]['outputs'][tlang] = { 'from_lang':slang, 'system_id':system_id, 'internal':{'fp':out_dtlx}, 'parsed':{'fp':out_html}, 'original':{'fp':out_original, 'format':fformat} }
        # Write file lists
        l_in_dtlx_fp = "%s/in_dtlx.lst" % tmp_dir
        self._stg.write_list(l_in_dtlx, l_in_dtlx_fp)
        l_in_txt_fp = "%s/in_txt.lst" % tmp_dir
        self._stg.write_list(l_in_txt, l_in_txt_fp)
        l_out_txt_fp = "%s/out_txt.lst" % tmp_dir
        self._stg.write_list(l_out_txt, l_out_txt_fp)
        l_out_dtlx_fp = "%s/out_dtlx.lst" % tmp_dir
        self._stg.write_list(l_out_dtlx, l_out_dtlx_fp)
        l_ex_out_dtlx_fp = "%s/out_ex_dtlx.lst" % tmp_dir
        self._stg.write_list(l_ex_out_dtlx, l_ex_out_dtlx_fp)
        return {'code':0, 'data':(len(l_in_dtlx), l_in_dtlx_fp, l_in_txt_fp, l_out_txt_fp, l_out_dtlx_fp, l_ex_out_dtlx_fp, is_html)}
 
    def generate_formatted_translations(self, up):
        doc_lang = up.manifest['language']
        for d in up.manifest['documents']:
            obj_id = d['external_id']
            obj_hash = up.control['ids'][obj_id]['hash']
            in_html = up.control['ids'][obj_id]['input']['parsed']['fp']
            for lang in filter(lambda l: l != doc_lang, up.control['ids'][obj_id]['outputs'].keys()):
                out_dtlx = up.control['ids'][obj_id]['outputs'][lang]['internal']['fp']
                out_html = up.control['ids'][obj_id]['outputs'][lang]['parsed']['fp']    
                out_original = up.control['ids'][obj_id]['outputs'][lang]['original']['fp']    
                fformat = up.control['ids'][obj_id]['format']
                res = self._doc_parsers[fformat].regenerate_document(out_dtlx, in_html, out_html, out_original)
                if res['code'] != 0:
                    return res
        return {'code':0}
            
 
    def generate_empty_translations(self, up, slang, tlang):
        for d in up.manifest['documents']:
            obj_id = d['external_id']
            obj_hash = up.control['ids'][obj_id]['hash']
            in_dtlx = up.control['ids'][obj_id]['input']['internal']['fp']
            in_html = up.control['ids'][obj_id]['input']['parsed']['fp']
            out_dtlx = self._stg.output_file_abspath('internal.tl.%s-%s.dtlx' % (slang, tlang), up.id, obj_hash)
            res = create_empty_dtlx(in_dtlx, out_dtlx, obj_id, slang, tlang)
            if res['code'] != 0:
                return res
            out_html = self._stg.output_file_abspath('parsed.tl.%s-%s.html' % (slang, tlang), up.id, obj_hash)
            fformat = up.control['ids'][obj_id]['format']
            out_original = self._stg.output_file_abspath('original.tl.%s-%s.%s' % (slang, tlang, fformat), up.id, obj_hash)
            res = self._doc_parsers[fformat].regenerate_document(out_dtlx, in_html, out_html, out_original)
            if res['code'] != 0:
                return res
            up.control['ids'][obj_id]['outputs'][tlang] = { 'from_lang':slang, 'system_id':None, 
                                                            'internal':{'fp':out_dtlx}, 
                                                            'parsed':{'fp':out_html}, 
                                                            'original':{'fp':out_original, 'format':fformat} }
        return {'code':0}

     
    def launch_translation(self, up, system_id, slang, tlang, wait_jobs=[]):
        if up.manifest.get("test_mode", False):
            if self._verbose: self._log("Test mode: will generate an empty %s translation ..." % (tlang))
            res = self.generate_empty_translations(up, slang, tlang)
            if res['code'] != 0:
                return res
            return {'code':0, 'jobs':[]}
        else:
            pair = "%s-%s" % (slang, tlang)
            if self._verbose: self._log("Launching '%s' translation system (System ID %d) ..." % (pair, system_id))
            tmp_dir = self._stg.tmp_tl_pair_dir(up.id, slang, tlang)
            res = self.prepare_translation(up, slang, tlang, system_id, tmp_dir)
            if res['code'] != 0:
                return res
            ndocs, l_in_dtlx_fp, l_in_txt_fp, l_out_txt_fp, l_out_dtlx_fp, l_ex_out_dtlx_fp, is_html = res['data']
            if ndocs == 0:
                if self._verbose: self._log("\t- No documents to be translated.")
                return {'code':0, 'jobs':[]}
            job_prefix = "%s.%s.%s.%s-%s" % (self._cfg['SCH']['JOB_NAME_PREFIX'], self._cfg['STO']['DB_NAME'], up.id, slang, tlang)
            all_jids = []
            # dtlx2txt
            dtlx2txt_tmp_dir = "%s/dtlx2txt" % tmp_dir
            my_job_name = "%s.dtlx2txt" % job_prefix
            log = "%s/dtlx2txt.log" % tmp_dir
            if up.manifest.get("force", False):
                l_ex_out_dtlx_fp = None # If force == True, do not take into account existing translations
            res = dtlx2txt(self._jbs, slang, tlang, l_in_dtlx_fp, l_in_txt_fp, l_out_dtlx_fp, dtlx2txt_tmp_dir, ex_dtlx_lst=l_ex_out_dtlx_fp, log_file=log, job_name=my_job_name, hold_job_ids=wait_jobs) 
            if res['code'] != 0:
                return res
            jids = res['jids']
            all_jids += jids
            if self._debug:
                self._log("DTLX2TXT process launched:")
                self._log("\tCommand: %s" % res['cmd'])
                self._log("\tJob ID(s): %s" % str(jids))
                self._log("\tLog file: %s" % log)
            # Translation
            mt = self._sys['MT'][pair]['dsys'][system_id]['module'] # MT module
            my_tmp_dir = "%s/translate" % tmp_dir
            my_job_name = "%s.translate" % job_prefix
            log = "%s/translate.log" % tmp_dir
            res = mt.translate(self._jbs, l_in_txt_fp, l_out_txt_fp, my_tmp_dir, html=is_html, log_file=log, job_name=my_job_name, hold_job_ids=jids)
            if res['code'] != 0:
                return res
            jids = res['jids']
            all_jids += jids
            if self._debug:
                self._log("Translation process launched:")
                self._log("\tCommand: %s" % res['cmd'])
                self._log("\tJob ID(s): %s" % str(jids))
                self._log("\tLog file: %s" % log)
            # txt2dtlx
            my_tmp_dir = "%s/txt2dtlx" % tmp_dir
            my_job_name = "%s.txt2dtlx" % job_prefix
            log = "%s/txt2dtlx.log" % tmp_dir
            res = txt2dtlx(self._jbs, tlang, dtlx2txt_tmp_dir, l_out_txt_fp, l_out_dtlx_fp, my_tmp_dir, log_file=log, job_name=my_job_name, hold_job_ids=jids) 
            if res['code'] != 0:
                return res
            jids = res['jids']
            all_jids += jids
            if self._debug:
                self._log("TXT2DTLX process launched:")
                self._log("\tCommand: %s" % res['cmd'])
                self._log("\tJob ID(s): %s" % str(jids))
                self._log("\tLog file: %s" % log)
            up.control['jobs'] += all_jids
            return {'code':0, 'jobs':all_jids}
 
    
    def launch_translations(self, up):
        jobs = []
        pair_jobids = {}
        target_jobids = {}
        sys = up.control['systems']
        for pair in self.mtsort(sys['mt'].keys()):
            sl = pair.split('-')[0]
            tl = pair.split('-')[1]
            tlpath = sys['mt'][pair]
            if len(tlpath) == 1:
                if pair in pair_jobids:
                    if self._verbose: self._log("%s translation was explicitly requested but '%s' translation system was already launched." % (tl, pair))
                else:
                    system_id = sys["mt"][pair][0]['id']
                    wait_jobs = target_jobids.get(sl, [])
                    res = self.launch_translation(up, system_id, sl, tl, wait_jobs=wait_jobs)
                    if res['code'] != 0:
                        return res
                    jobs = jobs + res['jobs']
                    pair_jobids[pair] = res['jobs']
                    target_jobids[tl] = res['jobs']
            else:
                if self._verbose: self._log("Will Launch several MT systems to generate a '%s' translation..." % (pair))
                wait_jobs = target_jobids.get(sl, [])
                for step in tlpath:
                    spair = step['p']
                    ssl = spair.split('-')[0]
                    stl = spair.split('-')[1]
                    if step.get('skip', False):
                        if self._verbose: self._log(" - Skipping translation step '%s' ('skip' option provided)." % (spair))
                    elif spair in pair_jobids:
                        if self._verbose: self._log(" - '%s' translation system was already launched." % (stl, spair))
                        wait_jobs = pair_jobids[spair]
                    else:
                        system_id = step['id']
                        if self._verbose: self._log(" - Launching '%s' translation system (System ID %d) ..." % (spair, system_id))
                        res = self.launch_translation(up, system_id, ssl, stl, wait_jobs=wait_jobs)
                        if res['code'] != 0:
                            return res
                        jobs = jobs + res['jobs']
                        wait_jobs = res['jobs']         
                        pair_jobids[spair] = res['jobs']
                        target_jobids[stl] = res['jobs']
        return {'code':0, 'jobs':jobs}
    
   
    def store_data(self, up):
        res = self._stg.begin_transaction(up.op_code)
        if res['code'] != 0:
            return res
        doc_lang = up.manifest['language']
        owner_id = up.control['user']['id']
        for d in up.manifest['documents']:
           obj_id = d['external_id'] 
           obj_hash = up.control['ids'][obj_id]['hash']
           fformat = up.control['ids'][obj_id]['format']
           outputs = up.control['ids'][obj_id]['outputs']
           existing_tl = up.control['ids'][obj_id]['existing_tl']['indexes']['by_trg']
           if doc_lang in outputs: 
               title = d.get('title', None)
               src_orig = outputs[doc_lang]['original']['fp']
               src_html = outputs[doc_lang]['parsed']['fp']
               src_dtlx = outputs[doc_lang]['internal']['fp']
               sents = up.control['ids'][obj_id]['stats']['s']
               words = up.control['ids'][obj_id]['stats']['w']
               chars = up.control['ids'][obj_id]['stats']['c']
               if up.op_code == TextUpload.OP_CODE_NEW:
                   res = self._stg.insert_document(obj_hash, obj_id, doc_lang, title, fformat, sents, words, chars, src_orig, src_html, src_dtlx, owner_id)
               else:
                   res = self._stg.update_document(obj_hash, title, fformat, sents, words, chars, src_orig, src_html, src_dtlx)
               if res['code'] != 0:
                   return res
           for lang in filter(lambda l: l != doc_lang, outputs.keys()):
               from_lang = outputs[lang]['from_lang']
               system_id = outputs[lang]['system_id']
               tl_orig = outputs[lang]['original']['fp']
               tl_html = outputs[lang]['parsed']['fp']
               tl_dtlx = outputs[lang]['internal']['fp']
               res = compute_dtlx_tl_stats(tl_dtlx) 
               if res['code'] != 0:
                   return res
               stats = res['data']
               status = stats['status']
               sup_ratio = stats['sup_ratio']
               if lang in existing_tl:
                   tid = existing_tl[lang]
                   res = self._stg.update_translation(obj_hash, tid, status, sup_ratio, from_lang, system_id, fformat, tl_orig, tl_html, tl_dtlx)
               else:
                   res = self._stg.insert_translation(obj_hash, lang, status, sup_ratio, from_lang, system_id, fformat, tl_orig, tl_html, tl_dtlx)
               if res['code'] != 0:
                   return res
        res = self._stg.end_transaction()
        if res['code'] != 0:
           return res
        self._get_documents() # Update internal documents list
        return {'code':0}          

    def check_user_quota(self, up):
        user_id = up.control['user']['id']
        docs = up.control['quota']['d']
        sents = up.control['quota']['s']
        words = up.control['quota']['w']
        res = self._stg.get_user_quota_documents(user_id)
        if res['code'] != 0:
            return res
        if res['data']['q_docs'] != None: # if quota == None (NULL) -> Unlimited quota
            if res['data']['q_docs'] - docs < 0:
                return {'code':2230, 'info':'Unsufficient quota (remaining %d document files, but uploaded %d document files).' % (res['data']['q_docs'], docs)}
            #elif res['data']['q_sents'] - sents < 0:
            #    return {'code':2231, 'info':'Unsufficient quota (remaining %d sentences, but uploaded %d sentences).' % (res['data']['q_sents'], sents)}
            elif res['data']['q_words'] - words < 0:
                return {'code':2232, 'info':'Unsufficient quota (remaining %d words, but uploaded %d words).' % (res['data']['q_words'], words)}
        res = self._stg.update_user_quota_documents(user_id, docs, sents, words)
        if res['code'] != 0:
            return res
        return {'code':0}
    
    ##---------- Upload processing functions -------------#
    
    
    def process_failed_uploads(self):
        for up in self._uploads['failed']:
            try:
                td = (datetime.now() - up.last_ts).days
                if td >= self._cfg['GEN']['RM_ERR_UP_DAYS']:
                    if self._verbose: self._log("[*] Deleting upload ID %s ..." % up.id)
                    up_dir = "%s/%s" % (self._stg.uploads_mp, up.id)
                    res = self._stg.delete_upload(up.id, up_dir)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    if self._verbose: self._log("Upload ID %s succesfully deleted." % up.id)
            except:
                self._bug("Upload ID: %s" % up.id)
                continue
    
    
    def process_finished_uploads(self):
        for up in self._uploads['finished']:
            try:
                td = (datetime.now() - up.last_ts).days
                if td >= self._cfg['GEN']['RM_FIN_UP_DAYS']:
                    if self._verbose: self._log("[*] Deleting upload ID %s ..." % up.id)
                    up_dir = "%s/%s" % (self._stg.uploads_mp, up.id)
                    res = self._stg.delete_upload(up.id, up_dir)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    if self._verbose: self._log("Upload ID %s succesfully deleted." % up.id)
            except:
                self._bug("Upload ID: %s" % up.id)
                continue
    
   
    def cancel_upload(self, up):
        res = jbs.delete_jobs(up.control['jobs'])
        if res['code'] != 0:
            warn(res['info'])
            print_excpt()
        return {'code':0} 

    def process_running_uploads(self):
        # Get Running jobs
        res = self._jbs.get_running_jobs()
        if res['code'] != 0:
            self._error(res['info'], exit_code=res['code'])
        rjobs = res['data']
        # For each upload
        for up in self._uploads['running']:
            try:
                if self._verbose: self._log("[*] Checking running upload ID %s ..." % up.id)
                if up.cancel:
                    if self._verbose: self._log("Oh nooo! Cancel request!")
                    res = self.cancel_upload(up)
                    if res['code'] != 0:
                        up.info = "Could not cancel upload."
                        up.status = TextUpload.STCODE_CANCEL_ERROR
                        up.error_code = res['code']
                        self.ingest_error(up)
                        continue
                    up.info = "Upload successfully cancelled."
                    up.status = TextUpload.STCODE_CANCELLED
                    self.ingest_update(up)
                    continue
                res = self.check_running_processes(rjobs, up.control['jobs'])
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                job_status = res['data']
                if job_status == Scheduler.ST_QUEUED:
                    if self._verbose: self._log("Processes queued.")
                    up.info = "Processes queued."
                    self.ingest_update(up)
                elif job_status == Scheduler.ST_RUNNING:
                    if self._verbose: self._log("Processes running.")
                    up.info = "Processes running."
                    self.ingest_update(up)
                elif job_status == Scheduler.ST_ERROR:
                    if up.status == TextUpload.STCODE_TL_PROC:
                        up.info = "Translation Error."
                        up.status = TextUpload.STCODE_TL_ERROR 
                        up.error_code = 2001
                        self.ingest_error(up)
                        continue
                    else:
                        self._bug("Unexpected upload status %d" % (job_status), exit=True)
                elif job_status == Scheduler.ST_FINISHED:
                    if up.status == TextUpload.STCODE_TL_PROC:
                        if self._verbose: self._log("Translation processes finished.")
                        ret = self.generate_formatted_translations(up)
                        if ret['code'] != 0:
                            up.info = ret['info']
                            up.status = TextUpload.STCODE_DB_ERROR 
                            up.error_code = ret['code']
                            self.ingest_error(up)
                            continue
                        ret = self.store_data(up)
                        if ret['code'] != 0:
                            up.info = ret['info']
                            up.status = TextUpload.STCODE_DB_ERROR 
                            up.error_code = ret['code']
                            self.ingest_error(up)
                            continue
                        if up.op_code == TextUpload.OP_CODE_NEW:
                            up.info = "New document(s) successfully processed."
                            up.status = TextUpload.STCODE_NEW_FINISHED 
                        else:
                            up.info = "Document(s) successfully updated."
                            up.status = TextUpload.STCODE_UPD_FINISHED 
                        self.ingest_update(up)
                    else:
                        self._bug("Unexpected upload status %d" % (job_status), exit=True)
            except:
                self._bug("Upload ID: %s" % up.id)
                continue

    def delete_documents(self, up):
        res = self._stg.begin_transaction(up.op_code)
        if res['code'] != 0:
            return res
        for d in up.manifest['documents']:
            obj_id = d['external_id']
            obj_hash = up.control['ids'][obj_id]['hash']
            user_id = up.control['user']['id']
            obj_int_id = self._usr_docs[user_id][obj_id]['id']
            res = self._stg.delete_document(obj_int_id, obj_hash)
            if res['code'] != 0:
                return res
        res = self._stg.end_transaction()
        if res['code'] != 0:
            return res
        self._get_documents() # Update internal documents list
        return {'code':0}


    def process_new_uploads(self):
        for up in self._uploads['new']:
            try:
                if self._verbose: 
                    self._log("[*] Processing new upload ID %s ..." % up.id)
                if self._debug:
                    self._log("Manifest JSON:")
                    sys.stderr.write("\n%s\n" % json.dumps(up.manifest, indent=4, encoding='utf-8'))
                # Get user info
                res = self.get_user_info(up.user_id)
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                up.control['user'] = res['data']
                if up.cancel:
                    if self._verbose: self._log("Oh nooo! Cancel request!")
                    up.info = "Upload successfully cancelled."
                    up.status = TextUpload.STCODE_CANCELLED
                    self.ingest_update(up)
                    continue
                if up.op_code == up.OP_CODE_DELETE:
                    ret = self.check_documents(up)
                    if ret['code'] != 0:
                        up.info = ret['info']
                        up.status = TextUpload.STCODE_MP_ERROR 
                        up.error_code = ret['code']
                        self.ingest_error(up)
                        continue
                    ret = self.delete_documents(up)
                    if ret['code'] != 0:
                        up.info = ret['info']
                        up.status = TextUpload.STCODE_MP_ERROR 
                        up.error_code = ret['code']
                        self.ingest_error(up)
                        continue
                    up.info = "Document(s) successfully deleted."
                    up.status = TextUpload.STCODE_DEL_FINISHED 
                    self.ingest_update(up)
                elif up.op_code in (up.OP_CODE_NEW, up.OP_CODE_UPDATE):
                    # Create TMP dir
                    res = self._stg.create_tmp_dirs(up.id)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    # Create Outputs dir
                    res = self._stg.create_outputs_dir(up.id)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    # Check documents
                    ret = self.check_documents(up)
                    if ret['code'] != 0:
                        up.info = ret['info']
                        up.status = TextUpload.STCODE_MP_ERROR 
                        up.error_code = ret['code']
                        self.ingest_error(up)
                        continue
                    # Get input documents
                    ret = self.get_input_documents(up)
                    if ret['code'] != 0:
                        up.info = ret['info']
                        up.status = TextUpload.STCODE_MP_ERROR 
                        up.error_code = ret['code']
                        self.ingest_error(up)
                        continue
                    if up.manifest.get('generate', True):
                        if up.op_code == TextUpload.OP_CODE_UPDATE:
                            ret = self.get_existing_translations(up)
                            if ret['code'] != 0:
                                up.info = ret['info']
                                up.status = TextUpload.STCODE_MP_ERROR 
                                up.error_code = ret['code']
                                self.ingest_error(up)
                                continue
                        else:
                            if not(up.manifest.get('test_mode', False)):
                                ret = self.check_user_quota(up) # only for new operations
                                if ret['code'] != 0:
                                    up.info = ret['info']
                                    up.status = TextUpload.STCODE_MP_ERROR 
                                    up.error_code = ret['code']
                                    self.ingest_error(up)
                                    continue
                        # Determine translation tasks
                        ret = self.set_tl_tasks(up)
                        if ret['code'] != 0:
                            up.info = ret['info']
                            up.status = TextUpload.STCODE_MP_ERROR 
                            up.error_code = ret['code']
                            self.ingest_error(up)
                            continue
                        # Prepare and launch translation tasks
                        ret = self.launch_translations(up)
                        if ret['code'] != 0:
                            up.info = ret['info']
                            up.status = TextUpload.STCODE_MP_ERROR 
                            up.error_code = ret['code']
                            self.ingest_error(up)
                            continue
                        jids = ret['jobs']
                        if len(jids) > 0:
                            up.control['jobs'] += jids
                            up.info = "Translation(s) in progress. It may take several hours until it finishes."
                            up.status = TextUpload.STCODE_TL_PROC 
                            self.ingest_update(up)
                        else:
                            if self._verbose: self._log("No translation processes were launched.")
                            ret = self.store_data(up)
                            if ret['code'] != 0:
                                up.info = ret['info']
                                up.status = TextUpload.STCODE_DB_ERROR 
                                up.error_code = ret['code']
                                self.ingest_error(up)
                                continue
                            if up.op_code == TextUpload.OP_CODE_NEW:
                                up.info = "New document(s) successfully processed."
                                up.status = TextUpload.STCODE_NEW_FINISHED 
                            else:
                                up.info = "Document(s) successfully updated."
                                up.status = TextUpload.STCODE_UPD_FINISHED 
                            self.ingest_update(up)
                    else: # Generate option disabled:
                        if self._verbose: self._log("Generate option was disabled.")
                        ret = self.store_data(up)
                        if ret['code'] != 0:
                            up.info = ret['info']
                            up.status = TextUpload.STCODE_DB_ERROR 
                            up.error_code = ret['code']
                            self.ingest_error(up)
                            continue
                        if up.op_code == TextUpload.OP_CODE_NEW:
                            up.info = "New document(s) successfully processed."
                            up.status = TextUpload.STCODE_NEW_FINISHED 
                        else:
                            up.info = "Document(s) successfully updated."
                            up.status = TextUpload.STCODE_UPD_FINISHED 
                        self.ingest_update(up)
                    # Append TextUpload object to "running", 
                    # to avoid processing the same Doc IDs in different upload IDs (see check_documents())
                    self._uploads['running'].append(up)
                    if self._debug:
                        self._log("Control JSON:")
                        sys.stderr.write("\n%s\n" % json.dumps(up.control, indent=4, encoding='utf-8'))
                else:
                     # Unknown OPCODE
                     self._bug("Unkown operation code %d" % up.op_code)

            except:
                self._bug("Upload ID: %s" % up.id)
                continue
    


if __name__ == "__main__":

    DEFAULT_CFG_FP = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))
    DEFAULT_CFG_FN = "config.ini"
    SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

    # Define Option Parser
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option("-v", "--verbose", dest="verb", action="store_true", default=False, help="Verbose power on!")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="Debug mode")
    parser.add_option("-e", "--disable-errors", dest="errors", action="store_false", default=True, help="Disable ingest errors")
    parser.add_option("-C", "--config-file", dest="config_file", type="string", default=DEFAULT_CFG_FP, help="Configuration file. Default: %s" % DEFAULT_CFG_FN)
    parser.add_option("-D", "--database", dest="db_name", type="string", default=None, help="Database with which to work. Default: specified in config file")
    (opts, args) = parser.parse_args()


    core = ISTextCore(opts.config_file, verbose=opts.verb, debug=opts.debug, commit_errors=opts.errors, script_dir=SCRIPT_DIR)

    try:
        # Remove finished uploads
        if opts.verb: core._log("----------- Checking failed uploads ------------")
        core.process_failed_uploads()
    
        # Remove finished uploads
        if opts.verb: core._log("---------- Checking finished uploads -----------")
        core.process_finished_uploads()
    
        # Check running uploads
        if opts.verb: core._log("----------- Checking running uploads -----------")
        core.process_running_uploads()
    
        # Check new uploads
        if opts.verb: core._log("------------ Checking new uploads --------------")
        core.process_new_uploads()
    
        if opts.verb: core._log("------------------- EXIT 0 ---------------------")
    except: 
        out = StringIO()
        traceback.print_exc(file=out)
        exception_str = out.getvalue().strip()
        print exception_str
        if core._mailer != None:
            core._mailer.send_mail_ingest_bug(exception_str)

