# -*- coding: utf-8 -*-
#
#  Copyright 2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

#---------- Module Imports -------------#

# Standard Libraries
import sys
import os
import traceback
import re
import json
import base64
from optparse import OptionParser
from datetime import datetime

# Custom module imports
from modules.base.text import TextSplitter
from ISCore import ISCore, ISCoreException
from scheduler import Scheduler
from storage import Storage

class WebUpload():

    OP_CODE_NEW = 0
    OP_CODE_UPDATE = 1
    OP_CODE_DELETE = 2
    OP_CODE_CANCEL = 3
    
    STCODE_INGESTED = 0
    STCODE_MP_PROC = 1
    STCODE_TL_PROC = 2
    STCODE_NEW_FINISHED = 3
    STCODE_UPD_FINISHED = 20
    STCODE_DEL_FINISHED = 30
    STCODE_CANCELLED = 40
    STCODE_UNK_ERROR = 100
    STCODE_MP_ERROR = 101
    STCODE_TL_ERROR = 102
    STCODE_DB_ERROR = 103
    STCODE_UPD_ERROR = 200
    STCODE_DEL_ERROR = 300
    STCODE_CANCEL_ERROR = 400
    ERR_STCODES = [ STCODE_UNK_ERROR, STCODE_MP_ERROR, STCODE_TL_ERROR, 
                    STCODE_DB_ERROR, STCODE_UPD_ERROR, STCODE_DEL_ERROR, 
                    STCODE_CANCEL_ERROR ]
    RUN_STCODES = [ STCODE_MP_PROC, STCODE_TL_PROC ]
    FIN_STCODES = [ STCODE_NEW_FINISHED, STCODE_UPD_FINISHED, STCODE_DEL_FINISHED, STCODE_CANCELLED ]

    def __init__(self, up_id, op_code, ts, last_ts, web_key, manifest, user_id, status=None, info=None, control_data=None, error_code=None, cancel_request=False, debug=False):
        self.id = up_id
        self.web_key = web_key
        self.op_code = op_code
        self.status = status
        self.prev_status = status
        self.info = info
        self.error_code = error_code
        self.ts = ts 
        self.last_ts = last_ts 
        self.user_id = user_id
        self.cancel = cancel_request
        self.debug = debug
        self.manifest = manifest
        if control_data == None:
            self._init_control()
        else:
            self.control = control_data
 
    def __repr__(self):
        return "([WebUpload Object] <id = %s> <op_code = %d> <status = %d> <info = %s>)" % (self.id, self.op_code, self.status, self.info)

    def _init_control(self):
        self.control = { 'jobs':{'all':[], 'lang':{}}, 'website':{}, 'texts':{}, 'user':{}, 'systems':{}, 'quota':{'w':0, 'c':0} }

class ISWebCore(ISCore):

    def __init__(self, config_file, db=None, verbose=False, debug=False, commit_errors=True, script_dir=None):
        ISCore.__init__(self, config_file, ISCore.OBJ_WEB, db, verbose, debug, commit_errors, script_dir)
        self._uploads = None
        self._load_systems_mt()
        self._get_uploads()
 

    def _get_uploads(self): 
        self._uploads = { 'failed':[], 'finished':[], 'running':[], 'new':[] }
        res = self._stg.get_uploads()
        if res['code'] != 0:
            raise ISCoreException(res['info'])
        for r in res['data']:
            manif = json.loads(base64.b64decode(r['manifest']), encoding='utf-8')
            control = None
            if r['control_data'] != None:
                 control = json.loads(base64.b64decode(r['control_data']), encoding='utf-8')
            up = WebUpload(r['id'], r['op_code'], r['ts'], r['last_ts'], r['hash'], manif, r['user_id'], status=r['up_status'], info=r['info'], control_data=control, cancel_request=r['cancel_request'], debug=r['debug'])
            if self._debug and not(up.debug): # do not process regular upload under debug mode
                #self._log("[Debug mode] Skipping regular upload ID '%s'" % up.id)
                continue
            elif not(self._debug) and up.debug: # do not process debug uploads under normal mode 
                #self._log("Skipping debug upload ID '%s' " % up.id)
                continue
            if up.status in up.ERR_STCODES:
                self._uploads['failed'].append(up)
            elif up.status in up.FIN_STCODES:
                self._uploads['finished'].append(up)
            elif up.status in up.RUN_STCODES:
                self._uploads['running'].append(up)
            elif up.status == up.STCODE_INGESTED:
                self._uploads['new'].append(up)
            else:
                raise ISCoreException("Unknown status code %d for Upload ID %s" % (up.status, up.id))
 
    def ingest_update(self, up):
        res = self._stg.update_upload__new(up.id, up.status, up.info, up.control, up.error_code)
        if res['code'] != 0:
            self._error("Cannot update upload.", exit_code=res['code'])
        self._log("[Status Update] %s" %(up.info))
   
    def ingest_error(self, up, user=True):
        self._log("[Ingest Error %d] %s\n" %(up.error_code, up.info.encode('utf-8')))
        self._print_excpt(send_mail=False)
        if self._debug:
            self._log("Control JSON:")
            sys.stderr.write("\n%s\n" % json.dumps(up.control, indent=4, encoding='utf-8'))
        if self._commit_errors:
            user_id = up.control['user']['id']
            words = up.control['quota']['w']
            # Restore quota
            res = self._stg.restore_user_quota_documents(user_id, 0, 0, words)
            if res['code'] != 0:
                return res
            # Update upload status
            res = self._stg.update_upload__new(up.id, up.status, up.info, up.control, up.error_code)
            if res['code'] != 0:
                self._error("Cannot update upload.", exit_code=res['code'])
            self._log("[Status Update] %s" %(up.info))
            # Send Mail
            if self._mailer != None:
                user_name = up.control["user"]["name"]
                if self._cfg['MAI']['SAEM'] and self._cfg['MAI']['AAE'] != None:
                    res = self._mailer.send_mail_admin_error(up.id, up.ts, up.error_code, up.info, user_name)
                    if res['code'] != 0:
                        self._warn("Cannot send mail (Error Code: %d, Info: %s)." % (res['code'], res['info']))
                        self._print_excpt()
        
    def text_exists(self, web_id, text_hash):
        if user_id not in self._usr_docs.keys():
            return False
        elif obj_id not in self._usr_docs[user_id].keys():
            return False
        return True

    def check_website(self, up):
        res = self._stg.get_website_info(up.web_key, up.user_id)
        if res['code'] != 0:
            return res
        data = res['data']
        if len(data) == 0:
            return {'code':2283, 'info':'Website Key %s does not exist for user ID %d' % (web_key, user_id)}
        up.control['website']['id'] = data[0]['id']
        up.control['website']['name'] = data[0]['name']
        up.control['website']['language'] = data[0]['language']
        up.control['website']['options'] = data[0]['options']
        return {'code':0}
        

    def check_texts(self, up):
        res = self._stg.get_website_source_texts(up.control['website']['id'], up.manifest['texts'])
        if res['code'] != 0:
            return res
        ret_texts = res['data'] 
        for thash in up.manifest['texts']:
            if thash not in ret_texts.keys():
                return {'code':2225, 'info':'Text hash "%s" does not exist on Website ID %s.' % (thash, up.control['website']['id'])}
            tid = ret_texts[thash][0]
            ttxt = ret_texts[thash][1].decode('utf-8')
            nwords = len(ttxt.strip().split())
            nchars = len(filter(unicode.isalnum, ttxt))
            up.control['texts'][thash] = {'id':tid, 'txt':ttxt, 'stats':{'w':nwords, 'c':nchars}, 'tl':{}, 'outputs':{}}
            res = self._stg.get_website_translated_texts(up.control['texts'][thash]['id'])
            if res['code'] != 0:
                return res
            avail_tl = res['data']
            if len(avail_tl) > 0:
                for tt in avail_tl:
                    ttid = tt['id']
                    ttlang = tt['language']
                    ttstatus = tt['status']
                    tttext = tt['text']
                    up.control['texts'][thash]['tl'][ttlang] = {'id':ttid, 'st':ttstatus, 'txt':tttext}
            else: # No TL available yet, then add quota
                up.control['quota']['w'] += up.control['texts'][thash]['stats']['w']
                up.control['quota']['c'] += up.control['texts'][thash]['stats']['c']
        return {'code':0}
              
     
    def set_tl_tasks(self, up):
        web_lang = up.control['website']['language']
        user_id = up.control['user']['id']
        user_is_admin = up.control['user']['is_admin']
        res = self.get_user_custom_default_systems(user_id)
        if res['code'] != 0:
            return res
        duser_def_sys = res['data']
        regen_opt = up.manifest.get("re-generate", False)
        force_opt = up.manifest.get("force", False)
        req_langs_opt = up.manifest.get('requested_langs', None)
        use_system_ids = { 'mt':{} }
        if req_langs_opt == None:
            code = 3070
            info = "Missing requested_langs option." 
            return {'code':code, 'info':info}
        else: 
            if not(isinstance(req_langs_opt, dict)):
                return {'code':3070, 'info':"Bad format for 'requested_langs' option"}
            for rl in req_langs_opt.keys():
                d = req_langs_opt[rl]
                if not(isinstance(d,dict)):
                    return {'code':3070, 'info':"Bad format for 'requested_langs' option"}
                for k in d.keys():
                    if d[k] == None:
                        d[k] = {}
                if 'tlpath' in d: 
                    pair = "%s-%s" %(web_lang, rl)
                    lpath = []
                    sl = web_lang
                    for ds in d['tlpath']:
                        if not(isinstance(ds, dict)):
                            code = 3076
                            info = "Invalid format for 'requested_langs' -> 'tlpath' option"
                            return {'code':code, 'info':info}                           
                        if 'l' not in ds.keys():
                            code = 3076
                            info = "Invalid format for 'requested_langs' -> 'tlpath' option"
                            return {'code':code, 'info':info}
                        ppair = "%s-%s" % (sl, ds['l'])
                        skip = ds.get("skip", False)
                        if ppair not in self._sys['MT'].keys():
                            code = 3077
                            info = "There are not %s MT systems available to indirectly translate media file into %s language" % (ppair, rl)
                            return {'code':code, 'info':info}
                        sid = ds.get('sid', None)
                        if sid != None:
                            try:
                                sid = int(sid)
                            except:
                                return {'code':3078, 'info':'requested_langs -> "sid" must be an integer'}
                            dsys = self._sys['MT'][ppair]['dsys']
                            if sid not in dsys:
                                code = 3078
                                info = "MT System ID %s does not exist for translation direction %s" % (str(sid), ppair)
                                return {'code':code, 'info':info}
                            s = dsys[sid]
                            if not(s['public']) and not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                                    code = 3079
                                    info = "You are not allowed to explicitly make use of the %s MT System ID %s" % (ppair, str(sid))
                                    return {'code':code, 'info':info}
                            lpath.append({ 'p':ppair, 'id':sid, 'skip':skip})
                        else:
                            if ppair in duser_def_sys['mt'].keys():
                                lpath.append({ 'p':ppair, 'id':duser_def_sys['mt'][ppair], 'skip':skip})
                            else:
                                lpath.append({ 'p':ppair, 'id':self._sys['MT'][ppair]['default'], 'skip':skip})
                        sl = ds['l']
                    use_system_ids['mt'][pair] = lpath
                else:
                    clsid = d.get('sid', None)
                    if clsid != None:
                        try:
                            clsid = int(clsid)
                        except:
                            return {'code':3075, 'info':'requested_langs -> "sid" must be an integer'}
                        try:
                            pair = self._sys['MT']['_ids'][clsid]
                        except:
                            code = 3075
                            info = "MT System ID %s does not exist" % (str(clsid))
                            return {'code':code, 'info':info}
                        tl = pair.split("-")[1]
                        if rl != tl:
                            code = 3075
                            info = "MT System ID %s (%s) does not generate translations into the requested language %s" % (str(clsid), pair, rl)
                            return {'code':code, 'info':info}
                    else:
                        pair = "%s-%s" %(web_lang, rl)
                    if pair not in self._sys['MT'].keys():
                        code = 3073
                        info = "There are not %s MT systems available to directly translate media file into %s language" % (pair, rl)
                        return {'code':code, 'info':info}
                    if clsid != None: # Client has specfied a System ID
                        dsys = self._sys['MT'][pair]['dsys']
                        if clsid not in dsys:
                            code = 3075
                            info = "MT System ID %s does not exist for translation direction %s" % (str(clsid), pair)
                            return {'code':code, 'info':info}
                        s = dsys[clsid]
                        if not(s['public']) and not(user_is_admin) and user_id not in s['allowed_users']: # Not admin and not allowed to use that system
                            code = 3074
                            info = "You are not allowed to explicitly make use of the %s MT System ID %s" % (pair, str(clsid))
                            return {'code':code, 'info':info}
                        use_system_ids['mt'][pair] = [ { 'p':pair, 'id':clsid } ]
                    else:
                        if pair in duser_def_sys['mt'].keys():
                            use_system_ids['mt'][pair] = [ { 'p':pair, 'id':duser_def_sys['mt'][pair] } ]
                        else:
                            use_system_ids['mt'][pair] = [ { 'p':pair, 'id':self._sys['MT'][pair]['default'] } ]
            up.control['systems'] = use_system_ids                
        return {'code':0}
    

    def generate_empty_translations(self, up, tlang):
        for thash in up.control['texts']:
           if tlang not in up.control['texts'][thash]['tl']:
               source_text_id = up.control['texts'][thash]['id']
               text = " "
               self._stg.insert_website_translation(source_text_id, tlang, text)


    def create_input_document(self, lang, txt, fp):
        text_splitter = TextSplitter(lang)
        sents = []
        indexes = []
        l_txt = txt.split('\n') 
        for line in l_txt:
            spl_line = text_splitter.split_by_sentences(line)
            sents += spl_line
            indexes.append(len(spl_line))
        self._stg.write_input_text(sents, fp)
        return indexes
 
    def prepare_translation(self, up, slang, tlang, system_id, tmp_dir):
        pair = "%s-%s" % (slang, tlang)
        web_lang = up.control['website']['language']
        l_in_txt = []
        l_out_txt = []
        res = self._stg.create_tmp_tl_pair_dir(up.id, slang, tlang)
        if res['code'] != 0: 
            return res
        for thash in up.control['texts']:
            # Check existing supervised translations
            if tlang in up.control['texts'][thash]['tl']:
                status = up.control['texts'][thash]['tl'][tlang]['st']
                if status == Storage.WEB_TL_STATUS_MAN and not(up.manifest.get('force', False)):
                    continue # Do not retranslate if translation is manual and 'force' is disabled.

            res = self._stg.create_tmp_tl_pair_object_dir(up.id, slang, tlang, thash)
            if res['code'] != 0: 
                return res

            if slang == web_lang:
                txt = up.control['texts'][thash]['txt']
                in_txt_fp = self._stg.tmp_tl_pair_object_file_abspath(up.id, slang, tlang, thash, "in.txt")
                indexes = self.create_input_document(slang, txt, in_txt_fp)
            else:
                # Source from existing translations or generated translations (wait)
                if slang in up.control['texts'][thash]['tl']: # Existing translation
                    txt = up.control['texts'][thash]['tl'][tlang]['txt']
                    in_txt_fp = self._stg.tmp_tl_pair_object_file_abspath(up.id, slang, tlang, thash, "in.txt")
                    indexes = self.create_input_document(slang, txt, in_txt_fp)
                elif slang in up.control['texts'][thash]['outputs']: # Translation to be generated by IS
                    in_txt_fp = up.control['texts'][thash]['outputs'][slang]['out_fp']
                    indexes = up.control['texts'][thash]['outputs'][slang]['indexes']

            out_txt_fp = self._stg.tmp_tl_pair_object_file_abspath(up.id, slang, tlang, thash, "out.txt")
            up.control['texts'][thash]['outputs'][tlang] = {'src_lang':slang, 'in_fp':in_txt_fp, 'indexes':indexes, 'out_fp':out_txt_fp}
            l_in_txt.append(in_txt_fp)
            l_out_txt.append(out_txt_fp)
        # Write file lists
        l_in_txt_fp = "%s/in_txt.lst" % tmp_dir
        self._stg.write_list(l_in_txt, l_in_txt_fp)
        l_out_txt_fp = "%s/out_txt.lst" % tmp_dir
        self._stg.write_list(l_out_txt, l_out_txt_fp)
        return {'code':0, 'data':(len(l_in_txt), l_in_txt_fp, l_out_txt_fp)}
 
    def launch_translation(self, up, system_id, slang, tlang, wait_jobs=[]):
        if up.manifest.get("test_mode", False):
            if self._verbose: self._log("Test mode: will generate an empty %s translation ..." % (tlang))
            try:
                self.generate_empty_translations(up, tlang)
            except:
                return {'code':2291, 'info':'Could not generate empty translations'}
            return {'code':0, 'jobs':[]}
        else:
            jids = []
            pair = "%s-%s" % (slang, tlang)
            if self._verbose: self._log("Launching '%s' translation system (System ID %d) ..." % (pair, system_id))
            tmp_dir = self._stg.tmp_tl_pair_dir(up.id, slang, tlang)
            res = self.prepare_translation(up, slang, tlang, system_id, tmp_dir)
            if res['code'] != 0:
                return res
            ndocs, l_in_txt_fp, l_out_txt_fp = res['data']
            if ndocs == 0:
                if self._verbose: self._log("\t- No documents to be translated.")
                return {'code':0, 'jobs':[]}
            job_prefix = "%s.%s.%s.%s-%s" % (self._cfg['SCH']['JOB_NAME_PREFIX'], self._cfg['STO']['DB_NAME'], up.id, slang, tlang)
            all_jids = []
            # Translation
            mt = self._sys['MT'][pair]['dsys'][system_id]['module'] # MT module
            my_tmp_dir = "%s/translate" % tmp_dir
            my_job_name = "%s.translate" % job_prefix
            log = "%s/translate.log" % tmp_dir
            res = mt.translate(self._jbs, l_in_txt_fp, l_out_txt_fp, my_tmp_dir, html=False, log_file=log, job_name=my_job_name, hold_job_ids=jids)
            if res['code'] != 0:
                return res
            jids = res['jids']
            all_jids += jids
            if self._debug:
                self._log("Translation process launched:")
                self._log("\tCommand: %s" % res['cmd'])
                self._log("\tJob ID(s): %s" % str(jids))
                self._log("\tLog file: %s" % log)
            up.control['jobs']['all'] += all_jids
            return {'code':0, 'jobs':all_jids}
 
    
    def launch_translations(self, up):
        jobs = []
        pair_jobids = {}
        target_jobids = {}
        sys = up.control['systems']
        for pair in self.mtsort(sys['mt'].keys()):
            sl = pair.split('-')[0]
            tl = pair.split('-')[1]
            tlpath = sys['mt'][pair]
            if len(tlpath) == 1:
                if pair in pair_jobids:
                    if self._verbose: self._log("%s translation was explicitly requested but '%s' translation system was already launched." % (tl, pair))
                else:
                    system_id = sys["mt"][pair][0]['id']
                    wait_jobs = target_jobids.get(sl, [])
                    res = self.launch_translation(up, system_id, sl, tl, wait_jobs=wait_jobs)
                    if res['code'] != 0:
                        return res
                    jobs = jobs + res['jobs']
                    pair_jobids[pair] = res['jobs']
                    target_jobids[tl] = res['jobs']
                    up.control['jobs']['lang'][tl] = {'ids':res['jobs'], 'st':Scheduler.ST_QUEUED}
            else:
                if self._verbose: self._log("Will Launch several MT systems to generate a '%s' translation..." % (pair))
                wait_jobs = target_jobids.get(sl, [])
                for step in tlpath:
                    spair = step['p']
                    ssl = spair.split('-')[0]
                    stl = spair.split('-')[1]
                    if step.get('skip', False):
                        if self._verbose: self._log(" - Skipping translation step '%s' ('skip' option provided)." % (spair))
                    elif spair in pair_jobids:
                        if self._verbose: self._log(" - '%s' translation system was already launched." % (stl, spair))
                        wait_jobs = pair_jobids[spair]
                    else:
                        system_id = step['id']
                        if self._verbose: self._log(" - Launching '%s' translation system (System ID %d) ..." % (spair, system_id))
                        res = self.launch_translation(up, system_id, ssl, stl, wait_jobs=wait_jobs)
                        if res['code'] != 0:
                            return res
                        jobs = jobs + res['jobs']
                        wait_jobs = res['jobs']         
                        pair_jobids[spair] = res['jobs']
                        target_jobids[stl] = res['jobs']
                        up.control['jobs']['lang'][stl] = {'ids':res['jobs'], 'st':Scheduler.ST_QUEUED}
        return {'code':0, 'jobs':jobs}
    

    def store_translations(self, up, tlang):
        for thash in up.control['texts']:
            if tlang in up.control['texts'][thash]['outputs']:
                source_text_id = up.control['texts'][thash]['id']
                indexes = up.control['texts'][thash]['outputs'][tlang]['indexes']
                out_fp =  up.control['texts'][thash]['outputs'][tlang]['out_fp']
                ret = self._stg.read_output_text(out_fp)
                beg = 0; sents = []
                for i in xrange(len(indexes)):
                    end = beg + indexes[i]
                    mysents = ret[beg:end]
                    mytext = " ".join(mysents)
                    sents.append(mytext)
                    beg = end
                text = "\n".join(sents)
                self._stg.insert_website_translation(source_text_id, tlang, text)


    def check_user_quota(self, up):
        user_id = up.control['user']['id']
        words = up.control['quota']['w']
        res = self._stg.get_user_quota_documents(user_id)
        if res['code'] != 0:
            return res
        if res['data']['q_docs'] != None: # if quota == None (NULL) -> Unlimited quota
            #if res['data']['q_docs'] - docs < 0:
            #    return {'code':2230, 'info':'Unsufficient quota (remaining %d document files, but uploaded %d document files).' % (res['data']['q_docs'], docs)}
            #elif res['data']['q_sents'] - sents < 0:
            #    return {'code':2231, 'info':'Unsufficient quota (remaining %d sentences, but uploaded %d sentences).' % (res['data']['q_sents'], sents)}
            if res['data']['q_words'] - words < 0:
                return {'code':2232, 'info':'Unsufficient quota (remaining %d words, but uploaded %d words).' % (res['data']['q_words'], words)}
        res = self._stg.update_user_quota_documents(user_id, 0, 0, words)
        if res['code'] != 0:
            return res
        return {'code':0}
    
    ##---------- Upload processing functions -------------#
    
    
    def process_failed_uploads(self):
        for up in self._uploads['failed']:
            try:
                td = (datetime.now() - up.last_ts).days
                if td >= self._cfg['GEN']['RM_ERR_UP_DAYS']:
                    if self._verbose: self._log("[*] Deleting upload ID %s ..." % up.id)
                    up_dir = "%s/%s" % (self._stg.uploads_mp, up.id)
                    res = self._stg.delete_upload(up.id, up_dir)
                    if res['code'] != 0:
                        self._warn(res['info'])
                    if self._verbose: self._log("Upload ID %s succesfully deleted." % up.id)
            except:
                self._bug("Upload ID: %s" % up.id)
                continue
    
    
    def process_finished_uploads(self):
        for up in self._uploads['finished']:
            try:
                td = (datetime.now() - up.last_ts).days
                if td >= self._cfg['GEN']['RM_FIN_UP_DAYS']:
                    if self._verbose: self._log("[*] Deleting upload ID %s ..." % up.id)
                    up_dir = "%s/%s" % (self._stg.uploads_mp, up.id)
                    res = self._stg.delete_upload(up.id, up_dir)
                    if res['code'] != 0:
                        self._error(res['info'], exit_code=res['code'])
                    if self._verbose: self._log("Upload ID %s succesfully deleted." % up.id)
            except:
                self._bug("Upload ID: %s" % up.id)
                continue
    
   
    def cancel_upload(self, up):
        res = jbs.delete_jobs(up.control['jobs'])
        if res['code'] != 0:
            warn(res['info'])
            print_excpt()
        return {'code':0} 

    def process_running_uploads(self):
        # Get Running jobs
        res = self._jbs.get_running_jobs()
        if res['code'] != 0:
            self._error(res['info'], exit_code=res['code'])
        rjobs = res['data']
        # For each upload
        for up in self._uploads['running']:
            try:
                if self._verbose: self._log("[*] Checking running upload ID %s ..." % up.id)
                if up.cancel:
                    if self._verbose: self._log("Oh nooo! Cancel request!")
                    res = self.cancel_upload(up)
                    if res['code'] != 0:
                        up.info = "Could not cancel upload."
                        up.status = WebUpload.STCODE_CANCEL_ERROR
                        up.error_code = res['code']
                        self.ingest_error(up)
                        continue
                    up.info = "Upload successfully cancelled."
                    up.status = WebUpload.STCODE_CANCELLED
                    self.ingest_update(up)
                    continue
               
                queued = False; running = False; error = False; finished = False;
                for tlang in up.control['jobs']['lang']:
                    t_jobs = up.control['jobs']['lang'][tlang]['ids']
                    t_prev_status = up.control['jobs']['lang'][tlang]['st']
                    if t_prev_status in (Scheduler.ST_QUEUED, Scheduler.ST_RUNNING):
                        res = self.check_running_processes(rjobs, t_jobs)
                        if res['code'] != 0:
                            self._error(res['info'], exit_code=res['code'])
                        job_status = res['data']
                        up.control['jobs']['lang'][tlang]['st'] = job_status
                        if job_status == Scheduler.ST_QUEUED:
                            queued = True
                        elif job_status == Scheduler.ST_RUNNING:
                            running = True
                        elif job_status == Scheduler.ST_ERROR:
                            error = True
                        elif job_status == Scheduler.ST_FINISHED:
                            finished = True
                            if up.status == WebUpload.STCODE_TL_PROC:
                                if self._verbose: self._log("'%s' translation processes finished." % (tlang))
                                try:
                                     self.store_translations(up, tlang)
                                except:
                                     self._bug("Cannot store translations")
                                     up.control['jobs']['lang'][tlang]['st'] = Scheduler.ST_ERROR
                                     error = True
                            else:
                                self._bug("Unexpected upload status %d" % (job_status), exit=True)
                    elif t_prev_status == Scheduler.ST_ERROR:
                        error = True
                    elif t_prev_status == Scheduler.ST_FINISHED:  
                        finished = True

                if queued:
                    if not(running):
                        up.info = "Processes queued."
                    else:
                        up.info = "Processes running."
                    self.ingest_update(up)
                elif running:
                    up.info = "Processes running."
                    self.ingest_update(up)
                elif error:
                    up.info = "Translation Error."
                    up.status = WebUpload.STCODE_TL_ERROR 
                    up.error_code = 2001
                    self.ingest_error(up)
                elif finished:
                    up.info = "Processes finished."
                    up.status = WebUpload.STCODE_NEW_FINISHED
                    self.ingest_update(up)
                else:
                    self._bug("Unknow global status of Upload ID: %s" % up.id)
                    continue
            except:
                self._bug("Upload ID: %s" % up.id)
                continue


    def process_new_uploads(self):
        for up in self._uploads['new']:
            try:
                if self._verbose: 
                    self._log("[*] Processing new upload ID %s ..." % up.id)
                if self._debug:
                    self._log("Manifest JSON:")
                    sys.stderr.write("\n%s\n" % json.dumps(up.manifest, indent=4, encoding='utf-8'))
                # Get user info
                res = self.get_user_info(up.user_id)
                if res['code'] != 0:
                    self._error(res['info'], exit_code=res['code'])
                up.control['user'] = res['data']
                if up.cancel:
                    if self._verbose: self._log("Oh nooo! Cancel request!")
                    up.info = "Upload successfully cancelled."
                    up.status = WebUpload.STCODE_CANCELLED
                    self.ingest_update(up)
                    continue
                # Check Website
                res = self.check_website(up)
                if res['code'] != 0:
                    up.info = res['info']
                    up.status = WebUpload.STCODE_MP_ERROR 
                    up.error_code = res['code']
                    self.ingest_error(up)
                    continue
                # Check texts
                ret = self.check_texts(up)
                if ret['code'] != 0:
                    up.info = ret['info']
                    up.status = WebUpload.STCODE_MP_ERROR 
                    up.error_code = ret['code']
                    self.ingest_error(up)
                    continue

                # Check user quota
                ret = self.check_user_quota(up)
                if ret['code'] != 0:
                    up.info = ret['info']
                    up.status = WebUpload.STCODE_MP_ERROR 
                    up.error_code = ret['code']
                    self.ingest_error(up)
                    continue

                # Determine translation tasks
                ret = self.set_tl_tasks(up)
                if ret['code'] != 0:
                    up.info = ret['info']
                    up.status = WebUpload.STCODE_MP_ERROR 
                    up.error_code = ret['code']
                    self.ingest_error(up)
                    continue

                # Prepare and launch translation tasks
                ret = self.launch_translations(up)
                if ret['code'] != 0:
                    up.info = ret['info']
                    up.status = WebUpload.STCODE_MP_ERROR 
                    up.error_code = ret['code']
                    self.ingest_error(up)
                    continue

                jids = ret['jobs']
                if len(jids) > 0:
                    up.info = "Translation(s) in progress. It may take several hours until it finishes."
                    up.status = WebUpload.STCODE_TL_PROC 
                    self.ingest_update(up)
                else:
                    if self._verbose: self._log("No translation processes were launched.")
                    up.info = "New translations successfully processed (No translation processes were launched)."
                    up.status = WebUpload.STCODE_NEW_FINISHED 
                    self.ingest_update(up)

                    # Append WebUpload object to "running", 
                    # to avoid processing the same Doc IDs in different upload IDs (see check_documents())
                    self._uploads['running'].append(up)
                    if self._debug:
                        self._log("Control JSON:")
                        sys.stderr.write("\n%s\n" % json.dumps(up.control, indent=4, encoding='utf-8'))

            except:
                self._bug("Upload ID: %s" % up.id)
                continue
    


if __name__ == "__main__":

    DEFAULT_CFG_FP = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))
    DEFAULT_CFG_FN = "config.ini"
    SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

    # Define Option Parser
    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option("-v", "--verbose", dest="verb", action="store_true", default=False, help="Verbose power on!")
    parser.add_option("-d", "--debug", dest="debug", action="store_true", default=False, help="Debug mode")
    parser.add_option("-e", "--disable-errors", dest="errors", action="store_false", default=True, help="Disable ingest errors")
    parser.add_option("-C", "--config-file", dest="config_file", type="string", default=DEFAULT_CFG_FP, help="Configuration file. Default: %s" % DEFAULT_CFG_FN)
    parser.add_option("-D", "--database", dest="db_name", type="string", default=None, help="Database with which to work. Default: specified in config file")
    (opts, args) = parser.parse_args()

    core = ISWebCore(opts.config_file, verbose=opts.verb, debug=opts.debug, commit_errors=opts.errors, script_dir=SCRIPT_DIR)

    try:
        # Remove finished uploads
        if opts.verb: core._log("----------- Checking failed uploads ------------")
        core.process_failed_uploads()
    
        # Remove finished uploads
        if opts.verb: core._log("---------- Checking finished uploads -----------")
        core.process_finished_uploads()
    
        # Check running uploads
        if opts.verb: core._log("----------- Checking running uploads -----------")
        core.process_running_uploads()
    
        # Check new uploads
        if opts.verb: core._log("------------ Checking new uploads --------------")
        core.process_new_uploads()
    
        if opts.verb: core._log("------------------- EXIT 0 ---------------------")
    except: 
        out = StringIO()
        traceback.print_exc(file=out)
        exception_str = out.getvalue().strip()
        print exception_str
        if core._mailer != None:
            core._mailer.send_mail_ingest_bug(exception_str)

