#!/usr/bin/env python
# -*- coding: utf-8 -*-

# libtldoc: Library to parse and manipulate EMMA TEI format
# Copyright 2014 MLLP
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from xml.dom import minidom
from datetime import datetime
from string import join, split
import cStringIO
import re
import sys
import copy

# Class Modelling #

class Document:

    STATUS_NOT_TRANSLATED = "not_translated"
    STATUS_AUTOMATIC = "fully_automatic"
    STATUS_MIXED = "partially_human"
    STATUS_MANUAL = "fully_human"
    ALLOWED_STATUS_VALUES = [ STATUS_NOT_TRANSLATED, STATUS_AUTOMATIC, STATUS_MIXED, STATUS_MANUAL ]

    AUTHOR_TYPE_AUTO = "automatic"
    AUTHOR_TYPE_MANUAL = "human"
    ALLOWED_AUTHOR_TYPE_VALUES = [ AUTHOR_TYPE_AUTO, AUTHOR_TYPE_MANUAL ]

    FILTER_NO_FILTER = 0
    FILTER_GET_FORMER = 1
    FILTER_GET_LAST = 2
    FILTER_GET_BEST = 3
    ALLOWED_FILTER_CODES = [ FILTER_NO_FILTER, FILTER_GET_FORMER, FILTER_GET_LAST, FILTER_GET_BEST ]

    def __init__(self):
        self.docId = None
        self.srcLang = None
        self.trgLang = None
        self.status = None
        self.sup_ratio = 0.0
        self.title = ""
        self.sentences = []

    def __repr__(self):
        return self.toXML(False)

    def toXML(self, filtered=False):
        out = cStringIO.StringIO()
        wn(out, '<?xml version="1.0" encoding="utf-8"?>')
        wn(out, '<tt xmlns:tl="translectures.eu">') 
        wn(out, '<head>')
        wn(out, '<tl:d dI="%s" sL="%s" tL="%s" st="%s" sR="%.2f" t="%s"/>' % (self.docId, self.srcLang, self.trgLang, self.status, self.sup_ratio, self.title))
        wn(out, '</head>')
        wn(out, '<body>')
        for s in self.sentences:
            s.printXML(out, filtered)
        wn(out, '</body>')
        wn(out, '</tt>')
        data = out.getvalue()
        out.close()
        return data

    def update(self, Dmod, authorId, authorConf, authorType, confMeasure, timeStamp, highlight=False):
        for s in self.sentences:
            D = Dmod.get(s.sentId, None)
            if D != None:
                # Update source, if needed
                srcD = D.get('o', None)
                if srcD != None:
                  rawText = srcD.get("r", None)
                  if rawText != None: s.origin.rawText = rawText
                  formattedText = srcD.get("f", None)
                  if formattedText != None: s.origin.formattedText = formattedText
                # Update translation
                tlD = D['t'] 
                t_aId = None
                for t in s.translations:
                    if t.authorId == authorId:
                        t_aId = t
                        break
                if t_aId != None:
                    rawText = tlD.get("r", None)
                    if rawText != None: t_aId.rawText = rawText
                    formattedText = tlD.get("f", None)
                    if formattedText != None: t_aId.formattedText = formattedText
                    t_aId.authorConf = authorConf
                    t_aId.confMeasure = confMeasure
                    t_aId.timeStamp = timeStamp
                else:
                    new_t = Translation()
                    rawText = tlD.get("r", None)
                    if rawText != None: new_t.rawText = rawText
                    formattedText = tlD.get("f", None)
                    if formattedText != None: new_t.formattedText = formattedText
                    new_t.authorId = authorId
                    new_t.authorType = authorType
                    new_t.authorConf = authorConf
                    new_t.confMeasure = confMeasure
                    new_t.timeStamp = timeStamp
                    s.translations.append(new_t)
                if highlight:
                    s.highlight = True

    def updateStatus(self):
        self.status = self.STATUS_NOT_TRANSLATED
        self.filter(self.FILTER_GET_BEST)
        fAuto = False
        fHuman = False
        fEmpty = False
        nh = 0
        for s in self.sentences:
            if s.filteredTranslation != None:
                if s.filteredTranslation.authorType == self.AUTHOR_TYPE_AUTO:
                    fAuto = True
                elif s.filteredTranslation.authorType == self.AUTHOR_TYPE_MANUAL: 
                    fHuman = True
                    nh += 1
            else:
                fEmpty = True
                break
        if fEmpty: 
            self.status = self.STATUS_NOT_TRANSLATED
            self.sup_ratio = 0
        else:
            if fAuto and not(fHuman): self.status = self.STATUS_AUTOMATIC
            elif fAuto and fHuman: self.status = self.STATUS_MIXED
            elif not(fAuto) and fHuman: self.status = self.STATUS_MANUAL
            self.sup_ratio = nh / float(len(self.sentences)) * 100
                    
    def filter(self, filtPol):
        if filtPol not in self.ALLOWED_FILTER_CODES:
             raise Exception, 'Unknown filter code %d' % filtPol
        for s in self.sentences:
            s.filter(filtPol)
        
    def swap(self):
        if self.status == Document.STATUS_NOT_TRANSLATED:
            raise Exception, 'No translations available for this DTLX file.'
        self.filter(self.FILTER_GET_BEST)
        for s in self.sentences:
            t = Translation()
            t.authorType = s.filteredTranslation.authorType
            t.authorId = s.filteredTranslation.authorId
            t.authorConf = s.filteredTranslation.authorConf
            t.confMeasure = s.filteredTranslation.confMeasure
            t.rawText = s.origin.rawText
            s.origin.rawText = s.filteredTranslation.rawText
            t.formattedText = s.origin.formattedText
            s.origin.formattedText = s.filteredTranslation.formattedText
            s.translations = [t]
            s.filteredTranslation = t
        src = self.srcLang
        trg = self.trgLang
        self.srcLang = trg
        self.trgLang = src
        self.updateStatus()
         
    def translations_to_source(self):
        if self.status == Document.STATUS_NOT_TRANSLATED:
            raise Exception, 'No translations available for this DTLX file.'
        self.filter(self.FILTER_GET_BEST)
        for s in self.sentences:
            s.origin.rawText = s.filteredTranslation.rawText
            s.origin.formattedText = s.filteredTranslation.formattedText
            s.translations = []
            s.filteredTranslation = None
        src = self.srcLang
        trg = self.trgLang
        self.srcLang = trg
        self.trgLang = src
        self.status = Document.STATUS_NOT_TRANSLATED
           
    def copy_translations_given_origin(self, dtlx, allowed_author_types=[AUTHOR_TYPE_AUTO, AUTHOR_TYPE_MANUAL]):
        dtlx.filter(self.FILTER_GET_BEST)
        tgi = [ i for i in xrange(len(dtlx.sentences)) ]
        for s in self.sentences:
            i = 0
            found = False
            while i < len(tgi) and not(found):
                cs = dtlx.sentences[tgi[i]]
                if cs.origin.formattedText.strip() == s.origin.formattedText.strip() and cs.filteredTranslation.authorType in allowed_author_types: 
                    s.translations.append(cs.filteredTranslation)
                    tgi.pop(i)
                    found = True
                i += 1
        self.updateStatus()

    def copy_translations_given_sentId(self, dtlx, allowed_status=[AUTHOR_TYPE_AUTO, AUTHOR_TYPE_MANUAL]):
        dtlx.filter(self.FILTER_GET_BEST)
        for cs in dtlx.sentences:
            i = 0
            found = False
            while i < len(self.sentences) and not(found):
                s = self.sentences[i]
                if s.sentId == cs.sentId:
                    s.translations.append(cs.filteredTranslation)
                    found = True
                i += 1
            if not(found):
                raise Exception, 'Sentence id "%s" does not exist in this Document.' % cs.sentId
        self.updateStatus()
                   
    def empty_translations(self):
        for s in self.sentences:
            t = Translation()
            t.authorType = self.AUTHOR_TYPE_AUTO
            t.authorId = "empty_translation"
            t.authorConf = 0
            t.confMeasure = 0
            t.rawText = ""
            t.formattedText = ""
            s.translations = [t]
            s.filteredTranslation = t
        self.updateStatus()


class Sentence:

    def __init__(self, sentId):
        self.sentId = sentId
        self.origin = None
        self.translations = []
        self.filteredTranslation = None
        self.highlight = False 

    def __repr__(self):
        out = cStringIO.StringIO()
        self.printXML(out, False)
        data = out.getvalue()
        out.close()
        return data


    def filter(self, filtPol):
        self.filteredTranslation = None
        if filtPol == Document.FILTER_NO_FILTER: # No filter
            pass 
        if filtPol == Document.FILTER_GET_FORMER: # Former alternatives
            lts = datetime(2500,01,01,00,00)
            for t in self.translations:
                if t.timeStamp < lts:
                    self.filteredTranslation = t
                    lts = t.timeStamp
        elif filtPol == Document.FILTER_GET_LAST: # Last alternatives
            lts = datetime(1970,01,01,00,00)
            for t in self.translations:
                if t.timeStamp > lts:
                    self.filteredTranslation = t
                    lts = t.timeStamp
        elif filtPol == Document.FILTER_GET_BEST: # Best and last alternatives
            lts = datetime(1970,01,01,00,00)
            bcM = -1
            baC = -1
            fhuman = False
            for t in self.translations:
                if t.authorType == Document.AUTHOR_TYPE_AUTO and not(fhuman):
                    if t.confMeasure > bcM or (t.confMeasure == bcM and t.timeStamp > lts):
                        self.filteredTranslation = t
                        bcM = t.confMeasure
                        lts = t.timeStamp
                elif t.authorType == Document.AUTHOR_TYPE_MANUAL:
                    fhuman = True
                    if t.authorConf > baC or (t.authorConf == baC and t.timeStamp > lts):
                        self.filteredTranslation = t
                        baC = t.authorConf
                        lts = t.timeStamp


    def printXML(self, out, filtered):
        w(out, '<tl:s') 
        w(out, ' sI="%s"' % self.sentId)
        if self.highlight:
            w(out,' h="%d"' % (int(self.highlight)))
        wn(out,'>')
        self.origin.printXML(out)
        if filtered:
            self.filteredTranslation.printXML(out)
        else:
            for t in self.translations:
                t.printXML(out)
        wn(out, '</tl:s>')


class Origin:

    def __init__(self):
        self.rawText = None
        self.formattedText = None

    def __repr__(self):
        out = cStringIO.StringIO()
        self.printXML(out)
        data = out.getvalue()
        out.close()
        return data

    def printXML(self, out):
        wn(out, '<tl:o>')
        if self.formattedText != None:
            wn(out, '<tl:f>')
            wn(out, '<![CDATA[%s]]>' % re.sub(r"]]>", r"]]]]><![CDATA[>", self.formattedText))
            wn(out, '</tl:f>')
        wn(out, '<tl:r>')
        wn(out, escape(self.rawText))
        wn(out, '</tl:r>')
        wn(out, '</tl:o>')


class Translation:

    def __init__(self):
        self.authorType = None
        self.authorId = None
        self.authorConf = 0.1
        self.confMeasure = 0.1
        self.timeStamp = datetime.now()
        self.rawText = None
        self.formattedText = None
        self.wordAligns = None

    def __repr__(self):
        out = cStringIO.StringIO()
        self.printXML(out)
        data = out.getvalue()
        out.close()
        return data

    def __timestamp2string(self):
        tt = self.timeStamp.timetuple()
        return '%04d-%02d-%02dT%02d:%02d:%02d'%(tt[0],tt[1],tt[2],tt[3],tt[4],tt[5])

    def printXML(self, out):
        w(out, '<tl:t aT="%s" aI="%s" aC="%s" cM="%s" tS="%s"' % (self.authorType, self.authorId, self.authorConf, self.confMeasure, self.__timestamp2string()))
        if self.wordAligns != None:
            w(out, ' wA="%s"' % " ".join([ str(x) for x in self.wordAligns ]))
        wn(out, '>')
        if self.formattedText != None:
            wn(out, '<tl:f>')
            wn(out, '<![CDATA[%s]]>' % self.formattedText)
            wn(out, '</tl:f>')
        wn(out, '<tl:r>')
        wn(out, escape(self.rawText))
        wn(out, '</tl:r>')
        wn(out, '</tl:t>')


# Parser #


class FormatError(Exception):
    pass


class Parser:

    def __init__(self):
        pass

    def parse(self, fn):
        xml = minidom.parse(fn)
        doc = self.__parseXML(xml)
        return doc
    
    def parseString(self, s):
        xml = minidom.parseString(s)
        doc = self.__parseXML(xml)
        return doc

    def __parseXML(self, xml):
        doc = Document()
        tt = xml.getElementsByTagName("tt")[0]
        head = tt.getElementsByTagName("head")[0]
        self.__parseHead(head, doc)
        body = tt.getElementsByTagName("body")[0]
        self.__parseBody(body, doc)
        return doc

    def __parseHead(self, xml, doc):
        found = False
        for n in xml.childNodes:
            if isinstance(n, minidom.Element):
                if n.tagName == "tl:document" or n.tagName == "tl:d":
                    self.__parseDocument(n, doc)
                    found = True
                else:
                    raise FormatError, 'Expected <tl:document/> or <tl:d> at <head>, but found another unexpected element.'
        if not(found):
            raise FormatError, '<tl:document> or <tl:d> tag not found at <head>.'


    def __parseDocument(self, xml, doc):
        # documentID
        doc.docId = xml.getAttribute('dI') 
        if doc.docId == "": doc.docId = xml.getAttribute('documentId')
        if doc.docId == "": raise FormatError, 'Attribute "documentId" or "dI" not found at <tl:document> or <tl:d>.'
        # srcLang
        doc.srcLang = xml.getAttribute('sL') 
        if doc.srcLang == "": doc.srcLang = xml.getAttribute('srcLang')
        if doc.srcLang == "": raise FormatError, 'Attribute "srcLang" or "sL" not found at <tl:document> or <tl:d>.'
        # trgLang
        doc.trgLang = xml.getAttribute('tL') 
        if doc.trgLang == "": doc.trgLang = xml.getAttribute('trgLang')
        if doc.trgLang == "": raise FormatError, 'Attribute "trgLang" or "tL" not found at <tl:document> or <tl:d>.'
        # status
        doc.status = xml.getAttribute('st') 
        if doc.status == "": doc.status = xml.getAttribute('status')
        if doc.status == "": raise FormatError, 'Attribute "status" or "st" not found at <tl:document> or <tl:d>.'
        if doc.status not in Document.ALLOWED_STATUS_VALUES: raise FormatError, 'Unallowed value for attribute "status" at <tl:document> or <tl:d>.'
        # ttitle
        doc.title = xml.getAttribute('t') 
        if doc.title == "": doc.title = xml.getAttribute('title')

    def __parseBody(self, xml, doc):
        for n in xml.childNodes:
            if isinstance(n, minidom.Element):
                if n.tagName == "tl:sentence" or n.tagName == "tl:s":
                    self.__parseSentence(n, doc)
                else:
                    raise FormatError, 'Expected <tl:sentence> or <tl:s> element under <body>, but found another element.'
	
    def __parseSentence(self, xml, doc):
        # sentenceID
        sI = xml.getAttribute('sI')
        if sI == "": sI = xml.getAttribute('sentenceId')
        if sI == "": raise FormatError, 'Attribute "sentenceId" or "sI" not found at <tl:sentence> or <tl:s>.'
        s = Sentence(sI)
        for n in xml.childNodes:
            if isinstance(n, minidom.Element):
                if n.tagName == "tl:origin" or n.tagName == "tl:o":
                    self.__parseOrigin(n, s)
                elif n.tagName == "tl:translation" or n.tagName == "tl:t":
                    self.__parseTranslation(n, s)
                else:
                    raise FormatError, 'Expected <tl:origin> or <tl:translation> element under <tl:sentence>, but found another element.'
        if s.origin == None:
            raise FormatError, 'Missing <tl:origin> under <tl:sentence>.'
        doc.sentences.append(s)


    def __parseRawText(self, xml, obj):
        txt = ""
        for n in xml.childNodes:
            if isinstance(n, minidom.Text): 
                txt = n.data.strip()
                if txt == "": continue
                break
        obj.rawText = txt

	
    def __parseFormattedText(self, xml, obj):
        data = ""
        for n in xml.childNodes:
            if isinstance(n, minidom.CDATASection): 
                data = n.data.strip()
                if data == "": continue
                break
        obj.formattedText = data

	
    def __parseOrigin(self, xml, sent):
        o = Origin()
        for n in xml.childNodes:
            if isinstance(n, minidom.Element):
                 if n.tagName == "tl:rawtext" or n.tagName == "tl:r":
                     self.__parseRawText(n, o)
                 elif n.tagName == "tl:formattedtext" or n.tagName == "tl:f":
                     self.__parseFormattedText(n, o)
                 else:
                     raise FormatError, 'Expected <tl:r> or <tl:f> under <tl:origin>, but found something else.'
        if o.rawText == None:
            raise FormatError, 'Expected at least a <tl:r> under <tl:origin>, but not found.'
        sent.origin = o


    def __parseTranslation(self, xml, sent):
        # authorType
        aT = xml.getAttribute('aT') 
        if aT == "": aT = xml.getAttribute('authorType')
        if aT == "": raise FormatError, 'Attribute "authorType" or "aT" not found at <tl:translation> or <tl:t>.'
        if aT not in Document.ALLOWED_AUTHOR_TYPE_VALUES: raise FormatError, 'Unallowed value for attribute "authorType" at <tl:translation> or <tl:t>.'
        # authorId
        aI = xml.getAttribute('aI') 
        if aI == "": aI = xml.getAttribute('authorId')
        if aI == "": raise FormatError, 'Attribute "authorId" or "aI" not found at <tl:translation> or <tl:t>.'
        # authorConf
        aC = xml.getAttribute('aC') 
        if aC == "": aC = xml.getAttribute('authorConf')
        if aC == "": raise FormatError, 'Attribute "authorConf" or "aC" not found at <tl:translation> or <tl:t>.'
        # confMeasure
        cM = xml.getAttribute('cM') 
        if cM == "": cM = xml.getAttribute('confMeasure')
        if cM == "": raise FormatError, 'Attribute "confMeasure" or "cM" not found at <tl:translation> or <tl:t>.'
        # timeStamp
        tS = xml.getAttribute('tS') 
        if tS == "": tS = xml.getAttribute('timeStamp')
        if tS == "": raise FormatError, 'Attribute "timeStamp" or "tS" not found at <tl:translation> or <tl:t>.'
        try:
            tS = datetime.strptime(tS, "%Y-%m-%dT%H:%M:%S") 
        except:
            raise FormatError, 'Invalid format of attribute "timeStamp" or "tS" at <tl:translation> or <tl:t>.'
        # wordAligns
        wA = xml.getAttribute('wA')
        if wA == "": wA = xml.getAttribute('wordAligns')
        if wA == "": wA = None
        if wA != None:
            try:
                doc.wordAligns = [ int(x) for x in xml.getAttribute('wA').split(' ') ]
            except:
                raise FormatError, 'Bad format for attribute "wordAlings" at <tl:translation> or <tl:t>.'
        # Text
        t = Translation()
        for n in xml.childNodes:
            if isinstance(n, minidom.Element):
                 if n.tagName == "tl:rawtext" or n.tagName == "tl:r":
                     self.__parseRawText(n, t)
                 elif n.tagName == "tl:formattedtext" or n.tagName == "tl:f":
                     self.__parseFormattedText(n, t)
                 else:
                     raise FormatError, 'Expected <tl:r> or <tl:f> under <tl:t>, but found something else.'
        if t.rawText == None:
            raise FormatError, 'Expected at least a <tl:r> under <tl:origin>, but not found.'
        t.authorId = aI
        t.authorType = aT
        t.authorConf = aC
        t.confMeasure = cM
        t.timeStamp = tS
        sent.translations.append(t)


# Aux functions #


def escape(string):
    """ escapes xml-reserved tokens. """
    return re.sub(r"&(?!#\d{4};|amp;|lt;|gt;|quot;|apos;)", "&amp;", string).replace('<','&lt;').replace('>','&gt;').replace('"','&quot;').replace("'",'&apos;')

    
""" Fast string generation functions """
def w(sIO, string): sIO.write(string.encode('utf-8'))
def wn(sIO, string): w(sIO, "%s\n" % string)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print "Usage: %s <xml>" % sys.argv[0]
        sys.exit(1)
    parser = Parser()
    doc = parser.parse(sys.argv[1])
    #doc.updateStatus()
    #print doc
    #Dmod = { "titolo_0":{"r":"kaka de vaka"} }
    #doc.update(Dmod, "UPV", 1, "human", 1, datetime.now())
    #print doc
    for s in doc.sentences:
      print s.origin.rawText.encode('utf-8')


