# -*- coding: utf-8 -*-
#
#  Copyright 2013, 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
from datetime import datetime
from libdfxp import *
from trs import *

DIR = "%s" % os.path.dirname(os.path.abspath(__file__))
DFXP2TXT = "%s/dfxp2txt.sh" % DIR
TXT2DFXP = "%s/txt2dfxp.sh" % DIR

def __parse_srt_time(s):
    ls = s.split(',')
    ms = int(ls[1])
    lls = ls[0].split(':')
    h = int(lls[0])
    m = int(lls[1])
    s = int(lls[2])
    return (h*3600) + (m*60) + s + (ms/1000.0)


def __parse_srt_times(s):
    b = __parse_srt_time(s.split('-->')[0].strip())
    e = __parse_srt_time(s.split('-->')[1].strip())
    return b,e


def __srt2syncs(src_fd):
    syncs = []
    l = src_fd.readline()
    while l != "":
        if l.strip() == "": 
            l = src_fd.readline()
            continue
        l = src_fd.readline() # Skip caption number.
        begin, end = __parse_srt_times(l)
        l = src_fd.readline()
        ltext = []
        while l.strip() != "":
            ltext.append(l.strip())
            l = src_fd.readline()
        syncs.append((begin, end, " ".join(ltext).decode('UTF-8')))
    return syncs


def __trs2syncs(trs): 
    syncs = []
    for e in trs.episodes:
        for s in e.sections:
            for t in s.turns:
                text = None
                begin = t.start_time
                for c in t.content:
                    if isinstance(c, TransSync):
                        if text != None:
                            aux = begin, c.time, text
                            syncs.append(aux)
                        text = None
                        begin = c.time
                    elif isinstance(c, TransEvent): pass
                    elif isinstance(c, unicode):
                        if c == '[sonido de fondo]' or c == '[background noise]': text = ""
                        else: text = c
                    else: raise Exception()
                aux = begin, t.end_time, text
                syncs.append(aux)
    return syncs


def __syncs2dfxp(syncs):
    dfxp = TlDfxp()
    dfxp.timeStamp = datetime.now()
    dfxp.authorId = "tL-Ingest-service"
    dfxp.authorType = "human"
    dfxp.authorConf = 1.0
    dfxp.begin = syncs[0][0]
    dfxp.end = syncs[len(syncs)-1][1]
    dfxp.dfxp = dfxp
    i = 1
    for s in syncs:
        d = TlRawText(s[2])
        seg = TlSegment(dfxp, i, s[0], s[1])
        seg.data.append(d)
        dfxp.data.append(seg)
        i = i + 1
    return dfxp


def __store_dfxp(dfxp, trg_file):
    trg_fd = open(trg_file, 'w')
    trg_fd.write(dfxp.toXml())
    trg_fd.close()



def dfxp2txt(jbs, in_dfxp_lst, out_txt_lst, tmp_dir, log_file=None, job_name=None, hold_job_ids=None):
    cmd = "%s -v -t %s %s %s" % (DFXP2TXT, tmp_dir, in_dfxp_lst, out_txt_lst)
    if log_file == None:
        log_file = "%s/dfxp2txt.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=2.0, time="00:30:00", output=log_file, job_name=job_name, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid], 'cmd':cmd, 'log':log_file}

def txt2dfxp(jbs, trg_lang, dfxp2txt_tmp_dir, in_txt_lst, in_dfxp_lst, out_dfxp_lst, tmp_dir, log_file=None, job_name=None, hold_job_ids=None, system_label=None):
    my_tmp_dir = "%s/txt2dfxp" % tmp_dir
    opts = ""
    if system_label != None:
        opts += "-s %s" % system_label
    cmd = "%s -v -t %s -T %s %s %s %s %s %s" % (TXT2DFXP, tmp_dir, trg_lang, opts, dfxp2txt_tmp_dir, in_dfxp_lst, in_txt_lst, out_dfxp_lst)
    if log_file == None:
        log_file = "%s/txt2dfxp.log" % tmp_dir
    res = jbs.submit_job(cmd, mem=2.0, time="00:30:00", output=log_file, job_name=job_name, hold_jid=hold_job_ids)
    if res['code'] != 0:
        return res
    jid = res['jid']
    return {'code':0, 'jids':[jid], 'cmd':cmd, 'log':log_file}



def srt2dfxp(src_file, trg_file):
    try:
        src_fd = open(src_file)
    except:
        return {'code':40000, 'info':'Cannot open srt file.'}
    try:
        syncs = __srt2syncs(src_fd)
    except:
        return {'code':40001, 'info':'Cannot parse srt file (bad srt format?)'}
    src_fd.close()
    if len(syncs) == 0:
        return {'code':40002, 'info':'No syncs found in srt file.'}
    try:
        dfxp = __syncs2dfxp(syncs)
    except:
        return {'code':40003, 'info':'Cannot generate dfxp object from parsed srt object.'}
    try:
        __store_dfxp(dfxp, trg_file)
    except:
        return {'code':40004, 'info':'Cannot store dfxp data in disk.'}
    return {'code':0}


def trs2dfxp(src_file, trg_file):
    try:
        src_fd = open(src_file)
    except:
        return {'code':40005, 'info':'Cannot open trs file.'}
    try:
        trs = load_trans_file(src_fd)
        syncs = __trs2syncs(trs)
    except:
        return {'code':40006, 'info':'Cannot parse trs file (bad trs format?).'}
    if len(syncs) == 0:
        return {'code':3, 'info':'No syncs found in trs file.'}
    try:
        dfxp = __syncs2dfxp(syncs)
    except:
        return {'code':40007, 'info':'Cannot generate dfxp object from parsed trs object.'}
    try:
        __store_dfxp(dfxp, trg_file)
    except:
        return {'code':40008, 'info':'Cannot store dfxp data in disk.'}
    return {'code':0}

def create_empty_dfxp(media_length, dest_fp, segment_length=10):
    dfxp = TlDfxp()
    dfxp.timeStamp = datetime.now()
    dfxp.authorId = "tL-Ingest-service"
    dfxp.authorType = "automatic"
    dfxp.authorConf = 1.0
    dfxp.confMeasure = 1.0
    dfxp.begin = 0
    dfxp.end = media_length
    dfxp.dfxp = dfxp
    i = 1
    for s in range(0, media_length, segment_length):
        d = TlRawText("")
        seg = TlSegment(dfxp, i, s, min(float(s)+float(segment_length)-0.1, media_length))
        seg.data.append(d)
        dfxp.data.append(seg)
        i = i + 1
    __store_dfxp(dfxp, dest_fp)
    return {'code':0}


