#!/bin/bash -e 

set -u

export LC_ALL="C.UTF-8"
CURRDIR=$PWD; SCRIPTDIR=$(cd ${0%/*} &> /dev/null; echo $PWD; cd $CURRDIR &> /dev/null)

# log
function log {
  [ $VERBOSE -eq 0 ] || echo -e $1 >&2
}

function error {
  echo -e "[EE] $1" >&2
}

PRGNAME=$(basename $0)
NARGS=3

# Show help
function print_help {
    TO=$1
    SHOWHELP=$2
    echo "$PRGNAME [options] <SRC_LANG> <IN_DTLX_LIST> <OUT_TXT_LIST>" > $TO
    echo "" > $TO
    [ $SHOWHELP -eq 0 ] && return
    cat <<EOF > $TO

  Options:

    -h: Shows this help
    -v: Verbose power on!
    -r: Remove tmp directory at the end. 
    -f: File mode. All lists are considered as a single file.
    -t <str>: Specify temporal directory. By default it is created in /tmp.
    -S <str>: Scripts directory. Default: dirname of this script.
    -e <LIST>: Provide a list of DTLX files containing supervised translations to be preserved, one per line for each input DTLX file. 
              Use "_" (underscore) in the i-th line if no DTLX file is provided for the i-th input DTLX file.
    -T <str>: Set target language (by default is read from DTLX files)
    -o <LIST>: List of file paths to write target DTLX files. If not provided, target DTLX files are not stored.
    -s: Allow to use the best translations of the input DTLX as source text, if needed. 
 
EOF
}

SCRIPTS="$SCRIPTDIR"
VERBOSE=0
TMP="NULL"
RMTMP=0
LIST=1
EX_LIST="NULL"
TRG="NULL"
OUT_DTLX_LIST="NULL"
SWAP=0

ARGS=`getopt -o "hvrt:S:fe:T:o:s" -n $PRGNAME -- "$@"`
eval set -- "$ARGS"
while true
do
    case $1 in
	-h) print_help /dev/stdout 1; exit 0;;
        -v) VERBOSE=1;;
        -t) TMP="$2";;
        -r) RMTMP=1;;
        -f) LIST=0;;
        -S) SCRIPTS="$(readlink -f $2)";;
        -e) EX_LIST="$(readlink -f $2)";;
        -T) TRG="$2";;
        -o) OUT_DTLX_LIST="$2";;
        -s) SWAP=1;;
	--) shift; break;;
    esac
    shift
done
[ $# != $NARGS ] && { print_help /dev/stderr 0; error "wrong number of arguments"; exit 1; }

SRC="$1"
INLIST="$(readlink -f $2)"
OUTLIST="$(readlink -f $3)"

export PATH=$SCRIPTS:$PATH

set -e

if [ "$TMP" = "NULL" ]; then 
  TMP=${TMPDIR:-/tmp}/$(basename $0 .sh).$$; mkdir $TMP 2>/dev/null;
  #trap "rm -rf $TMP 2>/dev/null" EXIT
else
  mkdir -p "$TMP"
fi

if [ $LIST -eq 1 ]; then
  [ $(cat "$INLIST" | wc -l) -ne $(cat "$OUTLIST" | wc -l) ] && error "Input and Output lists have different number of lines." && exit 1;
  [ $(cat "$INLIST" | wc -l) -eq 0 ] && error "Empty input list." && exit 1;
  if [ $EX_LIST != "NULL" ]; then
    [ $(cat "$INLIST" | wc -l) -ne $(cat "$EX_LIST" | wc -l) ] && error "Existing (-e option) and Input lists have different number of lines." && exit 1;
  else
    EX_LIST="$TMP/existing.lst"
    awk '{print "_"}' $INLIST > $EX_LIST
  fi
  if [ $OUT_DTLX_LIST != "NULL" ]; then
    [ $(cat "$INLIST" | wc -l) -ne $(cat "$OUT_DTLX_LIST" | wc -l) ] && error "Out DTLX (-o option) and Input lists have different number of lines." && exit 1;
  else
    OUT_DTLX_LIST="$TMP/out_dtlx.lst"
    awk '{print "_"}' $INLIST > $OUT_DTLX_LIST
  fi

else

  echo "$INLIST" > "$TMP/in.lst"
  INLIST="$TMP/in.lst"
  echo "$OUTLIST" > "$TMP/out.lst"
  OUTLIST="$TMP/out.lst"
  EX_LIST="$TMP/existing.lst"
  if [ $EX_LIST != "NULL" ]; then
    echo "$EX_LIST" > "$TMP/existing.lst"
  else
    echo "_" > "$TMP/existing.lst"  
  fi
  OUT_DTLX_LIST="$TMP/out_dtlx.lst"
  if [ $OUT_DTLX_LIST != "NULL" ]; then
    echo "$OUT_DTLX_LIST" > "$TMP/out_dtlx.lst"
  else
    echo "_" > "$TMP/out_dtlx.lst"
  fi
fi

set -x

################## Preprocess ###################

mkdir -p "$TMP"
cd "$TMP"

gopts=""
[ "$TRG" != "NULL" ] && gopts="$gopts -t $TRG"
[ $SWAP -eq 1 ] && gopts="$gopts -s"

id=0
paste -d '|' $INLIST $OUTLIST $EX_LIST $OUT_DTLX_LIST | while IFS='|' read f trgf exf outdtlx; do
  id=$[id+1]
  [ ! -s "$f" ] && error "File $f does not exist or is empty." && exit 1
  opts=""
  if [ "$exf" != "_" ]; then
    [ ! -s "$exf" ] && error "File $exf does not exist or is empty." && exit 1
    opts="$opts -e $exf"
  fi
  if [ "$outdtlx" != "_" ]; then
    opts="$opts -o $outdtlx"
  fi
  "$SCRIPTS/dtlx2txt.py" $gopts $opts $SRC "$f" | awk '{if(NF==1) print $1,"__EMPTY__"; else print}' > "$TMP/$id.tlftxt"
  if [ $(cat "$TMP/$id.tlftxt" | wc -l) -eq 0 ]; then
    echo "[WW] No input sentences. Skipping." 1>&2
    #cp "$f" "$trgf"
    #continue
  fi
  awk '{print $1}' "$TMP/$id.tlftxt" > "$TMP/$id.segids"
  awk '{for(i=2;i<=NF;i++) printf("%s ",$i); printf("\n");}' "$TMP/$id.tlftxt" > "$TMP/$id.ftxt"
  cp "$TMP/$id.ftxt" $trgf
done

[ $RMTMP -eq 1 ] && rm -rf "$TMP"

set +x 

cd $CURRDIR

exit 0
