# -*- coding: utf-8 -*-
#
#  Copyright 2013, 2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import smtplib
from email.mime.text import MIMEText
from datetime import datetime
import traceback


class Mailer(object):
 
    SUBJECT_PREFIX = "[MLLP-TTP]"
    SERVICE_NAME = "MLLP's Media Transcription and Translation Platform"
    PLATFORM_URL = "https://ttp.mllp.upv.es"
    BOTTOM_MSG = """The MLLP-UPV team
http://mllp.upv.es\n
This message was generated automatically. Should you have any doubts, don't hesitate to contact us at mllp-support@upv.es ."""


    def __init__(self, src_mail, smtp_server, admin_processing_mail=None, admin_error_mail=None, admin_finished_mail=None):
        """ Constructor 

          src_mail     --  Source e-mail address.
          smtp_server  --  SMTP Server name.

        """
        self.src_mail = src_mail
        self.smtp_server = smtp_server 
        self.admin_processing_mail = admin_processing_mail
        self.admin_error_mail = admin_error_mail
        self.admin_finished_mail = admin_finished_mail

    def send_mail_ingest_bug(self, exception_str, extra_info=""):
        SUBJECT = "%s Ingest Service Exception" % (self.SUBJECT_PREFIX)
        BODY = """%s\n%s""" % (extra_info, exception_str)
        return self._send_mail(SUBJECT, BODY, self.admin_error_mail)

    def _send_mail(self, subject, content, to, cc=None, cco=None):
        """Send email to the specified recipients about the status of an specific upload ID.

        Arguments:

         Returns a Dict -> {'code':INT, 'info':STR}
 
         """
        # Create e-mail data
        try:
            msg = MIMEText(content.encode('UTF-8'))
            msg['Subject'] = subject.encode('UTF-8')
            msg['From'] = self.src_mail
            msg['To'] = ",".join(to)
            addrl = []
            addrl = addrl + to
            if cc != None: 
                msg['cc'] = ",".join(cc)
                addrl = addrl + cc
            if cco != None: 
                msg['cco'] = ",".join(cco)
                addrl = addrl + cco
        except:
            traceback.print_exc()
            return {'code':30000, 'info':'Cannot prepare mail data.'}
        # Send e-mail
        try:
            smtp = smtplib.SMTP(self.smtp_server)
            smtp.sendmail(self.src_mail, addrl, msg.as_string())
        except:
            traceback.print_exc()
            return {'code':30001, 'info':'Cannot send mail.'}
            
        # Everything OK :)
        return {'code':0, 'info':''}


    def _now(self):
        return self._datetime2str(datetime.now())
  
    def _datetime2str(self, date):
        return date.strftime("%d/%m/%Y %H:%M:%S %Z")

    def _seconds2readable_str(self, s):
        if s < 60:
            ss = "s" if s > 1 else ""
            return "%d second%s" % (s, ss)
        elif s >= 60 and s < 3600:
            minutes = s/60
            ms = "s" if minutes > 1 else ""
            return "%d minute%s" % (minutes, ms)
        elif s >= 3600 and s < 86400:
            hours = s/3600
            hs = "s" if hours > 1 else ""
            minutes = (s%3600)/60
            ms = "s" if minutes > 1 else ""
            return "%d hour%s, %d minute%s" % (hours, hs, minutes, ms)
        else:
            days = s/86400
            ds = "s" if days > 1 else ""
            hours = (s%86400)/3600
            hs = "s" if hours > 1 else ""
            minutes = ((s%86400)%3600)/60
            ms = "s" if minutes > 1 else ""
            return "%d day%s, %d hour%s %d minute%s" % (days, ds, hours, hs, minutes, ms)

class SpeechMailer(Mailer):

    def __init__(self, src_mail, smtp_server, admin_processing_mail=None, admin_error_mail=None, admin_finished_mail=None):
        Mailer.__init__(self, src_mail, smtp_server, admin_processing_mail=admin_processing_mail, admin_error_mail=admin_error_mail, admin_finished_mail=admin_finished_mail)  

    def send_mail_user_processing(self, up_id, v_id, title, to, cc=None, cco=None, eta=None):

        SUBJECT = "%s Your media file is being processed" % (self.SUBJECT_PREFIX)
        BODY =  """Dear user,

Your media file is being processed within the %s. \
Please be patient, it may take some time. We'll send you an email when the process is complete.

  Detailed Info:

  - Upload ID: %s
  - Video ID: %s
  - Title: %s
""" % (self.SERVICE_NAME, up_id, v_id, title)

        if eta != None:
            BODY += "  - Estimated time to complete: %s\n" % (self._seconds2readable_str(eta))
        BODY += "\n%s" % self.BOTTOM_MSG

        return self._send_mail(SUBJECT, BODY, to, cc, cco)

    def send_mail_user_finished(self, up_id, v_id, title, to, cc=None, cco=None):

        SUBJECT = "%s Your media file is now available"  % (self.SUBJECT_PREFIX)
        BODY =  """Dear user,

Your media file has been successfully processed by the %s. \
Your transcriptions and translations are now available. If you are using our trial service, \
you can log in to your personal account to view, edit and download your subtitles:

  %s

Detailed info:

  - Upload ID: %s
  - Video ID: %s
  - Title: %s

%s
""" % (self.SERVICE_NAME, self.PLATFORM_URL, up_id, v_id, title, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, to, cc, cco)

    def send_mail_user_error(self, up_id, v_id, title, err_code, info, to, cc=None, cco=None):
        
        SUBJECT = "%s Oops! We have encountered an error with your media file"  % (self.SUBJECT_PREFIX)
        BODY =  """Dear user,

An unexpected error has occurred while processing your media file. Our technical team is looking into it now. \
If we need more information from you, we'll be in touch. Otherwise, bear with us, we'll let you know when your file is available.

  Detailed info:

  - Upload ID: %s
  - Video ID: %s
  - Title: %s
  - Error code: %d
  - Info: %s
  
Thank you for your patience.

%s
""" % (up_id, v_id, title, err_code, info, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, to, cc, cco)

    def send_mail_admin_processing(self, up_id, v_id, title, language, dbname, user_mail, username, eta=None):    

        if self.admin_processing_mail == None: 
            return {'code':30002, 'info':'Cannot send mail: no admin email provided.'}
        
        mail_str="<unknown>"
        if user_mail != None: mail_str = ','.join(user_mail)

        SUBJECT = "%s A user has uploaded a new media file"  % (self.SUBJECT_PREFIX)
        BODY =  """An existing user has uploaded a new media file to the %s and it is being processed.

  Detailed info:

  - User: %s
  - Upload ID: %s
  - Video ID: %s
  - Title: %s
  - Media language: %s
  - Timestamp: %s
""" %(self.SERVICE_NAME, username, up_id, v_id, title, language, self._now())

        if eta != None:
            BODY += "  - Estimated time to complete: %s\n" % (self._seconds2readable_str(eta))
        BODY += "\n%s" % self.BOTTOM_MSG

        return self._send_mail(SUBJECT, BODY, self.admin_processing_mail)


    def send_mail_admin_finished(self, up_id, v_id, title, language, dbname, user_mail, username):    
        
        if self.admin_finished_mail == None: 
            return {'code':30003, 'info':'Cannot send mail: no admin email provided.'}

        mail_str="<unknown>"
        if user_mail != None: mail_str = ','.join(user_mail)

        SUBJECT = "%s A user's media file is now available"  % (self.SUBJECT_PREFIX)
        BODY =  """An uploaded media file has been successfully processed by the %s.

  Detailed info:

  - User: %s
  - Upload ID: %s
  - Video ID: %s
  - Title: %s
  - Media language: %s
  - Timestamp: %s

%s
""" %(self.SERVICE_NAME, username, up_id, v_id, title, language, self._now(), self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, self.admin_finished_mail)

    def send_mail_admin_error(self, up_id, v_id, title, language, dbname, user_mail, st_code, err_code, info, username):

        if self.admin_error_mail == None: 
            return {'code':30004, 'info':'Cannot send mail: no admin email provided.'}

        mail_str="<unknown>"
        if user_mail != None: mail_str = ','.join(user_mail)
        
        SUBJECT = "%s Error on Upload ID %s" % (self.SUBJECT_PREFIX, up_id)
        BODY =  """An error was found with an uploaded media package.

  Detailed info:

  - User: %s
  - Upload ID: %s
  - Video ID: %s
  - Title: %s
  - Media language: %s
  - Status code: %d
  - Error code: %d
  - Info: %s
  - Timestamp: %s

%s
"""  %(username, up_id, v_id, title, language, st_code, err_code, info, self._now(), self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, self.admin_error_mail)



class TextMailer(Mailer):

    def __init__(self, src_mail, smtp_server, admin_processing_mail=None, admin_error_mail=None, admin_finished_mail=None):
        Mailer.__init__(self, src_mail, smtp_server, admin_processing_mail=admin_processing_mail, admin_error_mail=admin_error_mail, admin_finished_mail=admin_finished_mail)  

    def send_mail_user_processing(self, up_id, up_ts, language, documents, sentences, words, to, cc=None, cco=None):

        SUBJECT = "%s Your text documents are being processed"  % (self.SUBJECT_PREFIX)
        BODY =  """Dear user,

Your text documents are being processed within the %s. \
Please be patient, it will take some time. We'll send you an email when the process is complete.

  Detailed Info:

  - Upload ID: %s
  - Uploaded on: %s
  - Language: %s
  - Documents: %s
  - Sentences: %s
  - Words: %s

%s
""" % (self.SERVICE_NAME, up_id, self._datetime2str(up_ts), language, documents, sentences, words, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, to, cc, cco)

    def send_mail_user_finished(self, up_id, up_ts, language, documents, sentences, words, to, cc=None, cco=None):

        SUBJECT = "%s Your text documents are now available"  % (self.SUBJECT_PREFIX)
        BODY =  """Dear user,

Your document files have been successfully processed by the %s. \
Your translations are now available. If you are using our trial service, \
you can log in to your personal account to view, edit and download your translations:

  %s

Detailed info:

  - Upload ID: %s
  - Uploaded on: %s
  - Language: %s
  - Documents: %s
  - Sentences: %s
  - Words: %s

%s
""" % (self.SERVICE_NAME, self.PLATFORM_URL, up_id, self._datetime2str(up_ts), language, documents, sentences, words, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, to, cc, cco)

    def send_mail_user_error(self, up_id, up_ts, err_code, info, to, cc=None, cco=None):
        
        SUBJECT = "%s Oops! We have encountered an error with your document files"  % (self.SUBJECT_PREFIX)
        BODY =  """Dear user,

An unexpected error has occurred while processing your text document files. Our technical team is looking into it now. \
If we need more information from you, we'll be in touch. Otherwise, bear with us, we'll let you know when your files are available.

  Detailed info:

  - Upload ID: %s
  - Uploaded on: %s
  - Error code: %d
  - Info: %s
  
Thank you for your patience.

%s
""" % (up_id, self._datetime2str(up_ts), err_code, info, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, to, cc, cco)

    def send_mail_admin_processing(self, up_id, up_ts, language, documents, sentences, words, username):    

        if self.admin_processing_mail == None: 
            return {'code':30002, 'info':'Cannot send mail: no admin email provided.'}
        
        SUBJECT = "%s A user has uploaded a new set of text documents"  % (self.SUBJECT_PREFIX)
        BODY =  """An existing user has uploaded a set of text document files to the %s and they are being processed.

  Detailed info:

  - User: %s
  - Upload ID: %s
  - Uploaded on: %s
  - Language: %s
  - Documents: %s
  - Sentences: %s
  - Words: %s

%s
""" %(self.SERVICE_NAME, username, up_id, self._datetime2str(up_ts), language, documents, sentences, words, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, self.admin_processing_mail)


    def send_mail_admin_finished(self, up_id, up_ts, language, documents, sentences, words, username):    
        
        if self.admin_finished_mail == None: 
            return {'code':30003, 'info':'Cannot send mail: no admin email provided.'}

        SUBJECT = "%s New text documents are now available"  % (self.SUBJECT_PREFIX)
        BODY =  """A set of text documents has been successfully processed by the %s.

  Detailed info:

  - User: %s
  - Upload ID: %s
  - Uploaded on: %s
  - Finished on: %s
  - Language: %s
  - Documents: %s
  - Sentences: %s
  - Words: %s

%s
""" %(self.SERVICE_NAME, username, up_id, self._datetime2str(up_ts), self._now(), language, documents, sentences, words, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, self.admin_finished_mail)

    def send_mail_admin_error(self, up_id, up_ts, err_code, info, username):

        if self.admin_error_mail == None: 
            return {'code':30004, 'info':'Cannot send mail: no admin email provided.'}

        SUBJECT = "%s Error on Upload ID %s" % (self.SUBJECT_PREFIX, up_id)
        BODY =  """An error was found with an uploaded document package.

  Detailed info:

  - User: %s
  - Upload ID: %s
  - Uploaded on: %s
  - Failed on: %s
  - Error code: %d
  - Info: %s

%s
"""  %(username, up_id, self._datetime2str(up_ts), self._now(), err_code, info, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, self.admin_error_mail)


class AlignMailer(Mailer):

    def __init__(self, src_mail, smtp_server, admin_error_mail=None):
        Mailer.__init__(self, src_mail, smtp_server, admin_error_mail=admin_error_mail)  

    def send_mail_admin_error(self, va_id, va_ts, err_code, info):

        if self.admin_error_mail == None: 
            return {'code':30004, 'info':'Cannot send mail: no admin email provided.'}

        SUBJECT = "%s Error on Align ID %s" % (self.SUBJECT_PREFIX, va_id)
        BODY =  """An error was found with an alignment request.

  Detailed info:

  - Alignment ID: %s
  - Requested on: %s
  - Failed on: %s
  - Error code: %d
  - Info: %s

%s
"""  %(va_id, self._datetime2str(va_ts), self._now(), err_code, info, self.BOTTOM_MSG)

        return self._send_mail(SUBJECT, BODY, self.admin_error_mail)


