# -*- coding: utf-8 -*-
#
#  Copyright 2013 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from subprocess import call, check_call, check_output, CalledProcessError, STDOUT
import os
import traceback
from StringIO import StringIO

def print_exception():
    out = StringIO()
    traceback.print_exc(file=out)
    print out.getvalue().strip()

class Scheduler(object):

    ST_RUNNING = "r"
    ST_QUEUED = "q"
    ST_ERROR = "e"
    ST_FINISHED = "f"


    def __init__(self, localhost, status_cmd, submit_scr, submit_opts=""):
        """ Constructor """
        self.localhost = localhost # Localhost
        self.status_cmd = status_cmd # Command to get queue/jobs status
        self.submit_scr = submit_scr # Script for submitting jobs to queue
        self.submit_opts = submit_opts # Default, static options for submit script



    #------ Private methods ------#


 
    def __parse_status_output(self, res):
        """Parses output of status script and returns
           a dict mapping job ID -> job status.
           job status codes must be:
           r -> running
           q -> queued
           e -> error

        """
        jlist = res.split('\n')[2:]
        d_job_st = {}
        for j in jlist:
            r = j.split()
            try:
                jid = int(r[0])
                job_status = r[4]
            except IndexError:
                pass
            else:
                if job_status == "r": job_status = self.ST_RUNNING
                elif job_status == "hr": job_status = self.ST_RUNNING
                elif job_status == "t": job_status = self.ST_QUEUED
                elif job_status == "qw": job_status = self.ST_QUEUED
                elif job_status == "hqw": job_status = self.ST_QUEUED
                elif job_status == "Eqw": job_status = self.ST_ERROR
                d_job_st[jid] = job_status
        return d_job_st


    def __get_job_id(self, res):
        """Parses output of submit script and returns
           the submitted Job ID.

        """
        for l in res.split('\n'):
            ls = l.split()
            try:
                if ls[0] == "Job:": return int(ls[1])
            except IndexError:
                pass
        return None



    #------ Public methods ------#



    def get_running_jobs(self):
        """ Returns the list of all running jobs in a Dict
            mapping job IDs with job status

        """
        # Execute status script
        try:
            res = check_output(self.status_cmd, shell=True)
        except CalledProcessError:
            print_exception()
            return {'code':20000, 'info':'Cannot retrieve list of running jobs.'}
        # Parse output
        try:
            d_job_st = self.__parse_status_output(res)
        except:
            print_exception()
            return {'code':20001, 'info':'Cannot parse list of running jobs.'}
        # Return data
        return {'code':0, 'info':'', 'data':d_job_st}
  

    def execute_job(self, command, stdout=None, stderr=None):
        try:
            if stdout != None and stderr != None and stdout == stderr:
                stdout_fd = open(stdout, "w")
                stderr_fd = stdout_fd
            else: 
                if stdout != None:
                    stdout_fd = open(stdout, "w")
                else:
                    stdout_fd = None
                if stderr != None:
                    stderr_fd = open(stderr, "w")
                else:
                    stderr_fd = None
        except IOError:
            return {'code':20002, 'info':'Cannot open sdtout and stderr files.'}
        try:
            check_call(command, shell=True,
                       stdout=stdout_fd, 
                       stderr=stderr_fd)
        except CalledProcessError:
            print_exception()
            return {'code':20003, 'info':'Cannot execute command.'}
        return {'code':0}


    def execute_job_output(self, command):
        try:
            res = check_output(command, shell=True)
        except CalledProcessError:
            print_exception()
            return {'code':20004, 'info':'Cannot execute command.'}
        return {'code':0, 'data':res}


    def submit_job(self, command, mem=None, time=None, output=None, job_name=None, hold_jid=[], local=False, other_opts=""):
        """ Submits a job to a queue 

        Arguments:

         command     --  command to be submitted to queue (str)
         mem         --  memory requirements in GB (float)
         time        --  time requirements in format HH:MM:SS (str)
         output      --  file in which store the output of the job (str)
         job_name    --  Job name (str)
         hold_jid    --  List of job IDs to wait (list of ints)
         local       --  Submit job only to local machine queue
         other_opts  --  Other options (string)

        Returns:

         Dict -> {'code':INT, 'info':STR, 'data':DICT }
         where 'data' contains the submitted Job ID.
       
        """
        # Prepare call to submit script
        cmd = "%s" % (self.submit_scr)
        # Specify memory requirements
        if mem != None: cmd = "%s -m %.1f" % (cmd, mem)
        # Specify time requirements
        if time != None: cmd = "%s -t %s" % (cmd, time)
        # Add hold for a list of Job IDs
        if hold_jid != None and hold_jid != []: 
            wopt = ",".join(map(lambda x: str(x), hold_jid))
            cmd = "%s -w %s" % (cmd, wopt)
        # Specify job name
        if job_name != None: cmd = "%s -n %s" % (cmd, job_name)
        # Set log filename
        if output != None: cmd = "%s -o %s" % (cmd, output)
        # Send to local machine
        if local: cmd = "%s -a %s" % (cmd, self.localhost)
        # Add other options
        cmd = "%s %s %s" % (cmd, other_opts, self.submit_opts)
        # Add the command to be submitted to the queue
        cmd = "%s '%s'" % (cmd, command)
        # Execute call to submit script
        try:
            res = check_output(cmd, stderr=STDOUT, shell=True)
        except CalledProcessError:
            print_exception()
            return {'code':20005, 'info':'Cannot submit job to queue.'}
        # Get submmited job ID from the output
        job_id = self.__get_job_id(res)
        if job_id == None:
            return {'code':20006, 'info':'Cannot get job ID.'}
        # Return job ID
        return {'code':0, 'info':'', 'jid':job_id}


    def delete_jobs(self, jids_list):
        cmd = "qdel %s" %(",".join(map(lambda x: str(x), jids_list)))
        try:
            check_output(cmd, shell=True)
        except:
            return {'code':20007, 'info':'Could not delete all job IDs (maybe they finished before).'}
        return {'code':0}

