# -*- coding: utf-8 -*-
#
#  Copyright 2013 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import shutil
import psycopg2
import psycopg2.extras
from psycopg2.extras import Json
import json
import re
import os
import base64
import hashlib
from datetime import datetime
from subprocess import check_call, check_output, CalledProcessError
import traceback
from StringIO import StringIO
import zipfile
import time

def print_exception():
    out = StringIO()
    traceback.print_exc(file=out)
    print out.getvalue().strip()

class Storage(object):

    API_VERSION = 3

    OBJ_SPEECH = "videos"
    OBJ_TEXT = "documents"
    OBJ_TEXT_TL = "translations"
    OBJ_WEB = "websites"
    OBJ_VALIGN = "valign"

    FT_DOC_INTERNAL = "translation"
    FT_DOC_PARSED = "parsed_document"
    FT_DOC_ORIGINAL = "ori_document"

    DOC_TL_STATUS_NOTL = 'not_translated'
    DOC_TL_STATUS_AUTO = 'fully_automatic'
    DOC_TL_STATUS_MIX = 'partially_human'
    DOC_TL_STATUS_MAN = 'fully_human'

    WEB_TL_STATUS_AUTO = 'automatic'
    WEB_TL_STATUS_MAN = 'revised'

    OP_CODE_NEW = 0
    OP_CODE_UPDATE = 1
    OP_CODE_DELETE = 2
    OP_CODE_CANCEL = 3

    def __init__(self, object_type, hostname, unix_username, unix_groupname, db_name, db_username, db_hostname, db_password=None):
        """ Constructor """
        if object_type not in [self.OBJ_SPEECH, self.OBJ_TEXT, self.OBJ_WEB, self.OBJ_VALIGN]:
            raise Exception('Bad object type')
        self.p_object_type = object_type
        self.host = hostname
        self.unix_username = unix_username
        self.unix_groupname = unix_groupname
        self.db, self.cur = self.__dbconnection__(db_name, db_username, db_hostname, db_password)
        mps = self.__get_mountpoints__()
        if self.p_object_type == self.OBJ_SPEECH:
            self.media_mp = mps['media']
            self.trans_mp = mps['transcriptions']
            self.uploads_mp = mps['uploads']
        elif self.p_object_type == self.OBJ_TEXT:
            self.media_mp = None
            self.trans_mp = mps['text_documents']
            self.uploads_mp = mps['text_uploads']
        elif self.p_object_type == self.OBJ_WEB:
            self.media_mp = None
            self.trans_mp = None
            self.uploads_mp = mps['web_uploads']
        elif self.p_object_type == self.OBJ_VALIGN:
            self.media_mp = mps['media']
            self.trans_mp = None
            self.uploads_mp = mps['valign_uploads'] 
        self.systems_mp = mps['systems']
        # For Speech:
        self.caption_type = self.__get_caption_types()
        self.object_type = self.__get_object_types()
        self._transaction_on = False
        self._transaction_op = None
        self._transaction_ids = []
        self._transaction_ts = None


    #--------------- Private Methods ----------------#


    # DB connection & select methods


    def __dbconnection__(self, db_name, db_user, db_host, db_password=None):
        """ Creates a connection to a postgreSQL database. """
        if db_password != None:
            db = psycopg2.connect(database=db_name, user=db_user, host=db_host, password=db_password);
        else:
            db = psycopg2.connect(database=db_name, user=db_user, host=db_host);
        cur = db.cursor(cursor_factory=psycopg2.extras.DictCursor)
        return db, cur

    def __get_mountpoints__(self):
        self.cur.execute("""SELECT name, path 
                            FROM mount_points p 
                            INNER JOIN machines m ON m.id = p.machine_id 
                            WHERE m.hostname = %s""",(self.host,))
        mps = { x[0]:x[1] for x in self.cur.fetchall() }
        return mps
 
    def __get_caption_types(self):
        self.cur.execute("""SELECT description, id 
                             FROM caption_types""")
        rows = self.cur.fetchall()
        return { r[0]:r[1] for r in rows }
 

    def __get_object_types(self):
        self.cur.execute("""SELECT table_name, id 
                             FROM object_types""")
        rows = self.cur.fetchall()
        return { r[0]:r[1] for r in rows }

 
    def __rows2data__(self, rows, keys):
        """ Converts rows returned after a SQL query to a list of Dictionaries using 
            the supplied list of keys. Data columns and their corresponding keys should be in the
            same position on both lists.

        """
        data = []
        for row in rows:
            d = {}
            for i in xrange(len(row)):
               d[keys[i]] = row[i]
            data.append(d)
        return data


    # Directory names and paths


    def __removed_lecture_dirprefix(self):
        return "removed"
    
    
    def __updated_lecture_dirsufix(self):
        return "old"


    def __lec_subdir(self, lec_hash): 
        return "%s/%s" % (lec_hash[0], lec_hash)


    def __current_lecture_media_dir_abspath(self, lec_hash):
        return "%s/%s" % (self.media_mp, self.__lec_subdir(lec_hash))


    def __current_lecture_trans_dir_abspath(self, lec_hash):
        return "%s/%s" % (self.trans_mp, self.__lec_subdir(lec_hash))


    def __removed_lecture_media_dir_abspath(self, lec_hash):
        return "%s/%s/%s" % (self.media_mp, self.__removed_lecture_dirprefix(), self.__lec_subdir(lec_hash))


    def __removed_lecture_trans_dir_abspath(self, lec_hash):
        return "%s/%s/%s" % (self.trans_mp, self.__removed_lecture_dirprefix(), self.__lec_subdir(lec_hash))


    def __updated_lecture_media_dir_abspath(self, lec_hash):
        return "%s/%s/%s" % (self.media_mp, self.__lec_subdir(lec_hash), self.__updated_lecture_dirsufix())


    def __updated_lecture_trans_dir_abspath(self, lec_hash):
        return "%s/%s/%s" % (self.trans_mp, self.__lec_subdir(lec_hash), self.__updated_lecture_dirsufix())


    def __file_relpath(self, lec_hash, file_name):
        return "%s/%s" % (self.__lec_subdir(lec_hash), file_name)


    def __media_file_abspath(self, lec_hash, file_name):
        return "%s/%s" % (self.media_mp, self.__file_relpath(lec_hash, file_name))

    
    def __trans_file_abspath(self, lec_hash, file_name):
        return "%s/%s" % (self.trans_mp, self.__file_relpath(lec_hash, file_name))


    def __upload_dir(self, up_id):
        return "%s/%s" % (self.uploads_mp, up_id)


    def __media_package_fp(self, up_id):
        return "%s/mp.zip" % self.__upload_dir(up_id)


    def __media_package_extraction_dir(self, up_id):
        return "%s/mp" % self.__upload_dir(up_id)

    
    def __manifest_fp(self, up_id):
        return "%s/manifest.json" % self.__media_package_extraction_dir(up_id)


    def __manifest_backup_fp(self, up_id):
        return "%s/manifest.json.bak" % self.__media_package_extraction_dir(up_id)


    def __outputs_dir(self, up_id):
        return "%s/outputs" % self.__upload_dir(up_id)


    def __outputs_manifest_fp(self, up_id):
        return "%s/manifest.json" % self.__outputs_dir(up_id)


    def __tmp_dir(self, up_id):
        return "%s/tmp" % self.__upload_dir(up_id)


    def __tmp_tx_dir(self, up_id):
        return "%s/tx" % self.__tmp_dir(up_id)


    def __tmp_tl_dir(self, up_id):
        return "%s/tl" % self.__tmp_dir(up_id)


    def __tmp_align_dir(self, up_id):
        return "%s/align" % self.__tmp_dir(up_id)


    def __tmp_media_dir(self, up_id):
        return "%s/media" % self.__tmp_dir(up_id)

    def __tmp_tts_dir(self, up_id):
        return "%s/tts" % self.__tmp_dir(up_id)

    # Media Package decompression and md5sum checking 


    def __unzip_mp(self, mp_fp, dest):
        """Unzips Media Package file.

        Arguments:

         mp_fp -- Media Package file path
         dest  -- Destination directory

        Returns:
  
         True -> Succeed.
         False -> Failed.

        """

        unzip_cmd = "unzip -ouj %s -d %s" % (mp_fp, dest)
        try:
            check_call(unzip_cmd, shell=True, 
                       stdout=open("%s/unzip.out" % dest, "w"), 
                       stderr=open("%s/unzip.err" % dest, "w"))
        except CalledProcessError:
            print_exception()
            return False
        return True


    def __compute_md5sum(self, fp):
        md5 = hashlib.md5()
        with open(fp, 'rb') as f:
            for chunk in iter(lambda: f.read(128 * md5.block_size), b''):
                md5.update(chunk)
        return md5.hexdigest()

    def __check_md5sum(self, fp, md5_proof):
        try:
            md5sum = self.__compute_md5sum(fp)
        except:
            print_exception()
            return 1
        if md5sum != md5_proof:
           return 2
        return 0


    # Storage management

    def __create_media_dir(self, lec_hash):
        dp = "%s/%s" % (self.media_mp, self.__lec_subdir(lec_hash))
        return self.__make_dir(dp)
   

    def __rm_media_dir(self, lec_hash):
        dp = "%s/%s" % (self.media_mp, self.__lec_subdir(lec_hash))
        return self.__rm_dir(dp)

 
    def __create_trans_dir(self, lec_hash):
        dp = "%s/%s" % (self.trans_mp, self.__lec_subdir(lec_hash))
        return self.__make_dir(dp)


    def __rm_trans_dir(self, lec_hash):
        dp = "%s/%s" % (self.trans_mp, self.__lec_subdir(lec_hash))
        return self.__rm_dir(dp)


    def __make_dir(self, path):
        if not(os.path.exists(path)):
            try:
                os.makedirs(path, 0775)
            except:
                return False
        return True


    def __rm_dir(self, path):
        try:
            shutil.rmtree(path)
        except:
            return False
        return True
    
    def write_list(self, flist, fp):
        with open(fp, 'w') as fd:
            for l in flist:
                if l == None:
                    fd.write("_\n")
                else:
                    fd.write("%s\n" % l.encode('utf-8'))
    

    def __cp_uploads2media(self, up_id, lec_hash, file_name, md5_fn=True):
        src_fp = self.media_package_file_abspath(file_name, up_id)
        if md5_fn:
            try:
                md5sum = self.__compute_md5sum(src_fp)
            except:
                print_exception()
                return {'code':10100, 'info':'Cannot compute md5sum of file "%s". Please contact system administrator.' % src_fp}
            ext = os.path.splitext(file_name)[1]
            new_file_name = "%s%s" % (md5sum, ext)
        else:
            new_file_name = file_name
        trg_fp = self.__media_file_abspath(lec_hash, new_file_name)
        try:
            shutil.copy(src_fp, trg_fp)
        except:
            print_exception()
            return {'code':10100, 'info':'Cannot copy file "%s" to disk. Please contact system administrator.' % file_name}
        return {'code':0, 'data':new_file_name}


    def copy_file(self, sf, tf):
        try:
            shutil.copy(sf, tf)
        except:
            return {'code':10230, 'info':'Could not copy file'}
        return {'code':0, 'info':''}


    def __cp_uploads2trans(self, up_id, lec_hash, file_name):
        src_fp = self.media_package_file_abspath(file_name, up_id)
        trg_fp = self.__trans_file_abspath(lec_hash, file_name)
        try:
            shutil.copy(src_fp, trg_fp)
        except:
            print_exception()
            return {'code':10101, 'info':'Cannot copy file "%s" to disk. Please contact system administrator.' % file_name}
        return {'code':0, 'data':file_name}

    def __cp_outputs2media(self, up_id, lec_hash, file_name, md5_fn=True):
        src_fp = self.output_file_abspath(file_name, up_id)
        if md5_fn:
            try:
                md5sum = self.__compute_md5sum(src_fp)
            except:
                print_exception()
                return {'code':10100, 'info':'Cannot compute md5sum of file "%s". Please contact system administrator.' % src_fp}
            ext = os.path.splitext(file_name)[1]
            new_file_name = "%s%s" % (md5sum, ext)
        else:
            new_file_name = file_name
        trg_fp = self.__media_file_abspath(lec_hash, new_file_name)
        try:
            shutil.copy(src_fp, trg_fp)
        except:
            print_exception()
            return {'code':10102, 'info':'Cannot copy file "%s" to disk. Please contact system administrator.' % file_name}
        return {'code':0, 'data':new_file_name}
 

    def __cp_outputs2trans(self, up_id, lec_hash, file_name):
        src_fp = self.output_file_abspath(file_name, up_id)
        trg_fp = self.__trans_file_abspath(lec_hash, file_name)
        try:
            shutil.copy(src_fp, trg_fp)
        except:
            print_exception()
            return {'code':10103, 'info':'Cannot copy file "%s" to disk. Please contact system administrator.' % file_name}
        return {'code':0, 'data':file_name}


    def __mv_media2removed(self, lec_hash):
        src_dir = self.__current_lecture_media_dir_abspath(lec_hash)
        trg_dir = self.__removed_lecture_media_dir_abspath(lec_hash)
        try:
            shutil.move(src_dir, trg_dir)
        except:
            print_exception()
            return False
        return True

 
    def __mv_trans2removed(self, lec_hash):
        src_dir = self.__current_lecture_trans_dir_abspath(lec_hash)
        trg_dir = self.__removed_lecture_trans_dir_abspath(lec_hash) 
        try:
            shutil.move(src_dir, trg_dir)
        except:
            print_exception()
            return False
        return True
       

    def __mv_media2updated(self, lec_hash):
        src_dir = self.__current_lecture_media_dir_abspath(lec_hash)
        tmp_dir = self.__current_lecture_media_dir_abspath(lec_hash)+'.tmp'
        trg_dir = self.__updated_lecture_media_dir_abspath(lec_hash) 
        try:
            shutil.move(src_dir, tmp_dir)
            shutil.move(tmp_dir, trg_dir)
        except:
            print_exception()
            return False
        return True

 
    def __mv_trans2updated(self, lec_hash):
        src_dir = self.__current_lecture_trans_dir_abspath(lec_hash)
        tmp_dir = self.__current_lecture_trans_dir_abspath(lec_hash)+'.tmp'
        trg_dir = self.__updated_lecture_trans_dir_abspath(lec_hash) 
        try:
            shutil.move(src_dir, tmp_dir)
            shutil.move(tmp_dir, trg_dir)
        except:
            print_exception()
            return False
        return True
 

    def __set_media_permissions(self, lec_hash):
        dp = "%s/%s" % (self.media_mp, self.__lec_subdir(lec_hash))
        return self.__set_permissions(dp)


    def __set_trans_permissions(self, lec_hash):
        dp = "%s/%s" % (self.trans_mp, self.__lec_subdir(lec_hash))
        return self.__set_permissions(dp)


    def __set_permissions(self, dir_path, mode=2775):
        cmd = "chmod -R %d %s" % (mode, dir_path)
        try:
            check_call(cmd, shell=True)
        except CalledProcessError:
            print_exception()
            return False
        cmd = "chown -R %s:%s %s" % (self.unix_username, self.unix_groupname, dir_path)
        try:
            check_call(cmd, shell=True)
        except CalledProcessError:
            print_exception()
            return False
        return True 


    # Directory backup and restore functions


    def __backup_manifest(self, up_id):
        try:
            shutil.copyfile(self.__manifest_fp(up_id), self.__manifest_backup_fp(up_id))
        except:
            print_exception()
            return False
        return True


    def __backup(self, lec_hash):
        src_media_dir = self.__current_lecture_media_dir_abspath(lec_hash)
        trg_media_dir = self.__current_lecture_media_dir_abspath(lec_hash)+'.bak'
        src_trans_dir = self.__current_lecture_trans_dir_abspath(lec_hash)
        trg_trans_dir = self.__current_lecture_trans_dir_abspath(lec_hash)+'.bak'
        self.__rm_dir(trg_media_dir)
        self.__rm_dir(trg_trans_dir)
        try:
            shutil.copytree(src_media_dir, trg_media_dir)
            shutil.copytree(src_trans_dir, trg_trans_dir)
        except:
            print_exception()
            return False
        return True


    def __restore_backup(self, lec_hash):
        src_media_dir = self.__current_lecture_media_dir_abspath(lec_hash)+'.bak'
        trg_media_dir = self.__current_lecture_media_dir_abspath(lec_hash)
        if not(os.path.exists(src_media_dir)):
            return False
        src_trans_dir = self.__current_lecture_trans_dir_abspath(lec_hash)+'.bak'
        trg_trans_dir = self.__current_lecture_trans_dir_abspath(lec_hash)
        if not(os.path.exists(src_trans_dir)):
            return False
        try:
            self.__rm_dir(trg_media_dir)
            shutil.copytree(src_media_dir, trg_media_dir)
            self.__rm_dir(src_media_dir)
            self.__rm_dir(trg_trans_dir)
            shutil.copytree(src_trans_dir, trg_trans_dir)
            self.__rm_dir(src_trans_dir)
        except:
            print_exception()
            return False
        return True


    def __rm_backup(self, lec_hash):
        media_dir = self.__current_lecture_media_dir_abspath(lec_hash)+'.bak'
        trans_dir = self.__current_lecture_trans_dir_abspath(lec_hash)+'.bak'
        try:
            self.__rm_dir(media_dir)
            self.__rm_dir(trans_dir)
        except:
            print_exception()
            return False
        return True


    def __rollback_insert(self, lec_hash, del_dirs=True):
        try:
            self.db.rollback()
        except:
            print_exception()
            return 1 
        if del_dirs:
            try:
                self.__rm_media_dir(lec_hash)
                self.__rm_trans_dir(lec_hash)
            except:
                pass
        return 0

 
    def __rollback_update(self, lec_hash, del_dirs=True):
        try:
            self.db.rollback()
        except:
            print_exception()
            return 1
        if del_dirs:
            media_dir = self.__updated_lecture_media_dir_abspath(lec_hash)
            trans_dir = self.__updated_lecture_trans_dir_abspath(lec_hash)
            try:
                self.__rm_dir(media_dir)
                self.__rm_dir(trans_dir)
            except:
                pass
        res = self.__restore_backup(lec_hash)
        if not(res):
            return 2
        return 0


    def __rollback_delete(self, lec_hash, del_dirs=True):
        try:
            self.db.rollback()
        except:
            print_exception()
            return 1
        if del_dirs:
            media_dir = self.__removed_lecture_media_dir_abspath(lec_hash)
            trans_dir = self.__removed_lecture_trans_dir_abspath(lec_hash)
            try:
                self.__rm_dir(media_dir)
                self.__rm_dir(trans_dir)
            except:
                pass
        res = self.__restore_backup(lec_hash)
        if not(res):
            return 2
        return 0


    # Insert and copy functions


    def __insert_lecture(self, ext_id, dur, lang, lec_hash, user_id):
        self.cur.execute("""INSERT INTO videos (external_id,duration,language,hash,owner_id) 
                                   VALUES (%s,%s,%s, %s, %s) 
                                   RETURNING id""", (ext_id, dur, lang, lec_hash, user_id))
        lec_id = self.cur.fetchone()[0]
        return lec_id


    def __update_lecture(self, lec_id, dur, lang):
        self.cur.execute("""UPDATE videos set duration = %s, language = %s
                                   WHERE id = %s""", (dur, lang, lec_id))


    def __check_and_insert_file(self, obj_id, obj_type_id, file_type, file_format, path, is_url=False):
        self.cur.execute("""SELECT id 
                                   FROM files
                                   WHERE object_id = %s
                                   AND object_type_id = %s 
                                   AND file_type_id = %s
                                   AND file_format = %s""", (obj_id, obj_type_id, file_type, file_format))
        res = self.cur.fetchall()
        if len(res) > 0:
            file_id = res[0][0]
            self.cur.execute("""UPDATE files
                                       SET file_path = %s,
                                           is_url = %s
                                       WHERE id = %s""", (path, is_url, file_id))
        else:
            return self.__insert_file(obj_id, obj_type_id, file_type, file_format, path, is_url)

      
    def __insert_file(self, obj_id, obj_type_id, file_type, file_format, path, is_url=False):
        self.cur.execute("""INSERT INTO files (object_id, object_type_id, file_type_id, file_format, file_path, is_url) 
                                   VALUES (%s,%s,%s,%s,%s,%s) 
                                   RETURNING id""", (obj_id, obj_type_id, file_type, file_format, path, is_url))
        file_id = self.cur.fetchone()[0]
        return file_id

        
    def __insert_metadata(self, lec_id, title, topic, keywords, date):
        self.cur.execute("""INSERT INTO videos_metadata (id,title,topic,keywords,date) 
                                   VALUES (%s,%s,%s,%s,%s)
                                   RETURNING id""", (lec_id, title, topic, keywords, date))
        lec_id = self.cur.fetchone()[0]
        return lec_id


    def __insert_speaker(self, external_id, name, gender, email, spk_hash, user_id):
        self.cur.execute("""INSERT INTO speakers (external_id, name, gender, email, hash, owner_id) 
                                 VALUES (%s, %s, %s, %s, %s, %s)
                                 RETURNING id""",(external_id, name, gender, email, spk_hash, user_id));
        int_spk_id = self.cur.fetchone()[0]
        return int_spk_id
    

    def __insert_speaker_lecture(self, spk_id, lec_id):
        self.cur.execute("""INSERT INTO speakers_videos (speaker_id, video_id) 
                                   VALUES (%s,%s)""", (spk_id, lec_id))


    def __insert_user_lecture(self, user_id, lec_id):
        self.cur.execute("""INSERT INTO users_videos (user_id, video_id) 
                                   VALUES (%s,%s)""", (user_id, lec_id))


    def __insert_caption(self, lec_id, lang, ctype, sub_type, sup_ratio, system_id, from_lang):
        self.cur.execute("""SELECT id FROM captions
                                   WHERE video_id = %s
                                   AND language = %s""", (lec_id, lang))
        res = self.cur.fetchall()
        if len(res) == 0:
            self.cur.execute("""INSERT INTO captions (video_id,language,type,sub_type,sup_ratio,system_id, from_language) 
                              VALUES (%s,%s,%s,%s,%s,%s,%s)
                              RETURNING id""", (lec_id, lang, ctype, sub_type, sup_ratio, system_id, from_lang))
            cap_id = self.cur.fetchone()[0]
        else:
            cap_id = res[0][0]
            self.cur.execute("""UPDATE captions 
                              SET type = %s, sup_ratio = %s, system_id = %s, from_language = %s
                              WHERE id = %s""", (ctype, sup_ratio, system_id, from_lang, cap_id))
        return cap_id


    def __insert_audiotrack(self, lec_id, lang, at_type, system_id, from_subs, vgender):
        cap_id = None
        if from_subs:
            self.cur.execute("""SELECT id FROM captions
                                   WHERE video_id = %s
                                   AND language = %s""", (lec_id, lang))
            cap_id = self.cur.fetchone()[0]

        self.cur.execute("""SELECT id FROM audiotracks
                                   WHERE video_id = %s
                                   AND language = %s
                                   AND voice_type = %s""", (lec_id, lang, at_type))
        res = self.cur.fetchall()
        if len(res) == 0:
            self.cur.execute("""INSERT INTO audiotracks (video_id,caption_id,language,voice_type,voice_gender,system_id) 
                              VALUES (%s,%s,%s,%s,%s,%s)
                              RETURNING id""", (lec_id, cap_id, lang, at_type, vgender, system_id))
            at_id = self.cur.fetchone()[0]
        else:
            at_id = res[0][0]
            self.cur.execute("""UPDATE audiotracks
                              SET system_id = %s, caption_id = %s, voice_gender = %s
                              WHERE id = %s""", (system_id, cap_id, vgender, at_id))
        return at_id



    def __insert_and_copy_file(self, file_name, file_type, file_format, up_id, ext_id, object_id, object_type_id, lec_hash, check_duplicated=False, output_file=False, is_url=False):
        if is_url:
            path = file_name
        else:
            if object_type_id in (self.object_type['videos'], self.object_type['audiotracks']):
                if object_type_id == self.object_type['videos']:
                    md5_fn = True
                else:
                    md5_fn = False
                if output_file:
                    res = self.__cp_outputs2media(up_id, lec_hash, file_name, md5_fn)
                else:
                    res = self.__cp_uploads2media(up_id, lec_hash, file_name, md5_fn)
            elif object_type_id == self.object_type['captions']:
                if output_file:
                    res = self.__cp_outputs2trans(up_id, lec_hash, file_name)
                else:
                    res = self.__cp_uploads2trans(up_id, lec_hash, file_name)
            if res['code'] != 0:
                return res
            new_file_name = res['data']
            path = self.__file_relpath(lec_hash, new_file_name)

        if check_duplicated:
            try:
                self.__check_and_insert_file(object_id, object_type_id, file_type, file_format, path, is_url)
            except:
                print_exception()
                return {'code':10000, 'info':'Cannot insert file "%s" into DB. Please contact system administrator.' % file_name}
        else:
            try:
                self.__insert_file(object_id, object_type_id, file_type, file_format, path, is_url)
            except:
                print_exception()
                return {'code':10001, 'info':'Cannot insert file "%s" into DB. Please contact system administrator.' % file_name}
        return {'code':0}


    def __check_and_insert_speaker_lecture(self, speakers, lec_id, user_id):
        for s in speakers:
            try:
                spk_id = s["speaker_id"]
            except KeyError:
                return {'code':10003, 'info':'speaker_id key not found in speakers when inserting lecture into DB.'}
            try:
                spk_name = s["speaker_name"]
            except KeyError:
                return {'code':10089, 'info':'speaker_name key not found in speakers when inserting lecture into DB.'}
            spk_hash = self.get_hash(spk_id, user_id)
            res = self.get_speaker_info(spk_hash)
            if res['code'] != 0:
                return res
            if len(res['data']) == 0:
                if spk_name == None: spk_name = ""
                spk_gender = s.get("speaker_gender", None)
                spk_email = s.get("speaker_email", None)
                try:
                    int_spk_id = self.__insert_speaker(spk_id.encode('utf-8'), spk_name.encode('utf-8'), spk_gender, spk_email, spk_hash, user_id) 
                except:
                    print_exception()
                    return {'code':10004, 'info':'Cannot insert speaker into DB. Please contact System Administrator.'}
            elif len(res['data']) == 1:
                int_spk_id = res['data'][0]['id']
            else:
                return {'code':10005, 'info':'Unexpected result from DB when retrieving speaker info. Please contact System Administrator.'}
            try:
                self.__insert_speaker_lecture(int_spk_id, lec_id)
            except:
                print_exception()
                return {'code':10006, 'info':'Cannot insert speaker-lecture relation into DB. Please contact System Administrator.'}
        return {'code':0}


    def __insert_and_copy_media(self, media, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=False):
        # main media file(s)
        for m in media:
            try:
                fn = m['filename']
            except KeyError:
                return {'code':10007, 'info':'filename key not found on file. Please contact system administrator.'}
            try:
                ft = m['filetype']
            except KeyError:
                return {'code':10008, 'info':'filetype key not found on file "%s". Please contact system administrator.' % fn}
            try:
                ff = m['fileformat']
            except KeyError:
                return {'code':10084, 'info':'fileformat key not found on file "%s". Please contact system administrator.' % fn}
            outf = m.get('output_file', False)
            is_url = True if not(m.get('store', True)) else False
            fn = m['url'] if is_url else fn
            res = self.__insert_and_copy_file(fn, ft, ff, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=check_duplicated, output_file=outf, is_url=is_url)
            if res['code'] != 0:
                return res
        return {'code':0}


    def __insert_and_copy_slides(self, slides, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=False):
        for sl in slides:
            try:
                fn = sl['filename']
            except KeyError:
                return {'code':10009, 'info':'filename key not found on file. Please contact system administrator.'}
            try:
                ft = sl['filetype']
            except KeyError:
                return {'code':10010, 'info':'filetype key not found on file "%s". Please contact system administrator.' % fn}
            try:
                ff = sl['fileformat']
            except KeyError:
                return {'code':10085, 'info':'fileformat key not found on file "%s". Please contact system administrator.' % fn}
            outf = sl.get('output_file', False)
            if sl.get('store', True):
                res = self.__insert_and_copy_file(fn, ft, ff, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=check_duplicated, output_file=outf)
                if res['code'] != 0:
                   return res
        return {'code':0}


    def __insert_and_copy_attachments(self, attachments, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=False):
        # attachment
        for att in attachments:
            try:
                fn = att['filename']
            except KeyError:
                return {'code':10011, 'info':'filename key not found on file. Please contact system administrator.'}
            try:
                ft = att['filetype']
            except KeyError:
                return {'code':10012, 'info':'filetype key not found on file "%s". Please contact system administrator.' % fn}
            try:
                ff = att['fileformat']
            except KeyError:
                return {'code':10086, 'info':'fileformat key not found on file "%s". Please contact system administrator.' % fn}
            outf = att.get('output_file', False)
            res = self.__insert_and_copy_file(fn, ft, ff, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=check_duplicated, output_file=outf)
            if res['code'] != 0:
                return res
        return {'code':0}

    def has_supervised_subs(self, lec_hash, lang):
        self.cur.execute("""SELECT count(*) FROM captions WHERE video_id = (SELECT id from videos where hash = %s) AND language = %s AND type > 0""", (lec_hash, lang))
        ret = self.cur.fetchone()[0]
        return {'code':0, 'data': bool(ret) }


    def __insert_and_copy_captions(self, captions, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=False):
        for cap in captions:
            try:
                fn = cap['filename']
            except KeyError:
                return {'code':10013, 'info':'filename key not found on file. Please contact system administrator.'}
            try:
                ft = cap['filetype']
            except KeyError:
                return {'code':10014, 'info':'filetype key not found on file "%s". Please contact system administrator.' % fn}
            try:
                ff = cap['fileformat']
            except KeyError:
                return {'code':10087, 'info':'fileformat key not found on file "%s". Please contact system administrator.' % fn}
            try:
                lang = cap['language']
            except KeyError:
                return {'code':10015, 'info':'language key not found in caption. Please contact system administrator.'} 
            try:
                sub_type = cap['type']
            except KeyError:
                return {'code':10111, 'info':'sub_type key not found in caption. Please contact system administrator.'} 
            try:
                manual = cap['human']
            except KeyError:
                return {'code':10016, 'info':'manual key not found in caption. Please contact system administrator.'} 
            try:
                system_id = cap['system_id']
            except KeyError:
                return {'code':10111, 'info':'system_id key not found in caption. Please contact system administrator.'} 
            try:
                from_lang = cap['from_language']
            except KeyError:
                return {'code':10111, 'info':'from_language key not found in caption. Please contact system administrator.'} 

            ctype = self.caption_type['fully_automatic'] ######### -> Improvement!
            sup_ratio = 0
            if manual:
                ctype = self.caption_type['fully_human'] ########### -> Improvement!
                sup_ratio = 100
            try:
                cap_id = self.__insert_caption(lec_id, lang, ctype, sub_type, sup_ratio, system_id, from_lang) 
            except:
                print_exception()
                return {'code':10017, 'info':'Cannot insert caption into DB. Please contact system administrator.'}
            outf = cap.get('output_file', False)
            res = self.__insert_and_copy_file(fn, ft, ff, up_id, ext_id, cap_id, obj_type_id, lec_hash, check_duplicated=check_duplicated, output_file=outf)
            if res['code'] != 0:
                return res
        return {'code':0}

    def __insert_and_copy_audiotracks(self, audiotracks, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=False):
        for at in audiotracks:
            try:
                fn = at['filename']
            except KeyError:
                return {'code':10013, 'info':'filename key not found on file. Please contact system administrator.'}
            try:
                ft = at['filetype']
            except KeyError:
                return {'code':10014, 'info':'filetype key not found on file "%s". Please contact system administrator.' % fn}
            try:
                ff = at['fileformat']
            except KeyError:
                return {'code':10087, 'info':'fileformat key not found on file "%s". Please contact system administrator.' % fn}
            try:
                lang = at['language']
            except KeyError:
                return {'code':10015, 'info':'language key not found in caption. Please contact system administrator.'} 
            try:
                at_type = at['type']
            except KeyError:
                return {'code':10111, 'info':'sub_type key not found in caption. Please contact system administrator.'} 
            try:
                system_id = at['system_id']
            except KeyError:
                return {'code':10111, 'info':'system_id key not found in caption. Please contact system administrator.'} 

            try:
                from_subs = at['from_subs']
            except KeyError:
                return {'code':10111, 'info':'from_subs key not found in caption. Please contact system administrator.'} 

            try:
                vgender = at['voice_gender']
            except KeyError:
                return {'code':10111, 'info':'voice_gender key not found in caption. Please contact system administrator.'} 

            try:
                at_id = self.__insert_audiotrack(lec_id, lang, at_type, system_id, from_subs, vgender) 
            except:
                print_exception()
                return {'code':10017, 'info':'Cannot insert caption into DB. Please contact system administrator.'}
            outf = at.get('output_file', False)
            res = self.__insert_and_copy_file(fn, ft, ff, up_id, ext_id, at_id, obj_type_id, lec_hash, check_duplicated=check_duplicated, output_file=outf)
            if res['code'] != 0:
                return res
        return {'code':0}


    def __insert_and_copy_waveform(self, waveform, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=False):
        try:
            fn = waveform['filename']
        except KeyError:
            return {'code':10013, 'info':'filename key not found on file. Please contact system administrator.'}
        try:
            ft = waveform['filetype']
        except KeyError:
            return {'code':10014, 'info':'filetype key not found on file "%s". Please contact system administrator.' % fn}
        try:
            ff = waveform['fileformat']
        except KeyError:
            return {'code':10087, 'info':'fileformat key not found on file "%s". Please contact system administrator.' % fn}

        outf = waveform.get('output_file', False)
        res = self.__insert_and_copy_file(fn, ft, ff, up_id, ext_id, lec_id, obj_type_id, lec_hash, check_duplicated=check_duplicated, output_file=outf)
        if res['code'] != 0:
            return res
        return {'code':0}



    def __update_metadata(self, lec_id, title, topic, keywords, date):
        #self.cur.execute("""UPDATE videos_metadata 
        #                           SET (title,topic,keywords,date) =
        #                           (%s,%s,%s,%s)
        #                           WHERE id = %s""", (title, topic, keywords, date, lec_id))
        self.cur.execute("""UPDATE videos_metadata 
                                   SET (title,topic,keywords) =
                                   (%s,%s,%s)
                                   WHERE id = %s""", (title, topic, keywords, lec_id))

    def __delete_for_update(self, lec_id):
        self.cur.execute("""DELETE FROM files WHERE object_id = %s AND object_type_id = %s""", (lec_id, self.object_type['videos']))
        self.cur.execute("""DELETE FROM files WHERE object_id IN (SELECT id FROM captions WHERE video_id = %s) AND object_type_id = %s""", (lec_id, self.object_type['captions']))
        self.cur.execute("""DELETE FROM files WHERE object_id IN (SELECT id FROM audiotracks WHERE video_id = %s) AND object_type_id = %s""", (lec_id, self.object_type['audiotracks']))
        self.cur.execute("""DELETE FROM audiotracks WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM captions WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM speakers_videos WHERE video_id = %s""", (lec_id,))


    def __hard_delete_lecture_db(self, lec_id):
        self.cur.execute("""DELETE FROM files WHERE object_id = %s AND object_type_id = %s""", (lec_id, self.object_type['videos']))
        self.cur.execute("""DELETE FROM files WHERE object_id IN (SELECT id FROM captions WHERE video_id = %s) AND object_type_id = %s""", (lec_id, self.object_type['captions']))
        self.cur.execute("""DELETE FROM files WHERE object_id IN (SELECT id FROM audiotracks WHERE video_id = %s) AND object_type_id = %s""", (lec_id, self.object_type['audiotracks']))
        self.cur.execute("""DELETE FROM mods WHERE mod_session_id IN (select id from mod_sessions where object_id = %s AND object_type = %s)""", (lec_id, Storage.OBJ_SPEECH))
        self.cur.execute("""DELETE FROM mod_sessions WHERE object_id = %s AND object_type = %s""", (lec_id, Storage.OBJ_SPEECH))
        self.cur.execute("""DELETE FROM audiotracks WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM captions WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM speakers_videos WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM users_videos WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM groups_videos WHERE video_id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM videos_metadata WHERE id = %s""", (lec_id,))
        self.cur.execute("""DELETE FROM videos WHERE id = %s""", (lec_id,))


    def __delete_lecture_db(self, lec_id):
        self.cur.execute("""UPDATE files 
                                   SET file_path = concat(%s, file_path) 
                                   WHERE id IN 
                                         (SELECT f2.id 
                                          FROM files f2 
                                          WHERE f2.object_id = %s AND
                                                f2.object_type_id = %s)""", 
                           (self.__removed_lecture_dirprefix()+'/', lec_id, self.object_type['videos']))
        self.cur.execute("""UPDATE files 
                                   SET file_path = concat(%s, file_path) 
                                   WHERE id IN 
                                         (SELECT f2.id 
                                          FROM files f2 
                                          INNER JOIN captions c on c.id = f2.object_id 
                                          WHERE c.video_id = %s AND 
                                          f2.object_type_id = %s)""", 
                            (self.__removed_lecture_dirprefix()+'/', lec_id, self.object_type['captions']))
        self.cur.execute("""UPDATE videos_metadata 
                                   SET removed = True 
                                   WHERE id = %s""", (lec_id,))
        

    def __mark_lecture_deleted(self, lec_id):
        self.cur.execute("""UPDATE videos_metadata 
                                   SET removed = True 
                                   WHERE id = %s""", (lec_id,))
 

    #--------------- Public Methods ----------------#

    
    # Database information


    LECTURE_KEYS = [ 'id', 'external_id', 'language', 'duration', 'removed', 'title', 'user_id' ]

    def get_lecture_info(self, lec_hash):
        """ Returns lecture info. """
        try:
            self.cur.execute("""SELECT v.id, v.external_id, v.language, v.duration, m.removed, m.title, v.owner_id
                                FROM videos v
                                INNER JOIN videos_metadata m ON v.id = m.id
                                WHERE hash = %s""",(lec_hash,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10018, 'info':'Cannot get lecture info'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.LECTURE_KEYS)
        return {'code':0, 'data':data}

    
    SPEAKER_KEYS = [ 'id', 'external_id', 'name', 'gender', 'age', 'email' ]

    def get_speaker_info(self, spk_hash):
        """ Returns speaker info. """
        try:
            self.cur.execute("""SELECT id, external_id, name, gender, age, email 
                                FROM speakers
                                WHERE hash = %s""",(spk_hash,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10019, 'info':'Cannot get speaker info'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.SPEAKER_KEYS)
        return {'code':0, 'data':data}

 
    CAPTION_KEYS = [ 'id', 'language', 'type', 'system_id', 'from_language' ]

    def get_captions(self, lec_hash):
        """ Returns speaker info. """
        try:
            self.cur.execute("""SELECT c.id, c.language, type, system_id, from_language
                                FROM captions c
                                INNER JOIN videos v ON c.video_id = v.id
                                WHERE v.hash = %s""",(lec_hash,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10020, 'info':'Cannot get captions'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.CAPTION_KEYS)
        return {'code':0, 'data':data}
  
    DOC_TRANSLATION_KEYS = [ 'id', 'language', 'status', 'sup_ratio', 'from_language', 'from_translation_id', 'system_id', 'dtlx_fp' ]

    def get_document_translations(self, obj_hash):
        try:
            self.cur.execute("""SELECT t.id, t.language, t.status, t.sup_ratio, t.from_language, t.from_translation_id, t.system_id, '""" + self.trans_mp + "/" + """' || f.file_path
                                FROM translations t 
                                INNER JOIN documents d ON t.document_id = d.id 
                                INNER JOIN files f ON f.object_id = t.id
                                WHERE d.hash = %s
                                      AND f.object_type_id = (select id from object_types where table_name = %s) 
                                      AND f.file_type_id = (select id from file_types where name = %s)""",(obj_hash, self.OBJ_TEXT_TL, self.FT_DOC_INTERNAL))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10020, 'info':'Cannot get translations'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.DOC_TRANSLATION_KEYS)
        return {'code':0, 'data':data}
 

    AUDIOTRACKS_KEYS = [ 'id', 'language', 'voice_type', 'system_id' ]

    def get_audiotracks(self, lec_hash):
        """ Returns speaker info. """
        try:
            self.cur.execute("""SELECT a.id, a.language, a.voice_type, a.system_id 
                                FROM audiotracks a
                                INNER JOIN videos v ON a.video_id = v.id
                                WHERE v.hash = %s""",(lec_hash,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10020, 'info':'Cannot get audiotracks'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.AUDIOTRACKS_KEYS)
        return {'code':0, 'data':data}



    def get_file_types(self):
        """ Returns dict of file types

        """
        try:
            self.cur.execute("""SELECT name, id 
                                FROM file_types""")
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10022, 'info':'Cannot get file types list'}
        # Create List of Dicts and return Data object with the results
        data = { r[0]:r[1] for r in rows }
        return {'code':0, 'data':data}


    def get_file_formats(self):
        """ Returns dict of file types

        """
        try:
            self.cur.execute("""SELECT name
                                FROM file_formats""")
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10081, 'info':'Cannot get file formats list'}
        # Create List and return results
        data = [ r[0] for r in rows ]
        return {'code':0, 'data':data}


    FILES_KEYS = [ 'file_type_id', 'file_format', 'file_path' ]

    def get_files(self, lec_hash):
        """ Returns lecture info. """
        try:
            self.cur.execute("""SELECT f.file_type_id, f.file_format, f.file_path
                                FROM files f
                                INNER JOIN videos v ON v.id = f.object_id 
                                WHERE f.object_type_id = %s
                                AND v.hash = %s""", (self.object_type['videos'], lec_hash))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10023, 'info':'Cannot get files list.'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.FILES_KEYS)
        return {'code':0, 'data':data}


    USERS_KEYS = [ 'id', 'username', 'name', 'is_admin', 'email', 'password' ]

    def get_user(self, user_id):
        try:
            self.cur.execute("""SELECT id, username, name, admin, email, password 
                                 FROM users
                                 WHERE id = %s""", (user_id,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10095, 'info':'Cannot get user.'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.USERS_KEYS)
        return {'code':0, 'data':data}

    def get_user_by_username(self, user_name):
        try:
            self.cur.execute("""SELECT id, username, name, admin, email, password 
                                 FROM users
                                 WHERE username = %s""", (user_name,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10104, 'info':'Cannot get user.'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.USERS_KEYS)
        return {'code':0, 'data':data}




    def get_media_file_abspath(self, lec_hash):
        """ Returns media file path of a given lecture's external ID """
        try:
            self.cur.execute("""SELECT f.file_path, f.is_url, f.file_format 
                                FROM files f 
                                INNER JOIN videos v ON v.id = f.object_id 
                                WHERE f.object_type_id = %s 
                                      AND f.file_type_id = 0 
                                      AND f.file_format = 'mp4'
                                      AND v.hash = %s""", (self.object_type['videos'], lec_hash)) #### hardcoded file_type!!!!!!!!!!!
        except:
            print_exception()
            return {'code':10024, 'info':'Cannot get media file path.'}
        try:
            row = self.cur.fetchone()
        except:
            print_exception()
            return {'code':10025, 'info':'Media file not found on DB.'}
        if row == None:
            return {'code':10025, 'info':'Media file not found on DB.'}
        path = row[0]
        is_url = row[1]
        ext = row[2]
        if is_url:
            media_fp = path
        else:
            media_fp = self.get_data_file_abspath(path)
        return {'code':0, 'data':media_fp, 'is_url':is_url, 'ext':ext}


    def get_dfxp_file_abspath(self, lec_hash, lang):
        try:
            self.cur.execute("""SELECT f.file_path 
                                FROM files f 
                                INNER JOIN captions c ON c.id = f.object_id 
                                INNER JOIN videos v ON v.id = c.video_id 
                                WHERE f.object_type_id = %s 
                                      AND f.file_type_id = 4
                                      AND f.file_format = 'dfxp'
                                      AND v.hash = %s
                                      AND c.language = %s""", (self.object_type['captions'], lec_hash, lang)) ##### hardcoded file_type!!!!!!
        except:
            print_exception()
            return {'code':10026, 'info':'Cannot get dfxp file path.'}
        try:
            row = self.cur.fetchone()
        except:
            print_exception()
            return {'code':10027, 'info':'DFXP file not found on DB.'}
        dfxp_fp = self.get_trans_file_abspath(row[0])
        return {'code':0, 'data':dfxp_fp}

 
    SYSTEMS_KEYS = [ 'id', 'lang1', 'lang2', 'type', 'name', 'description', 'file_path', 'public', 'as_default', 'enabled', 'tts_voice_gender', 'user_list', 'features_tadapt', 'rtf', 'const_time' ]

    def get_systems(self):
        """ Returns systems """
        try:
            self.cur.execute("""SELECT id, lang1, lang2, type, name, description, module_path, public, as_default, enabled, tts_voice_gender,
                                       (SELECT array ( SELECT us.user_id 
                                                       FROM users_systems us 
                                                       WHERE us.system_id = s.id )) AS user_list, features_tadapt, rtf, const_time
                                FROM systems s""")
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10097, 'info':'Cannot get systems'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.SYSTEMS_KEYS)
        return {'code':0, 'data':data}

    def get_user_default_systems(self, user_id):
        try:
            self.cur.execute("""SELECT id, lang1, lang2, type, name, description, module_path, public, s.as_default, enabled, NULL
                                FROM users_systems us
                                INNER JOIN systems s ON s.id = us.system_id
                                WHERE us.user_id = %s
                                AND us.as_default = True""", (user_id,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10097, 'info':'Cannot get systems'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.SYSTEMS_KEYS)
        return {'code':0, 'data':data}
   


    # Media Package Management


    def create_media_package_extraction_dir(self, up_id):
        dest_dir = self.__media_package_extraction_dir(up_id)
        # Create dest_dir if does not exist
        if not(os.path.isdir(dest_dir)):
            try:
                os.makedirs(dest_dir, 0775)
            except OSError:
                return {'code':10211, 'info':'Cannot create destination dir for extracting Media Package.'}
        return {'code':0}

    def exists_media_package_file(self, up_id):
        return os.path.exists(self.__media_package_fp(up_id))

    def extract_media_package(self, up_id):
        mp_fp = self.__media_package_fp(up_id)
        if os.path.exists(mp_fp):
            dest_dir = self.__media_package_extraction_dir(up_id)
            # Create dest_dir if does not exist
            if not(os.path.isdir(dest_dir)):
                try:
                    os.makedirs(dest_dir, 0775)
                except OSError:
                    return {'code':10029, 'info':'Cannot create destination dir for extracting Media Package.'}
            # Unzip Media Package
            res = self.__unzip_mp(mp_fp, dest_dir)
            if not(res):
                return {'code':10030, 'info':'Cannot unzip Media Package.'}
            return {'code':0, 'exists':True} 
        else:
            return {'code':0, 'exists':False}

    def load_manifest(self, up_id):
        """Load into memory the contents of a manifest.json file

        Arguments:
    
         manif -- manifest.json file path.
    
        Returns:
     
         Dict -> {'code':INT, 'info':STR, 'data':DICT }
         where 'data' contains the parsed JSON data in a Dict.

        """
        manif_fp = self.__manifest_fp(up_id)
        # Load JSON data from manifest.json
        try:
            manif = json.load(open(manif_fp), encoding='UTF-8')
        except IOError:
            return {'code':10031, 'info':'manifest.json file not found in Media Package.'}
        except ValueError:
            print_exception()
            return {'code':10032, 'info':'Invalid JSON format: Cannot parse JSON object. Please check syntax.'}
        # Return JSON data
        return {'code':0, 'info':'', 'data':manif}


    def store_manifest(self, up_id, manif, backup=False):
        if backup:
            res = self.__backup_manifest(up_id)
            if not(res):
                return {'code':10033, 'info':'Cannot backup manifest.json file. Please contact system administrator.'}
        manif_fp = self.__manifest_fp(up_id)
        try:
            manif_fd = open(manif_fp, 'w')
        except IOError:
            return {'code':10034, 'info':'Cannot open output manifest.json file for write sanitized manifest. Please contact system administrator.'}
        try:
            manif_fd.write(json.dumps(manif, encoding='UTF-8', indent=4, separators=(',', ': '), ensure_ascii=False).encode('utf-8'))
        except:
            print_exception()
            return {'code':10035, 'info':'Cannot dump and store output sanitized manifest.json file. Please contact system administrator.'}
        try:
            manif_fd.close()
        except:
            print_exception()
            pass
        return {'code':0}



    def check_mp_file(self, up_id, fn, md5):
        fp = self.media_package_file_abspath(fn, up_id)
        if not(os.path.isfile(fp)):
            return {'code':10036, 'info':'File "%s" not found in media package.' % fn}
        res = self.__check_md5sum(fp, md5)
        if res == 1: 
            return {'code':10037, 'info':'Cannot compute md5sum of the file "%s". Please contact system administrator.' % fn}
        elif res == 2:
            return {'code':10038, 'info':'Non-matching md5sums of file "%s"' % fn}
        else:
            return {'code':0}


    def unzip_attached_package_file(self, up_id, fn):
        fp = self.media_package_file_abspath(fn, up_id)
        dest = fp + ".ex"
        if not(os.path.isfile(fp)):
            return {'code':10081, 'info':'File "%s" not found in media package.' % fn}
        try:
            zf = zipfile.ZipFile(fp, 'r', allowZip64=True)
        except:
            return {'code':10081, 'info':'Could not open attached zip file (possibly bad format).'}
        try:
            zf.extractall(path=dest)
        except:
            print_exception()
            return {'code':10081, 'info':'Could not extract attached zip file.'}
        return {'code':0}


    def rename_mp_file(self, up_id, fn, fn_new):
        fp = self.media_package_file_abspath(fn, up_id)
        if not(os.path.isfile(fp)):
            return {'code':10081, 'info':'File "%s" not found in media package.' % fn}
        fp_new = self.media_package_file_abspath(fn_new, up_id)
        #if os.path.isfile(fp_new):
        #    if not(os.path.isfile(fp)):
        #         return {'code':0}
        #    #else:
        #    #     return {'code':10082, 'info':'Cannot rename input file "%s". New file name "%s" already exists in media package.' % (fn, fn_new)}
        try:
            os.rename(fp, fp_new)
        #    shutil.move(fp, fp_new)
        except:
            print_exception()
            return {'code':10083, 'info':'Cannot rename input file "%s". Please contact system administrator' % fp}
        return {'code':0}


    # Ingest Outputs management


    def get_data_file_abspath(self, relpath):
        return "%s/%s" % (self.media_mp, relpath.decode('utf-8')) # Temporal fix for filenames with non-ascii characters

 
    def get_trans_file_abspath(self, relpath):
        return "%s/%s" % (self.trans_mp, relpath.decode('utf-8')) # Temporal fix for filenames with non-ascii characters


    def output_file_abspath(self, fn, up_id, subdir=None):
        if subdir != None:
            return "%s/%s/%s" % (self.__outputs_dir(up_id), subdir, fn)
        else:
            return "%s/%s" % (self.__outputs_dir(up_id), fn)


    def media_package_file_abspath(self, fn, up_id):
        return "%s/%s" % (self.__media_package_extraction_dir(up_id), fn)


    def create_outputs_dir(self, up_id):
        dp = self.__outputs_dir(up_id)
        res =  self.__make_dir(dp)
        if not(res):
            return {'code':10039, 'info':'Cannot create outputs dir. Please contact system administrator.'}
        return {'code':0}

    def get_outputs_subdir(self, up_id, subdir):
        return "%s/%s" % (self.__outputs_dir(up_id), subdir)

    def get_outputs_subdir_file_abspath(self, up_id, subdir, fn):
        return "%s/%s" % (self.get_outputs_subdir(up_id, subdir), fn)

    def create_outputs_subdir(self, up_id, subdir):
        dp = self.get_outputs_subdir(up_id, subdir)
        res =  self.__make_dir(dp)
        if not(res):
            return {'code':10039, 'info':'Cannot create outputs dir. Please contact system administrator.'}
        return {'code':0, 'data':dp}



    def store_outputs_manifest(self, up_id, manif):
        manif_fp = self.__outputs_manifest_fp(up_id)
        try:
            manif_fd = open(manif_fp, 'w')
        except IOError:
            return {'code':10040, 'info':'Cannot open output manifest.json file for write. Please contact system administrator.'}
        try:
            json.dump(manif, manif_fd, encoding='UTF-8', ensure_ascii=False)
        except:
            print_exception()
            return {'code':10041, 'info':'Cannot dump and store output manifest.json file. Please contact system administrator.'}
        try:
            manif_fd.close()
        except:
            print_exception()
            pass
        return {'code':0}


    def load_outputs_manifest(self, up_id):
        manif_fp = self.__outputs_manifest_fp(up_id)
        # Load JSON data from manifest.json
        try:
            manif_fd = open(manif_fp)
        except IOError:
            return {'code':10042, 'info':'Cannot open output manifest.json file for reading. Please contact system administrator.'}
        try:
            manif = json.load(manif_fd, encoding='UTF-8')
        except:
            print_exception()
            return {'code':10043, 'info':'Invalid format of output manifest.json. '}
        try:
            manif_fd.close()
        except:
            print_exception()
            pass
        # Return JSON data
        return {'code':0, 'info':'', 'data':manif}


    def tmp_tx_file_abspath(self, up_id, file_path):
        return "%s/%s" % (self.__tmp_tx_dir(up_id), file_path)

    def tmp_tts_file_abspath(self, up_id, file_path):
        return "%s/%s" % (self.__tmp_tts_dir(up_id), file_path)

    def tmp_tx_dir(self, up_id):
        return self.__tmp_tx_dir(up_id)

    def tmp_align_dir(self, up_id):
        return self.__tmp_align_dir(up_id)

    def tmp_tts_dir(self, up_id):
        return self.__tmp_tts_dir(up_id)

    def tmp_media_dir(self, up_id):
        return self.__tmp_media_dir(up_id)


    def tmp_tl_file_abspath(self, up_id, file_path):
        return "%s/%s" % (self.__tmp_tl_dir(up_id), file_path)


    def tmp_tl_dir(self, up_id):
        return self.__tmp_tl_dir(up_id)


    def tmp_tl_pair_dir(self, up_id, src_lang, trg_lang):
        return "%s/%s-%s" % (self.__tmp_tl_dir(up_id), src_lang, trg_lang)

  
    def tmp_tl_pair_object_dir(self, up_id, src_lang, trg_lang, obj_hash):
        return "%s/objects/%s" % (self.tmp_tl_pair_dir(up_id, src_lang, trg_lang), obj_hash)


    def tmp_tl_pair_file_abspath(self, up_id, src_lang, trg_lang, file_path):
        return "%s/%s" % (self.tmp_tl_pair_dir(up_id, src_lang, trg_lang), file_path)

    def tmp_tl_pair_object_file_abspath(self, up_id, src_lang, trg_lang, obj_hash, file_path):
        return "%s/%s" % (self.tmp_tl_pair_object_dir(up_id, src_lang, trg_lang, obj_hash), file_path)

    def tmp_tts_lang_dir(self, up_id, lang):
        return "%s/%s" % (self.__tmp_tts_dir(up_id), lang)

    def tmp_tts_lang_file_abspath(self, up_id, lang, file_path):
        return "%s/%s" % (self.tmp_tts_lang_dir(up_id, lang), file_path)

    def tmp_media_file_abspath(self, up_id, file_path):
        return "%s/%s" % (self.__tmp_media_dir(up_id), file_path)


    def create_tmp_dirs(self, up_id):
        if self.p_object_type == self.OBJ_SPEECH:
            tx_dir = self.__tmp_tx_dir(up_id)
            res =  self.__make_dir(tx_dir)
            if not(res):
                return {'code':10044, 'info':'Cannot create transcription temporal dir. Please contact system administrator.'}
            align_dir = self.__tmp_align_dir(up_id)
            res =  self.__make_dir(align_dir)
            if not(res):
                return {'code':10045, 'info':'Cannot create alignment temporal dir. Please contact system administrator.'}
            tl_dir = self.__tmp_tl_dir(up_id)
            res =  self.__make_dir(tl_dir)
            if not(res):
                return {'code':10045, 'info':'Cannot create translation temporal dir. Please contact system administrator.'}
            media_dir = self.__tmp_media_dir(up_id)
            res =  self.__make_dir(media_dir)
            if not(res):
                return {'code':10046, 'info':'Cannot create media temporal dir. Please contact system administrator.'}
            tts_dir = self.__tmp_tts_dir(up_id)
            res =  self.__make_dir(tts_dir)
            if not(res):
                return {'code':10044, 'info':'Cannot create TTS temporal dir. Please contact system administrator.'}
        elif self.p_object_type == self.OBJ_VALIGN:
            align_dir = self.__tmp_align_dir(up_id)
            res =  self.__make_dir(align_dir)
            if not(res):
                return {'code':10045, 'info':'Cannot create alignment temporal dir. Please contact system administrator.'}
        else:
            tl_dir = self.__tmp_tl_dir(up_id)
            res =  self.__make_dir(tl_dir)
            if not(res):
                return {'code':10045, 'info':'Cannot create translation temporal dir. Please contact system administrator.'}
        return {'code':0}


    def create_tmp_tl_pair_dir(self, up_id, src_lang, trg_lang):
        tl_pair_dir = self.tmp_tl_pair_dir(up_id, src_lang, trg_lang)
        res =  self.__make_dir(tl_pair_dir)
        if not(res):
            return {'code':10047, 'info':'Cannot create translation temporal dir. Please contact system administrator.'}
        return {'code':0}
 
    def create_tmp_tl_pair_object_dir(self, up_id, src_lang, trg_lang, obj_hash):
        tl_pair_dir = self.tmp_tl_pair_object_dir(up_id, src_lang, trg_lang, obj_hash)
        res =  self.__make_dir(tl_pair_dir)
        if not(res):
            return {'code':10047, 'info':'Cannot create translation temporal dir. Please contact system administrator.'}
        return {'code':0}
 

    def create_tmp_tts_lang_dir(self, up_id, lang):
        tts_lang_dir = self.tmp_tts_lang_dir(up_id, lang)
        res =  self.__make_dir(tts_lang_dir)
        if not(res):
            return {'code':10047, 'info':'Cannot create TTS lang temporal dir. Please contact system administrator.'}
        return {'code':0}
   
  

    # Insert, update and delete lecture methods

    def get_hash(self, obj_id, user_id):
        h = hashlib.sha256()
        if isinstance(obj_id, unicode):
            obj_id = obj_id.encode('utf-8')
        if user_id != None:
            h.update("%d|%s" % (user_id, obj_id))
        else:
            h.update("%s" % obj_id)
        return h.hexdigest()


    def insert_lecture(self, up_id, metadata, media, slides, attachments, captions, audiotracks, waveform, user_id):
        # Lecture
        try:
            ext_id = metadata["external_id"]
        except KeyError:
            return {'code':10048, 'info':'external_id key not found in metadata when inserting lecture into DB.'}
        try:
            dur = metadata["duration"]
        except KeyError:
            return {'code':10049, 'info':'duration key not found in metadata when inserting lecture into DB.'}
        try:
            lang = metadata["language"]
        except KeyError:
            return {'code':10050, 'info':'language key not found in metadata when inserting lecture into DB.'}
        try:
            speakers = metadata["speakers"]
        except KeyError:
            return {'code':10091, 'info':'"speakers" key not found in metadata when inserting lecture into DB.'}
        lec_hash = self.get_hash(ext_id, user_id)
        res = self.get_lecture_info(lec_hash)
        if res['code'] != 0:
            return res
        if len(res['data']) > 0:
            return {'code':10088, 'info':"External ID '%s' already exists in Database." % ext_id}
        try:
            lec_id = self.__insert_lecture(ext_id, dur, lang, lec_hash, user_id)
        except:
            print_exception()
            self.__rollback_insert(lec_hash, del_dirs=False)
            return {'code':10051, 'info':'Cannot insert new video into DB. Please contact system administrator.'}

        # Lecture Metadata
        title = metadata.get('title', None)
        if title != None: title = title.encode('UTF-8')
        topic = metadata.get('topic', None)
        if topic != None: topic = topic.encode('UTF-8')
        kw = metadata.get('keywords', None)
        if kw != None: kw = kw.encode('UTF-8')
        date = metadata.get('date', None)
        try:
            self.__insert_metadata(lec_id, title, topic, kw, date)
        except:
            print_exception()
            self.__rollback_insert(lec_hash, del_dirs=False)
            return {'code':10052, 'info':'Cannot insert lecture\'s metadata into DB. Please contact system administrator.'}

        ## Users-Videos
        #if user_id != None:
        #    try:
        #        self.__insert_user_lecture(user_id, lec_id)
        #    except:
        #        print_exception()
        #        self.__rollback_insert(lec_hash, del_dirs=False)
        #        return {'code':10094, 'info':'Cannot insert new video-user relation into DB. Please contact system administrator.'}
    
        # create media and trans directory
        res = self.__create_media_dir(lec_hash) 
        if not(res):
            return {'code':10053, 'info':'Cannot create media dir. Please contact system administrator.'}
        res = self.__create_trans_dir(lec_hash) 
        if not(res):
            return {'code':10054, 'info':'Cannot create trans dir. Please contact system administrator.'}

        # media
        res = self.__insert_and_copy_media(media, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=False)
        if res['code'] != 0:
            self.__rollback_insert(lec_hash)
            return res

        # Speaker 
        res = self.__check_and_insert_speaker_lecture(speakers, lec_id, user_id)
        if res['code'] != 0:
            self.__rollback_insert(lec_hash)
            return res

        # Slides
        res = self.__insert_and_copy_slides(slides, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash)
        if res['code'] != 0:
            self.__rollback_insert(lec_hash)
            return res

        # Attachments
        res = self.__insert_and_copy_attachments(attachments, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=False)
        if res['code'] != 0:
            self.__rollback_insert(lec_hash)
            return res

        # captions
        res = self.__insert_and_copy_captions(captions, up_id, ext_id, lec_id, self.object_type['captions'], lec_hash, check_duplicated=False)
        if res['code'] != 0:
            self.__rollback_insert(lec_hash)
            return res

        # audiotracks
        res = self.__insert_and_copy_audiotracks(audiotracks, up_id, ext_id, lec_id, self.object_type['audiotracks'], lec_hash, check_duplicated=False)
        if res['code'] != 0:
            self.__rollback_insert(lec_hash)
            return res

        # waveform
        if waveform != None:
            res = self.__insert_and_copy_waveform(waveform, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=False)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res


        # Set directory and file permissions
        res = self.__set_media_permissions(lec_hash)
        if not(res):
            self.__rollback_insert(lec_hash)
            return {'code':10055, 'info':'Cannot set permissions in media directory. Please contact system administrator.'}
        res = self.__set_trans_permissions(lec_hash)
        if not(res):
            self.__rollback_insert(lec_hash)
            return {'code':10056, 'info':'Cannot set permissions in trans directory. Please contact system administrator.'}
        # Commit changes to DB
        try:
            self.db.commit()
        except:
            print_exception()
            self.__rollback_insert(lec_hash)
            return {'code':10057, 'info':'Cannot commit transaction to DB. Please contact system administrator.'}
        return {'code':0 } 

  
    def update_lecture(self, up_id, metadata, media, slides, attachments, captions, audiotracks, waveform, user_id, regen=False): 
        try:
            ext_id = metadata["external_id"]
        except KeyError:
            return {'code':10058, 'info':'external_id key not found in metadata when inserting lecture into DB.'}
        lec_hash = self.get_hash(ext_id, user_id)
        res = self.get_lecture_info(lec_hash)
        if res['code'] != 0:
            return res
        if len(res['data']) == 0:
            return {'code':10059, 'info':'External ID \'%s\' does not exist in DB.' % ext_id}
        lec_id = res['data'][0]['id']

        res = self.__backup(lec_hash)
        if not(res):
            return {'code':10060, 'info':'Cannot backup directories. Please contact system administrator.'}

        # Update existing lecture with new data
        if media == None:
            if not(regen): # If it is not a regeneration of subtitles.
                # Lecture Metadata

                title = metadata.get('title', None)
                if title != None: title = title.encode('UTF-8')
                topic = metadata.get('topic', None)
                if topic != None: topic = topic.encode('UTF-8')
                kw = metadata.get('keywords', None)
                if kw != None: kw = kw.encode('UTF-8')
                date = metadata.get('date', None)
                try:
                    self.__update_metadata(lec_id, title, topic, kw, date)
                except:
                    print_exception()
                    self.__rollback_update(lec_hash, del_dirs=False)
                    return {'code':10061, 'info':'Cannot update lecture\'s metadata into DB. Please contact system administrator.'}

            # Slides
            res = self.__insert_and_copy_slides(slides, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=True)
            if res['code'] != 0:
                self.__rollback_update(lec_hash, del_dirs=False)
                return res

            # Attachments
            res = self.__insert_and_copy_attachments(attachments, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=False)
            if res['code'] != 0:
                self.__rollback_update(lec_hash, del_dirs=False)
                return res

            # captions
            res = self.__insert_and_copy_captions(captions, up_id, ext_id, lec_id, self.object_type['captions'], lec_hash, check_duplicated=True)
            if res['code'] != 0:
                self.__rollback_update(lec_hash, del_dirs=False)
                return res

            # audiotracks
            res = self.__insert_and_copy_audiotracks(audiotracks, up_id, ext_id, lec_id, self.object_type['audiotracks'], lec_hash, check_duplicated=True)
            if res['code'] != 0:
                self.__rollback_update(lec_hash)
                return res

            # waveform
            if waveform != None:
                res = self.__insert_and_copy_waveform(waveform, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=True)
                if res['code'] != 0:
                    self.__rollback_update(lec_hash)
                    return res

            # Set directory and file permissions
            res = self.__set_media_permissions(lec_hash)
            if not(res):
                self.__rollback_update(lec_hash)
                return {'code':10062, 'info':'Cannot set permissions in media directory. Please contact system administrator.'}
            res = self.__set_trans_permissions(lec_hash)
            if not(res):
                self.__rollback_update(lec_hash)
                return {'code':10063, 'info':'Cannot set permissions in trans directory. Please contact system administrator.'}

        # Re-recording
        else:
            try:
                ext_id = metadata["external_id"]
            except KeyError:
                return {'code':10048, 'info':'external_id key not found in metadata when updating lecture into DB.'}
            try:
                dur = metadata["duration"]
            except KeyError:
                return {'code':10049, 'info':'duration key not found in metadata when updating lecture into DB.'}
            try:
                lang = metadata["language"]
            except KeyError:
                return {'code':10050, 'info':'language key not found in metadata when updating lecture into DB.'}
            try:
                speakers = metadata["speakers"]
            except KeyError:
                return {'code':10091, 'info':'"speakers" key not found in metadata when updating lecture into DB.'}

            # Get lecture ID
            lec_hash = self.get_hash(ext_id, user_id)
            res = self.get_lecture_info(lec_hash)
            if res['code'] != 0:
                return res
            if len(res['data']) == 0:
                return {'code':10088, 'info':"External ID '%s' does not exist in Database." % ext_id}
            lec_id = res['data'][0]['id']
            
            # Update lecture
            try:
                self.__update_lecture(lec_id, dur, lang)
            except:
                print_exception()
                self.__rollback_update(lec_hash, del_dirs=False)
                return {'code':10095, 'info':'Cannot update lecture in DB. Please contact system administrator.'}
    
            # Update Metadata
            title = metadata.get('title', None)
            if title != None: title = title.encode('UTF-8')
            topic = metadata.get('topic', None)
            if topic != None: topic = topic.encode('UTF-8')
            kw = metadata.get('keywords', None)
            if kw != None: kw = kw.encode('UTF-8')
            date = metadata.get('date', None)
            try:
                self.__update_metadata(lec_id, title, topic, kw, date)
            except:
                print_exception()
                self.__rollback_update(lec_hash, del_dirs=False)
                return {'code':10096, 'info':'Cannot update lecture\'s metadata into DB. Please contact system administrator.'}
    
            # Delete captions, speakers_videos, and files
            try:
                self.__delete_for_update(lec_id)
            except:
                print_exception()
                self.__rollback_update(lec_hash, del_dirs=False)
                return {'code':10097, 'info':'Cannot delete info from DB prior update operation. Please contact system administrator.'}
     
            # Backup old data dirs
            res = self.__mv_media2updated(lec_hash)
            if not(res):
                self.__rollback_update(lec_hash)
                return {'code':10065, 'info':'Cannot move existing media directory.'}
            res = self.__mv_trans2updated(lec_hash)
            if not(res):
                self.__rollback_update(lec_hash)
                return {'code':10066, 'info':'Cannot move existing trans directory.'}
    
            # create media and trans directory
            res = self.__create_media_dir(lec_hash) 
            if not(res):
                return {'code':10053, 'info':'Cannot create media dir. Please contact system administrator.'}
            res = self.__create_trans_dir(lec_hash) 
            if not(res):
                return {'code':10054, 'info':'Cannot create trans dir. Please contact system administrator.'}
    
            # media
            res = self.__insert_and_copy_media(media, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=False)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res
    
            # Speaker 
            res = self.__check_and_insert_speaker_lecture(speakers, lec_id, user_id)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res
    
            # Slides
            res = self.__insert_and_copy_slides(slides, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res
    
            # Attachments
            res = self.__insert_and_copy_attachments(attachments, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=False)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res
    
            # captions
            res = self.__insert_and_copy_captions(captions, up_id, ext_id, lec_id, self.object_type['captions'], lec_hash, check_duplicated=False)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res

            # audiotracks
            res = self.__insert_and_copy_audiotracks(audiotracks, up_id, ext_id, lec_id, self.object_type['audiotracks'], lec_hash, check_duplicated=False)
            if res['code'] != 0:
                self.__rollback_insert(lec_hash)
                return res
    
            # waveform
            if waveform != None:
                res = self.__insert_and_copy_waveform(waveform, up_id, ext_id, lec_id, self.object_type['videos'], lec_hash, check_duplicated=True)
                if res['code'] != 0:
                    self.__rollback_insert(lec_hash)
                    return res

            # Set directory and file permissions
            res = self.__set_media_permissions(lec_hash)
            if not(res):
                self.__rollback_insert(lec_hash)
                return {'code':10055, 'info':'Cannot set permissions in media directory. Please contact system administrator.'}
            res = self.__set_trans_permissions(lec_hash)
            if not(res):
                self.__rollback_insert(lec_hash)
                return {'code':10056, 'info':'Cannot set permissions in trans directory. Please contact system administrator.'}
            # Commit changes to DB
            try:
                self.db.commit()
            except:
                print_exception()
                self.__rollback_insert(lec_hash)
                return {'code':10057, 'info':'Cannot commit transaction to DB. Please contact system administrator.'}
            return {'code':0 } 
        
        # Remove backup
        res = self.__rm_backup(lec_hash)
        if not(res):
            return {'code':10068, 'info':'Cannot remove backup directories. Please contact system administrator.'}

        return {'code':0}


    def delete_lecture(self, ext_id, lec_hash):
        res = self.get_lecture_info(lec_hash)
        if res['code'] != 0:
            return res
        if len(res['data']) == 0:
            return {'code':10069, 'info':'External ID \'%s\' does not exist in DB.' % ext_id}
        lec_id = res['data'][0]['id']

        #res = self.__backup(lec_hash)
        #if not(res):
        #    return {'code':10070, 'info':'Cannot backup directories. Please contact system administrator.'}

        try:
        #    self.__delete_lecture_db(lec_id)
            self.__mark_lecture_deleted(lec_id)
        except:
            print_exception()
        #    self.__rollback_delete(lec_hash, del_dirs=False)
            return {'code':10071, 'info':'Cannot delete lecture in DB. Please contact system administrator.'}
        #res = self.__mv_media2removed(lec_hash)
        #if not(res):
        #    self.__rollback_delete(lec_hash)
        #    return {'code':10072, 'info':'Cannot move media directory. Please contact system administrator.'}
        #res = self.__mv_trans2removed(lec_hash)
        #if not(res):
        #    self.__rollback_delete(lec_hash)
        #    return {'code':10073, 'info':'Cannot move trans directory. Please contact system administrator.'}
        try:
            self.db.commit()
        except:
            print_exception()
        #    self.__rollback_delete(lec_hash)
            return {'code':10074, 'info':'Cannot commit transaction to DB. Please contact system administrator.'}

        #res = self.__rm_backup(lec_hash)
        #if not(res):
        #    return {'code':10075, 'info':'Cannot remove backup directories. Please contact system administrator.'}

        return {'code':0 } 

    def hard_delete_lecture(self, ext_id, lec_hash):
        res = self.get_lecture_info(lec_hash)
        if res['code'] != 0:
            return res
        if len(res['data']) == 0:
            return {'code':10069, 'info':'External ID \'%s\' does not exist in DB.' % ext_id}
        lec_id = res['data'][0]['id']

        res = self.__backup(lec_hash)
        if not(res):
            return {'code':10070, 'info':'Cannot backup directories. Please contact system administrator.'}

        try:
            self.__hard_delete_lecture_db(lec_id)
        except:
            print_exception()
            self.__rollback_delete(ext_id, del_dirs=False)
            return {'code':10071, 'info':'Cannot hard delete lecture in DB. Please contact system administrator.'}
        res = self.__rm_media_dir(lec_hash)
        if not(res):
            self.__rollback_delete(lec_hash)
            return {'code':10072, 'info':'Cannot move media directory. Please contact system administrator.'}
        res = self.__rm_trans_dir(lec_hash)
        if not(res):
            self.__rollback_delete(lec_hash)
            return {'code':10073, 'info':'Cannot move trans directory. Please contact system administrator.'}
        try:
            self.db.commit()
        except:
            print_exception()
            self.__rollback_delete(lec_hash)
            return {'code':10074, 'info':'Cannot commit transaction to DB. Please contact system administrator.'}

        res = self.__rm_backup(lec_hash)
        if not(res):
            return {'code':10075, 'info':'Cannot remove backup directories. Please contact system administrator.'}

        return {'code':0 } 




    
    # Uploads management


    UPLOADS_KEYS = [ 'id', 'hash', 'op_code', 'external_id', 'language', 'user_id', 'email', 'ts', 'up_status', 'info', 'error_code', 'jobs', 'last_ts', 'session_id', 'manifest', 'control_data', 'cancel_request', 'debug' ]

    def get_uploads(self):
        """ Returns list of all uploads. """
        # Get list of uploads
        try:
            self.cur.execute("""SELECT id, hash, op_code, external_id, language, user_id, email, timestamp, status_id, info, error_code,
                                       cluster_job_ids, last_status_update, session_id, manifest, control_data, cancel_request, debug
                                FROM uploads
                                WHERE object_type = %s
                                  AND api_version = %s
                                ORDER BY timestamp ASC""", (self.p_object_type, self.API_VERSION))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10076, 'info':'Cannot get uploads list'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.UPLOADS_KEYS)
        return {'code':0, 'data':data}


    def get_upload_by_hash(self, lec_hash):
        try:
            self.cur.execute("""SELECT id, hash, op_code, external_id, language, user_id, email, timestamp, status_id, info, error_code,
                                       cluster_job_ids, last_status_update, session_id, manifest, control_data
                                FROM uploads
                                WHERE hash = %s
                                      AND object_type = %s""", (lec_hash, self.p_object_type))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10090, 'info':'Cannot get uploads info'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.UPLOADS_KEYS)
        return {'code':0, 'data':data}


    def get_upload_by_id(self, up_id):
        try:
            self.cur.execute("""SELECT id, hash, op_code, external_id, language, user_id, email, timestamp, status_id, info, error_code,
                                       cluster_job_ids, last_status_update, session_id, manifest, control_data
                                FROM uploads
                                WHERE id = %s
                                      AND object_type = %s""", (up_id, self.p_object_type))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10090, 'info':'Cannot get uploads info'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.UPLOADS_KEYS)
        return {'code':0, 'data':data}


    def delete_upload(self, up_id, up_dir):
        """ Deletes upload from both DB and disk """
        try:
            self.cur.execute("""DELETE FROM uploads
                                WHERE id = %s""",(up_id,))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10077, 'info':'Cannot delete upload entry of DB.'}
        try:
            shutil.rmtree(up_dir)
        except OSError:
            return {'code':10078, 'info':'Cannot delete upload directory from disk.'}
        return {'code':0}


    def update_upload(self, up_id, external_id, lang, user_id, status_code, info, ts, jobs=None, error_code=None, session_id=None, eta=None):
        """ Updates upload status """
        query = """UPDATE uploads SET status_id = %s, external_id = %s, language = %s, user_id = %s, last_status_update = %s, info = %s, error_code = %s"""
        values = [status_code, external_id, lang, user_id, ts, info, error_code]

        if jobs != None:
            query += ", cluster_job_ids = %s"
            values.append(jobs)
        if session_id != None:
            query += ", session_id = %s"
            values.append(session_id)
        if eta != None:
            query += ", eta = %s"
            values.append(eta)
        query += " WHERE id = %s"
        values.append(up_id)
        tvalues = tuple(values)
        try:
            self.cur.execute(query, tvalues)
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10080, 'info':'Cannot update uploads table'} 
        return {'code':0}


    def get_user_quota(self, user_id):
        try:
            self.cur.execute("""SELECT quota_time, quota_videos, accum_time, accum_videos
                                FROM users_quota
                                WHERE user_id = %s""", (user_id,))
            row = self.cur.fetchone()
        except:
            print_exception()
            return {'code':10092, 'info':'Cannot get user quota.'}
        data = { 'seconds':row[0], 'num_videos':row[1], 'acc_time':row[2], 'acc_videos':row[3] }
        return {'code':0, 'data':data}


    def update_user_quota(self, up_id, user_id, duration):
        try:
            self.cur.execute("""UPDATE users_quota
                                 SET quota_time = quota_time - %s, quota_videos = quota_videos - 1, 
                                     accum_time = accum_time + %s, accum_videos = accum_videos + 1
                                 WHERE user_id = %s""",(duration, duration, user_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10093, 'info':'Cannot update user quota.'}
        try:
            self.cur.execute("""UPDATE uploads
                                 SET consumed_quota = %s
                                 WHERE id = %s""",(duration, up_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10101, 'info':'Cannot update consumed quota on uploads table.'}
        return {'code':0}


    def restore_user_quota(self, up_id, user_id):
        try:
            self.cur.execute("""SELECT consumed_quota
                                FROM uploads
                                WHERE id = %s""", (up_id,))
            row = self.cur.fetchone()
        except:
            print_exception()
            return {'code':10102, 'info':'Cannot get consumed user quota.'}
        duration = row[0]
        if duration != None:
            try:
                self.cur.execute("""UPDATE users_quota
                                     SET quota_time = quota_time + %s, quota_videos = quota_videos + 1, 
                                         accum_time = accum_time - %s, accum_videos = accum_videos - 1
                                     WHERE user_id = %s""",(duration, duration, user_id))
                self.db.commit()
            except:
                print_exception()
                self.db.rollback()
                return {'code':10093, 'info':'Cannot restore user quota.'}
            try:
                self.cur.execute("""UPDATE uploads
                                     SET consumed_quota = NULL
                                     WHERE id = %s""",(up_id,))
                self.db.commit()
            except:
                print_exception()
                self.db.rollback()
                return {'code':10101, 'info':'Cannot update consumed quota on uploads table.'}
        return {'code':0}


    def start_session(self, lec_hash, user_id, author_id, author_conf, author_name=None):
        try:
            self.cur.execute("""SELECT id
                                FROM mod_sessions 
                                WHERE video_id = (SELECT id FROM videos WHERE hash = %s) AND ended_at IS NULL""", (lec_hash,))
            rows = self.cur.fetchall()
        except:
            return {'code':10172, 'info':'Cannot check edit sessions.'}
 
        if len(rows) > 0:
            session_id = rows[0][0]
            return {'code':10173, 'info':'Cannot start updating media: there exists an ongoing user edit session.', 'data':session_id}

        now = datetime.now()
        try:
            self.cur.execute("""INSERT INTO mod_sessions 
                 (video_id, user_id, author_id, author_conf, author_name, author_type, started_at, last_update, last_alive, media_duration) 
                 VALUES 
                 ((select id from videos where hash = %s), %s, %s,%s,%s, %s, %s, %s, %s, (select duration from videos where hash = %s))
                 RETURNING id""",
               (lec_hash, user_id, author_id, author_conf, author_name, "computer", now, now, now, lec_hash))
            sid = self.cur.fetchone()[0]
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10174, 'info':'Cannot start session.'}
        return {'code':0, 'data':sid}

    def update_session(self, session_id):
        try:
            now = datetime.now()
            self.cur.execute("""UPDATE mod_sessions 
                 SET last_update = %s, last_alive = %s
                 WHERE id = %s""",
                (now, now, session_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10177, 'info':'Cannot update session.'}
        return {'code':0}


    def end_session(self, session_id, ended_by_id, ended_by_name=None):
        try:
            now = datetime.now()
            self.cur.execute("""UPDATE mod_sessions 
                 SET ended_at = %s, ended_by_id = %s, ended_by_name = %s, ended_by_type = 'computer', requires_revision = False, requires_postprocess = False, do_regenerate = False
                 WHERE id = %s""",
                (now, ended_by_id, ended_by_name, session_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10175, 'info':'Cannot end session.'}
        return {'code':0}

    def delete_session(self, session_id):
        try:
            self.cur.execute("""DELETE FROM mod_sessions WHERE id = %s""", (session_id,))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10176, 'info':'Cannot delete session.'}
        return {'code':0}




#####################################################################################
#####################################################################################
#                                  DOCUMENTS                                        #
#####################################################################################
#####################################################################################

    DOCUMENTS_KEYS = [ 'id', 'external_id', 'title', 'language', 'sentences', 'words', 'chars', 'format', 'owner_id', 'hash' ] 

    def get_documents(self):
        """ Returns list of all documents. """
        # Get list of uploads
        try:
            self.cur.execute("""SELECT id, external_id, title, language, sentences, words, chars, format, owner_id, hash
                                FROM documents""")
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10222, 'info':'Cannot get documents list'}
        # Create List of Dicts and return Data object with the results
        data = self.__rows2data__(rows, Storage.DOCUMENTS_KEYS)
        return {'code':0, 'data':data}

    def get_src_document(self, doc_hash, file_type):
        mp = self.trans_mp
        self.cur.execute("""SELECT f.file_path
                                FROM files f 
                                INNER JOIN documents d ON f.object_id = d.id 
                                WHERE d.hash = %s 
                                AND f.object_type_id = (select id from object_types where table_name = 'documents') 
                                AND f.file_type_id = (select id from file_types where name = %s)""",(doc_hash, file_type))
        rows = self.cur.fetchall()
        if len(rows) == 0:
            return {'code':10223, 'info':'Cannot get source document file'}
        elif len(rows) > 1:
            return {'code':10224, 'info':'Retrieved more than one document file'}
        rfp = rows[0][0]
        fp = '%s/%s' %(mp,rfp)
        return {'code':0, 'data':fp}

    def get_tl_document(self, tid, file_type):
        mp = self.trans_mp
        self.cur.execute("""SELECT f.file_path
                                FROM files f 
                                INNER JOIN translations t ON f.object_id = t.id 
                                WHERE t.id = %s 
                                AND f.object_type_id = (select id from object_types where table_name = 'translations') 
                                AND f.file_type_id = (select id from file_types where name = %s)""",(tid, file_type))
        rows = self.cur.fetchall()
        if len(rows) == 0:
            return {'code':10225, 'info':'Cannot get translated document file'}
        elif len(rows) > 1:
            return {'code':10226, 'info':'Retrieved more than one document file'}
        rfp = rows[0][0]
        fp = '%s/%s' %(mp,rfp)
        return {'code':0, 'data':fp}



    def __document_trans_dir_abspath(self, obj_hash):
        return "%s/%s/%s" % (self.trans_mp, obj_hash[0], obj_hash)

    def __rfpath(self, obj_hash, fp):
        return "%s/%s/%s" % (obj_hash[0], obj_hash, os.path.basename(fp))

    def __ffpath(self, mp, obj_hash, fp):
        return "%s/%s" % (mp, self.__rfpath(obj_hash, fp))
   
    def __copy_file(self, sfp, tfp):
        shutil.copy(sfp, tfp)

    def __insert_file__new(self, obj_id, obj_type_id, file_type, file_format, path, is_url=False):
        self.cur.execute("""INSERT INTO files (object_id, object_type_id, file_type_id, file_format, file_path, is_url) 
                                   VALUES (%s, (select id from object_types where table_name = %s), (select id from file_types where name = %s), %s, %s, %s) 
                                   RETURNING id""", (obj_id, obj_type_id, file_type, file_format, path, is_url))
        file_id = self.cur.fetchone()[0]
        return file_id

    def insert_document(self, obj_hash, obj_id, doc_lang, title, fformat, sents, words, chars, src_orig_file, src_html_file, src_dtlx_file, owner_id):
        date = datetime.now()
        try:
            self._transaction_add_document(obj_hash)
            self.__create_trans_dir(obj_hash) 

            # Insert Document object
            self.cur.execute("""INSERT INTO documents (external_id, language, sentences, words, chars, format, title, date, owner_id, hash) 
                                VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s) RETURNING id""",
                                (obj_id, doc_lang, sents, words, chars, fformat, title, date, owner_id, obj_hash))
            doc_id = self.cur.fetchone()[0]
            # Copy and insert original Document
            orfp = self.__rfpath(obj_hash, src_orig_file)
            offp = self.__ffpath(self.trans_mp, obj_hash, src_orig_file)
            self.__copy_file(src_orig_file, offp)
            self.__insert_file__new(doc_id, Storage.OBJ_TEXT, Storage.FT_DOC_ORIGINAL, fformat, orfp, False)
            # Copy and insert HTML Document
            prfp = self.__rfpath(obj_hash, src_html_file)
            pffp = self.__ffpath(self.trans_mp, obj_hash, src_html_file)
            self.__copy_file(src_html_file, pffp)
            self.__insert_file__new(doc_id, Storage.OBJ_TEXT, Storage.FT_DOC_PARSED, "html", prfp, False)
            # Copy and insert DTLX Document
            irfp = self.__rfpath(obj_hash, src_dtlx_file)
            iffp = self.__ffpath(self.trans_mp, obj_hash, src_dtlx_file)
            self.__copy_file(src_dtlx_file, iffp)
            self.__insert_file__new(doc_id, Storage.OBJ_TEXT, Storage.FT_DOC_INTERNAL, "xml", irfp, False)

            self.__set_trans_permissions(obj_hash)
        except:
            print_exception()
            self._restore()
            return {'code':100000, 'info':'Cannot insert document.'} 
        return {'code':0}


    def insert_translation(self, obj_hash, lang, status, sup_ratio, from_lang, system_id, fformat, tl_orig_file, tl_html_file, tl_dtlx_file):
        try:
            self._transaction_add_document(obj_hash)

            # Insert Translation object
            self.cur.execute("""INSERT INTO translations (document_id, language, status, sup_ratio, from_language, system_id) 
                                VALUES ((SELECT id FROM documents WHERE hash = %s), %s, %s, %s, %s, %s) RETURNING id""",
                                (obj_hash, lang, status, sup_ratio, from_lang, system_id))
            tid = self.cur.fetchone()[0]
            # Copy and insert original Document
            orfp = self.__rfpath(obj_hash, tl_orig_file)
            offp = self.__ffpath(self.trans_mp, obj_hash, tl_orig_file)
            self.__copy_file(tl_orig_file, offp)
            self.__insert_file__new(tid, Storage.OBJ_TEXT_TL, Storage.FT_DOC_ORIGINAL, fformat, orfp, False)
            # Copy and insert HTML Document
            prfp = self.__rfpath(obj_hash, tl_html_file)
            pffp = self.__ffpath(self.trans_mp, obj_hash, tl_html_file)
            self.__copy_file(tl_html_file, pffp)
            self.__insert_file__new(tid, Storage.OBJ_TEXT_TL, Storage.FT_DOC_PARSED, "html", prfp, False)
            # Copy and insert DTLX Document
            irfp = self.__rfpath(obj_hash, tl_dtlx_file)
            iffp = self.__ffpath(self.trans_mp, obj_hash, tl_dtlx_file)
            self.__copy_file(tl_dtlx_file, iffp)
            self.__insert_file__new(tid, Storage.OBJ_TEXT_TL, Storage.FT_DOC_INTERNAL, "xml", irfp, False)

            self.__set_trans_permissions(obj_hash)
        except:
            print_exception()
            self._restore()
            return {'code':100020, 'info':'Cannot insert translation.'} 
        return {'code':0}


    def update_document(self, obj_hash, title, fformat, sents, words, chars, src_orig_file, src_html_file, src_dtlx_file):
        date = datetime.now()
        try:
            self._transaction_add_document(obj_hash)

            # Update Document object
            self.cur.execute("""UPDATE documents 
                                  SET (sentences, words, chars, format, title, date) 
                                    = (%s, %s, %s, %s, COALESCE(%s, title), %s) 
                                  WHERE hash = %s
                                  RETURNING id""",
                                (sents, words, chars, fformat, title, date, obj_hash))
            doc_id = self.cur.fetchone()[0]
            # Delete previous original document (file extension can be different)
            res = self.get_src_document(obj_hash, Storage.FT_DOC_ORIGINAL)
            if res['code'] != 0:
                return res
            fp = res['data']
            os.remove(fp)
            # Copy updated original Document
            offp = self.__ffpath(self.trans_mp, obj_hash, src_orig_file)
            self.__copy_file(src_orig_file, offp)
            # Overwrite HTML Document
            pffp = self.__ffpath(self.trans_mp, obj_hash, src_html_file)
            self.__copy_file(src_html_file, pffp)
            # Overwrite DTLX Document
            iffp = self.__ffpath(self.trans_mp, obj_hash, src_dtlx_file)
            self.__copy_file(src_dtlx_file, iffp)

            self.__set_trans_permissions(obj_hash)
        except:
            print_exception()
            self._restore()
            return {'code':100010, 'info':'Cannot update document.'} 
        return {'code':0}


    def update_translation(self, obj_hash, tid, status, sup_ratio, from_lang, system_id, fformat, tl_orig_file, tl_html_file, tl_dtlx_file):
        try:
            self._transaction_add_document(obj_hash)

            # Update Document object
            self.cur.execute("""UPDATE translations 
                                SET (status, sup_ratio, from_language, system_id) 
                                  = (%s, %s, %s, %s)
                                WHERE id = %s""",
                                (status, sup_ratio, from_lang, system_id, tid))
            # Delete previous original document (file extension can be different)
            res = self.get_tl_document(tid, Storage.FT_DOC_ORIGINAL)
            if res['code'] != 0:
                return res
            fp = res['data']
            os.remove(fp)
            # Overwrite original Document
            offp = self.__ffpath(self.trans_mp, obj_hash, tl_orig_file)
            self.__copy_file(tl_orig_file, offp)
            # Overwrite HTML Document
            pffp = self.__ffpath(self.trans_mp, obj_hash, tl_html_file)
            self.__copy_file(tl_html_file, pffp)
            # Overwrite DTLX Document
            iffp = self.__ffpath(self.trans_mp, obj_hash, tl_dtlx_file)
            self.__copy_file(tl_dtlx_file, iffp)

            self.__set_trans_permissions(obj_hash)
        except:
            print_exception()
            self._restore()
            return {'code':100030, 'info':'Cannot update translation.'} 
        return {'code':0}


    def __hard_delete_document_db(self, obj_int_id):
        self.cur.execute("""DELETE FROM files WHERE object_id = %s AND object_type_id = (select id from object_types where table_name = %s)""", (obj_int_id, Storage.OBJ_TEXT))
        self.cur.execute("""DELETE FROM files WHERE object_id IN (SELECT id FROM translations WHERE document_id = %s) AND object_type_id = (select id from object_types where table_name = %s)""", (obj_int_id, Storage.OBJ_TEXT_TL))
        self.cur.execute("""DELETE FROM mods WHERE mod_session_id IN (select id from mod_sessions where object_id = %s and object_type = %s)""", (obj_int_id, Storage.OBJ_TEXT))
        self.cur.execute("""DELETE FROM mod_sessions WHERE object_id = %s and object_type = %s""", (obj_int_id, Storage.OBJ_TEXT))
        self.cur.execute("""DELETE FROM translations WHERE document_id = %s""", (obj_int_id,))
        self.cur.execute("""DELETE FROM users_documents WHERE document_id = %s""", (obj_int_id,))
        self.cur.execute("""DELETE FROM groups_documents WHERE document_id = %s""", (obj_int_id,))
        self.cur.execute("""DELETE FROM documents WHERE id = %s""", (obj_int_id,))

    def delete_document(self, obj_int_id, obj_hash):
        try:
            self._transaction_add_document(obj_hash)
            self.__hard_delete_document_db(obj_int_id)
            assert(self.__rm_trans_dir(obj_hash))
        except:
            print_exception()
            self._restore()
            return {'code':100100, 'info':'Could not delete document.'}

        return {'code':0 } 


    def begin_transaction(self, operation):
        if self._transaction_on:
            raise Exception("Already under transaction ID %s: must call end_transaction() function before." % self._transaction_ts)
        self._transaction_on = True
        self._transaction_op = operation
        self._transaction_ids = []
        self._transaction_ts = str(int(time.time()))
        #print "BEGIN TRANSACTION %s" % self._transaction_ts # FIXME
	return {'code':0}

    def _transaction_add_document(self, obj_hash):
        if not(self._transaction_on):
            raise Exception("Not under transaction: must call begin_transaction() function before.")
        if obj_hash not in (self._transaction_ids):
            self._transaction_ids.append(obj_hash)
            if self._transaction_op != Storage.OP_CODE_NEW: 
                self._backup_document(obj_hash)
            #print "ADDED DOCUMENT TO TRANSACTION: %s ((All: %s ))" % (obj_hash, str(self._transaction_ids)) # FIXME

    def _backup_document(self, obj_hash):
        src_trans_dir = self.__document_trans_dir_abspath(obj_hash)
        trg_trans_dir = self.__document_trans_dir_abspath(obj_hash)+'.bak.%s' % self._transaction_ts
        if not(os.path.exists(trg_trans_dir)):
            shutil.copytree(src_trans_dir, trg_trans_dir)
            #print "DIRECTORY BACKUPED: %s -> %s" % (src_trans_dir, trg_trans_dir) # FIXME

    def _restore_backup_document(self, obj_hash):
        src_trans_dir = self.__document_trans_dir_abspath(obj_hash)+'.bak.%s' % self._transaction_ts
        trg_trans_dir = self.__document_trans_dir_abspath(obj_hash)
        #print "RESTORING BACKUP: %s -> %s" % (src_trans_dir, trg_trans_dir) # FIXME
        self.__rm_dir(trg_trans_dir)
        shutil.copytree(src_trans_dir, trg_trans_dir)
        self.__rm_dir(src_trans_dir)

    def _rm_backup_document(self, obj_hash):
        trans_dir = self.__document_trans_dir_abspath(obj_hash)+'.bak.%s' % self._transaction_ts
        #print "REMOVING BACKUP: %s" % (trans_dir) # FIXME
        self.__rm_dir(trans_dir)

    def _restore(self):
        if not(self._transaction_on):
            raise Exception("Not under transaction: must call begin_transaction() function before.")
        #print "RESTORING TRANSACTION ((All: %s ))" % (str(self._transaction_ids)) # FIXME
        for obj_hash in self._transaction_ids:
             if self._transaction_op == Storage.OP_CODE_NEW:
                 self.__rm_trans_dir(obj_hash)
             else:
                 self._restore_backup_document(obj_hash)
        self.db.rollback()
        self._transaction_on = False
        self._transaction_ids = []
        self._transaction_op = None
        self._transaction_ts = None

    def end_transaction(self):
        if not(self._transaction_on):
            raise Exception("Not under transaction: must call begin_transaction() function before.")
        self.db.commit()
        if self._transaction_op != Storage.OP_CODE_NEW:
            for obj_hash in self._transaction_ids:
                 self._rm_backup_document(obj_hash)
        self._transaction_on = False
        self._transaction_ids = []
        self._transaction_op = None
        self._transaction_ts = None
	return {'code':0}

    QUOTA_DOCUMENTS_KEYS = [ 'q_docs', 'q_sents', 'q_words', 'a_docs', 'a_sents', 'a_words' ]

    def get_user_quota_documents(self, user_id):
        try:
            self.cur.execute("""SELECT quota_documents, quota_sentences, quota_words, accum_documents, accum_sentences, accum_words
                                FROM users_quota
                                WHERE user_id = %s""", (user_id,))
            rows = self.cur.fetchall()
        except:
            print_exception()
            return {'code':10092, 'info':'Cannot get user quota.'}
        data = self.__rows2data__(rows, Storage.QUOTA_DOCUMENTS_KEYS)[0]
        return {'code':0, 'data':data}


    def update_user_quota_documents(self, user_id, documents, sentences, words):
        try:
            #self.cur.execute("""UPDATE users_quota
            #                     SET quota_documents = quota_documents - %s, 
            #                         quota_sentences = quota_sentences - %s, 
            #                         quota_words = quota_words - %s 
            #                     WHERE user_id = %s AND quota_documents IS NOT NULL""",(documents, sentences, words, user_id))
            self.cur.execute("""UPDATE users_quota
                                 SET quota_documents = quota_documents - %s, 
                                     quota_words = quota_words - %s 
                                 WHERE user_id = %s AND quota_documents IS NOT NULL""",(documents, words, user_id))
            self.cur.execute("""UPDATE users_quota
                                 SET accum_documents = accum_documents + %s, 
                                     accum_sentences = accum_sentences + %s, 
                                     accum_words = accum_words + %s
                                 WHERE user_id = %s""" ,(documents, sentences, words, user_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10093, 'info':'Cannot update user quota.'}
        return {'code':0}

    def restore_user_quota_documents(self, user_id, documents, sentences, words):
        try:
            #self.cur.execute("""UPDATE users_quota
            #                     SET quota_documents = quota_documents + %s, 
            #                         quota_sentences = quota_sentences + %s, 
            #                         quota_words = quota_words + %s 
            #                     WHERE user_id = %s AND quota_documents IS NOT NULL""",(documents, sentences, words, user_id))
            self.cur.execute("""UPDATE users_quota
                                 SET quota_documents = quota_documents + %s, 
                                     quota_words = quota_words + %s 
                                 WHERE user_id = %s AND quota_documents IS NOT NULL""",(documents, words, user_id))
            self.cur.execute("""UPDATE users_quota
                                 SET accum_documents = accum_documents - %s, 
                                     accum_sentences = accum_sentences - %s, 
                                     accum_words = accum_words - %s
                                 WHERE user_id = %s""" ,(documents, sentences, words, user_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':10093, 'info':'Cannot update user quota.'}
        return {'code':0}


    def update_upload__new(self, up_id, status, info, control, error_code=None):
        now = datetime.now()
        control = base64.b64encode(json.dumps(control, encoding='utf-8'))
        try:
            self.cur.execute("""UPDATE uploads 
                                SET status_id = %s, info = %s, control_data = %s,
                                last_status_update = %s, error_code = %s
                                WHERE id = %s""",
                             (status, info, control, now, error_code, up_id))
            self.db.commit()
        except:
            print_exception()
            self.db.rollback()
            return {'code':100080, 'info':'Cannot update uploads table'} 
        return {'code':0}





#####################################################################################
#####################################################################################
#                                  WEBSITES                                         #
#####################################################################################
#####################################################################################

    def __rows2dict__(self, rows, key_pos=0, value_pos=1, null_values=False):
        return { r[key_pos]:r[value_pos] for r in rows if (r[value_pos] != None or null_values) }
    
    def __rows2dict_tuple__(self, rows):
        return { r[0]:tuple(r[1:]) for r in rows }

    WEB_WEBSITE_KEYS = [ 'id', 'name', 'language', 'options' ]

    def get_website_info(self, web_key, user_id):
        self.cur.execute("""SELECT id, name, language, options
                                       FROM websites w
                                       WHERE w.key = %s AND w.owner_id = %s""", (web_key, user_id))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, Storage.WEB_WEBSITE_KEYS)
        return {'code':0, 'data':data}

    def get_website_source_texts(self, web_id, hash_list):
        self.cur.execute("""SELECT hash, id, text
                                       FROM source_texts st
                                       WHERE st.website_id = %s
                                         AND st.hash IN %s""", (web_id, tuple(hash_list)))
        rows = self.cur.fetchall()
        data = self.__rows2dict_tuple__(rows)
        return {'code':0, 'data':data}

    WEB_TLTEXTS_KEYS = [ 'id', 'language', 'status', 'text' ]

    def get_website_translated_texts(self, text_id):
        self.cur.execute("""SELECT id, language, status, text
                                       FROM translated_texts tt
                                       WHERE tt.source_text_id = %s""", (text_id,))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, Storage.WEB_TLTEXTS_KEYS)
        return {'code':0, 'data':data}

    def write_input_text(self, sents, fp):
        with open(fp, 'w') as fd:
            for l in sents:
                fd.write("%s\n" % l.rstrip('\n').encode('utf-8'))

    def read_output_text(self, fp):
        ret = []
        with open(fp, 'r') as fd:
            for l in fd:
                ret.append(l.rstrip('\n').decode('utf-8'))
        return ret

    def insert_website_translation(self, text_id, lang, text):
         self.cur.execute("""INSERT INTO translated_texts 
                                (source_text_id, language, status, last_update, text)
                               VALUES (%s, %s, %s, %s, %s)""", (text_id, lang, self.WEB_TL_STATUS_AUTO, datetime.now(), text))
         self.db.commit()



#####################################################################################
#####################################################################################
#                              VIDEO_ALIGNMENTS                                     #
#####################################################################################
#####################################################################################


    def get_video_alignments(self):
        """ Returns list of all video_alignments. """
        self.cur.execute("""SELECT id, video_id, timestamp, eta, completed_on, 
                                   status_id, info, error_code, input_data, resegment,
                                   control_data, output_data, committed, cancelled, debug
                            FROM video_alignments
                            ORDER BY timestamp ASC""")
        rows = self.cur.fetchall()
        return {'code':0, 'data':rows}

    def delete_video_alignment(self, va_id):
        """ Deletes video_alignment from both DB and disk """
        self.cur.execute("""DELETE FROM video_alignments
                            WHERE id = %s""",(va_id,))
        self.db.commit()
        va_dir = "%s/%s" % (self.uploads_mp, va_id)
        try:
            shutil.rmtree(va_dir)
        except:
            print_exception()
        return {'code':0}


    def update_video_alignment(self, va_id, status, info, control, error_code=None, output=None, completed_on=None, eta=None):
        control = Json(control)
        if output != None:
            output = Json(output)
        self.cur.execute("""UPDATE video_alignments 
                            SET status_id = %s, info = %s, control_data = %s,
                                error_code = %s, output_data = %s, completed_on = %s,
                                eta = %s
                            WHERE id = %s""",
                         (status, info, control, error_code, output, completed_on, eta, va_id))
        self.db.commit()
        return {'code':0}


    def get_video_by_id(self, video_id):
        """ Returns lecture info. """
        self.cur.execute("""SELECT v.id, v.hash, v.external_id, v.language, v.duration, m.removed, m.title, v.owner_id
                            FROM videos v
                            INNER JOIN videos_metadata m ON v.id = m.id
                            WHERE v.id = %s""",(video_id,))
        row = self.cur.fetchone()
        if row == None:
            return {'code':1000, 'info':'Video ID %d not found' % video_id}
        return {'code':0, 'data':dict(row)}

    def _get_align_text_file_path(self, va_id):
        return "%s/text-align.lst" % self.__tmp_align_dir(va_id)

    def save_align_text_file(self, va_id, data):
        file_path = self._get_align_text_file_path(va_id)
        with open(file_path, 'w') as fd:
            fd.write(data.encode('utf-8'))
        return {'code':0, 'data':file_path}
   
    def load_output_json_file(self, json_fp):
        if not(os.path.exists(json_fp)):
            return {'code':1001, 'info':'Output JSON alignment file not found'}
        with open(json_fp, 'r') as fd:
            data = fd.read()
        try:
            ret = json.loads(data)
        except:
            return {'code':1002, 'info':'Could not parse output JSON alignment file'}
        return {'code':0, 'data':ret}
        

