#!/usr/bin/python
# -*- coding: UTF-8 -*-

import os
import sys
import psycopg2
import psycopg2.extras
from datetime import datetime, timedelta
from ConfigParser import ConfigParser
from optparse import OptionParser
import traceback
import unicodedata
import subprocess
from imp import load_source
import json
import base64
from uuid import uuid4

from lib.libdfxp import *
from lib.libdtlx import *
from lib.libwer import lev_changes_naive
from lib.libbleu import bleu_score
from lib.libtlp import TLPSpeechClient, TLPTextClient


API_VERSION = 3

def parse_config(cfg_fd): 

    # Create object
    config = ConfigParser(allow_no_value=True)
    config.readfp(cfg_fd)
    CFG = {}

    CFG['GEN'] = {}
    CFG['GEN']['ESH'] = int(config.get('general', 'end_session_human'))
    CFG['GEN']['ESHA'] = int(config.get('general', 'end_session_human_alive'))
    CFG['GEN']['ESC'] = int(config.get('general', 'end_session_computer'))
    CFG['GEN']['CWER'] = config.getboolean('general', 'compute_wer')
    CFG['GEN']['CCER'] = config.getboolean('general', 'compute_cer')
    CFG['GEN']['CBLE'] = config.getboolean('general', 'compute_bleu')
    CFG['GEN']['CTER'] = config.getboolean('general', 'compute_ter')
    CFG['GEN']['ATX'] = config.getboolean('general', 'align_transcriptions')
    CFG['DB'] = {}
    CFG['DB']['NAME'] = config.get('database', 'db_name')
    CFG['DB']['USER'] = config.get('database', 'db_user')
    CFG['DB']['HOST'] = config.get('database', 'db_host')
    if CFG['DB']['HOST'] != None: 
      if CFG['DB']['HOST'].strip() == "": CFG['DB']['DB_HOST'] = None
    CFG['DB']['PASS'] = config.get('database', 'db_passwd')
    if CFG['DB']['PASS'] != None: 
      if CFG['DB']['PASS'].strip() == "": CFG['DB']['DB_PASS'] = None
    CFG['API'] = {}
    CFG['API']['WSURL'] = config.get("libtlp", "web_service_url")
    CFG['API']['HTEN'] = config.getboolean("libtlp", "http_auth_enabled")
    CFG['API']['HTUS'] = config.get("libtlp", "http_username")
    CFG['API']['HTPW'] = config.get("libtlp", "http_password")
    CFG['API']['USR'] = config.get("libtlp", "api_username")
    CFG['API']['KEY'] = config.get("libtlp", "api_secret_key")
    CFG['API']['AID'] = config.get("libtlp", "author_id")
    CFG['API']['ANM'] = config.get("libtlp", "author_name")
    if CFG['API']['ANM'].strip() == '': CFG['API']['ANM'] = None
    # External Modules
    CFG['EXTM'] = {}
    # WER
    CFG['EXTM']['WER'] = config.get('external_modules', 'wer'); 
    if CFG['EXTM']['WER'].strip() != "": 
      if not(os.path.isabs(CFG['EXTM']['WER'])):
        CFG['EXTM']['WER'] = SCRIPT_DIR + "/" + CFG['EXTM']['WER']
      mod = load_source('wer', CFG['EXTM']['WER'])
      if "compute_metric" not in dir(mod):
         sys.stderr.write("[EE] WER external module does not feature a compute_metric() function.\n") 
         sys.exit(1)
      CFG['EXTM']['WER'] = mod
    else:
      CFG['EXTM']['WER'] = None
    # CER
    CFG['EXTM']['CER'] = config.get('external_modules', 'cer'); 
    if CFG['EXTM']['CER'].strip() != "": 
      if not(os.path.isabs(CFG['EXTM']['CER'])):
        CFG['EXTM']['CER'] = SCRIPT_DIR + "/" + CFG['EXTM']['CER']
      mod = load_source('cer', CFG['EXTM']['CER'])
      if "compute_metric" not in dir(mod):
         sys.stderr.write("[EE] CER external module does not feature a compute_metric() function.\n") 
         sys.exit(1)
      CFG['EXTM']['CER'] = mod
    else:
      CFG['EXTM']['CER'] = None
    # BLEU
    CFG['EXTM']['BLEU'] = config.get('external_modules', 'bleu'); 
    if CFG['EXTM']['BLEU'].strip() != "": 
      if not(os.path.isabs(CFG['EXTM']['BLEU'])):
        CFG['EXTM']['BLEU'] = SCRIPT_DIR + "/" + CFG['EXTM']['BLEU']
      mod = load_source('bleu', CFG['EXTM']['BLEU'])
      if "compute_metric" not in dir(mod):
         sys.stderr.write("[EE] BLEU external module does not feature a compute_metric() function.\n") 
         sys.exit(1)
      CFG['EXTM']['BLEU'] = mod
    else:
      CFG['EXTM']['BLEU'] = None
    # TER
    CFG['EXTM']['TER'] = config.get('external_modules', 'ter'); 
    if CFG['EXTM']['TER'].strip() != "": 
      if not(os.path.isabs(CFG['EXTM']['TER'])):
        CFG['EXTM']['TER'] = SCRIPT_DIR + "/" + CFG['EXTM']['TER']
      mod = load_source('ter', CFG['EXTM']['TER'])
      if "compute_metric" not in dir(mod):
         sys.stderr.write("[EE] TER external module does not feature a compute_metric() function.\n") 
         sys.exit(1)
      CFG['EXTM']['TER'] = mod
    else:
      CFG['EXTM']['TER'] = None
    # Websites
    CFG['WEB'] = {}
    CFG['WEB']['WALA'] = config.getint('websites', 'wait_time_last_added_min')
    CFG['WEB']['DT'] = config.getboolean('websites', 'delete_texts')
    CFG['WEB']['DLA'] = config.getint('websites', 'delete_texts_after_last_access_days')
    CFG['WEB']['DST'] = config.getboolean('websites', 'delete_texts_supervised_translations')
    CFG['WEB']['DCA'] = datetime.strptime(config.get('websites', 'delete_texts_check_at'), '%H:%M') 
    return CFG

def dbconnection(db_name, db_user, db_host, db_password=None):
    """ Creates a connection to a postgreSQL database. """
    if db_password != None:
        db = psycopg2.connect(database=db_name, user=db_user, host=db_host, password=db_password);
    else:
        db = psycopg2.connect(database=db_name, user=db_user, host=db_host);
    cur = db.cursor(cursor_factory=psycopg2.extras.DictCursor)
    return db, cur


def has_translations(vid):
    cur.execute("""SELECT count(*) 
                        FROM captions c
                        WHERE c.video_id = %s
                          AND c.sub_type = 'tl'""", (vid,))
    row = cur.fetchone()
    return bool(row[0])

def is_being_processed(vhash):
    cur.execute("""SELECT count(*) FROM uploads where hash = %s and op_code in (1,2) and status_id < 20""", (vhash,))
    row = cur.fetchone()
    return bool(row[0])

def is_being_edited_speech(vid):
    cur.execute("""SELECT count(*) FROM mod_sessions where object_type = 'videos' and object_id = %s and (author_conf = 100 or author_type = 'computer') and ended_at is null""", (vid,))
    row = cur.fetchone()
    return bool(row[0])

def is_being_edited_text(did):
    cur.execute("""SELECT count(*) FROM mod_sessions where object_type = 'documents' and object_id = %s and (author_conf = 100 or author_type = 'computer') and ended_at is null""", (did,))
    row = cur.fetchone()
    return bool(row[0])


def check_ongoing_sessions_speech(max_inactivity_period_human, max_inactivity_period_human_alive, max_inactivity_period_computer, ended_by_id="tlp-daemon", ended_by_name=None):
    cur.execute("""SELECT s.id, s.author_id, s.author_conf, s.author_name, s.author_type, s.mods, s.last_update, s.last_alive, v.external_id, u.username FROM mod_sessions s INNER JOIN videos v on v.id = s.object_id INNER JOIN users u ON u.id = v.owner_id WHERE s.object_type = 'videos' AND ended_at IS NULL ORDER BY started_at ASC""")
    rows = cur.fetchall()
    now = datetime.now()
    max_human_td = timedelta(minutes=max_inactivity_period_human)
    max_human_alive_td = timedelta(minutes=max_inactivity_period_human_alive)
    max_computer_td = timedelta(minutes=max_inactivity_period_computer)
    for r in rows:
        session_id = r[0]
        author_type = r[4]
        last_update = r[6]
        last_alive = r[7]
        external_id = r[8]
        username = r[9]
        diff = (now - last_update)
        diff_alive = (now - last_alive)
        if (author_type == 'human' and diff > max_human_td) or (author_type == 'human' and diff_alive > max_human_alive_td) or (author_type == 'computer' and diff > max_computer_td):

            sys.stderr.write("[II] Automatically closing Speech session ID %d (%s)\n" % (session_id, diff))

            try:
                tlpcli_sp.api_end_session(external_id, ended_by_id, 100, author_name=ended_by_name, session_id=session_id, force=True, su=username)
            except:
                traceback.print_exc()
                continue
            sys.stderr.write("%s\n" % tlpcli_sp.get_printable_response_data())


def check_ongoing_sessions_text(max_inactivity_period_human, max_inactivity_period_human_alive, max_inactivity_period_computer, ended_by_id="tlp-daemon", ended_by_name=None):
    cur.execute("""SELECT s.id, s.author_id, s.author_conf, s.author_name, s.author_type, s.mods, s.last_update, s.last_alive, d.external_id, u.username FROM mod_sessions s INNER JOIN documents d on d.id = s.object_id INNER JOIN users u ON u.id = d.owner_id WHERE s.object_type = 'documents' AND ended_at IS NULL ORDER BY started_at ASC""")
    rows = cur.fetchall()
    now = datetime.now()
    max_human_td = timedelta(minutes=max_inactivity_period_human)
    max_human_alive_td = timedelta(minutes=max_inactivity_period_human_alive)
    max_computer_td = timedelta(minutes=max_inactivity_period_computer)
    for r in rows:
        session_id = r[0]
        author_type = r[4]
        last_update = r[6]
        last_alive = r[7]
        external_id = r[8]
        username = r[9]
        diff = (now - last_update)
        diff_alive = (now - last_alive)
        if (author_type == 'human' and diff > max_human_td) or (author_type == 'human' and diff_alive > max_human_alive_td) or (author_type == 'computer' and diff > max_computer_td):

            sys.stderr.write("[II] Automatically closing Text session ID %d (%s)\n" % (session_id, diff))

            try:
                tlpcli_txt.api_end_session(external_id, ended_by_id, 100, author_name=ended_by_name, session_id=session_id, force=True, su=username)
            except:
                traceback.print_exc()
                continue
            sys.stderr.write("%s\n" % tlpcli_txt.get_printable_response_data())

# FIXME From Web Service! Should be linked...
def segment_overlapping(rb, re, hb, he):
  return (hb < rb and he > rb + 0.1) or (hb >= rb and he <= re) or (hb < re - 0.1 and he > re)

# FIXME From Web Service! Should be linked...
def merge_mods_speech(mods, session_mods): 
    for language in mods.keys():

        mod = mods[language]
        new_txt = mod.get('txt', [])
        new_del = list(set(mod.get('del', [])))
        new_rev_time = mod.get('rev_time', None)
        if new_rev_time != None:
            new_rev_time = int(new_rev_time)
    
        if language in session_mods:
    
            lang_mod = session_mods[language]
            lang_txt = lang_mod['txt']
            lang_del = lang_mod['del']
            lang_rev_time = lang_mod['rev_time']
    
            # Process mod's rev_time key
            if lang_rev_time != None and new_rev_time != None:
                lang_mod['rev_time'] += new_rev_time
                

            # Process mod's del key
            if len(new_del) > 0:
                new_lang_txt = []
                for seg in lang_txt: 
                    if seg['sI'] not in new_del: # Delete session segments => do not append it to the new list
                        new_lang_txt.append(seg)
                lang_del = list(set(lang_del + new_del)) # Update language session del list
                lang_mod['del'] = lang_del
                lang_mod['txt'] = new_lang_txt
                lang_txt = lang_mod['txt']
    
            # Process mod's txt key
            if len(new_txt) > 0:
    
                # 1. Drop edited segment IDs
                edited_segIDs = [ seg['sI'] for seg in new_txt ]
                new_lang_txt = []
                for seg in lang_txt: 
                    if seg['sI'] not in edited_segIDs: 
                        new_lang_txt.append(seg)
                lang_mod['txt'] = new_lang_txt
                lang_txt = lang_mod['txt']
    
                # 2. DO the magic stuff
                slmod = sorted(new_txt, key=lambda x: float(x['b']), reverse=True)
                cm = slmod.pop()
                ins = False
                new_lang_txt = []
                for seg in lang_txt:
    
                    while len(slmod) > 0 and float(cm['e']) < seg['b'] + 0.1:
                        if not(ins):
                            new_lang_txt.append(cm)
                        cm = slmod.pop()
                        ins = False
            
                    if len(slmod) == 0 and float(cm['e']) < seg['b'] + 0.1 and not(ins):
                        new_lang_txt.append(cm)
                        ins = True
            
                    if segment_overlapping(float(cm['b']), float(cm['e']), seg['b'], seg['e']) and not(ins):
                        new_lang_txt.append(cm)
                        ins = True
                    else:
                        new_lang_txt.append(seg)
            
                if not(ins):
                    new_lang_txt.append(cm)
              
                while len(slmod) > 0:
                    cm = slmod.pop()
                    new_lang_txt.append(cm)
                
                lang_mod['txt'] = new_lang_txt
            
        else:
    
            session_mods[language] = { 'txt':new_txt, 'del':new_del, 'rev_time':new_rev_time }
    
    return session_mods


def get_finished_sessions_speech():
    # Build query
    if CFG['GEN']['ATX'] == True:
        cond = "(s.postprocessed = False OR s.alignment_completed = False)"
    else:
        cond = "s.postprocessed = False"
    cur.execute("""SELECT s.id, v.id, v.external_id, v.language, v.hash, u.username, s.mods, s.do_regenerate, s.evaluated, s.postprocessed, s.aligned, s.video_alignments_id, s.mods_stats, s.session_type
                   FROM mod_sessions s
                   INNER JOIN videos v ON v.id = s.object_id
                   INNER JOIN users u ON v.owner_id = u.id
                   WHERE s.object_type = 'videos' AND s.ended_at IS NOT NULL AND s.requires_postprocess = True AND %s""" % cond)
    rows = cur.fetchall()
    d = {}
    for r in rows:
       mods = json.loads(base64.b64decode(r[6]), encoding="UTF-8")
       mods_stats = json.loads(base64.b64decode(r[12]), encoding="UTF-8")
       video_id = r[1]
       if video_id in d:
           d[video_id]['session_ids'].append(r[0])
           d[video_id]['regenerate'] = d[video_id]['regenerate'] and r[7]
           d[video_id]['subs_langs_mod'] = list(set(d[video_id]['subs_langs_mod'] + mods.keys()))
           d[video_id]['merged_mods'] = merge_mods_speech(mods, d[video_id]['merged_mods'])
           d[video_id]['evaluated'] = d[video_id]['evaluated'] and r[8]
           d[video_id]['postprocessed'] = d[video_id]['postprocessed'] and r[9]
           d[video_id]['session_type'] = list(set(d[video_id]['session_type'] + [r[13]]))
           if r[11] != None:
               d[video_id]['align_id'] = r[11]
           else:
               d[video_id]['subs_langs_mod_after_align'] = list(set(d[video_id]['subs_langs_mod_after_align'] + mods.keys()))
           for k in mods_stats:
               if d[video_id]['rev_times'][k] != None: 
                   my_rev_time = mods_stats[k].get('rev_time', None)
                   if k in d[video_id]['rev_times']:
                       if my_rev_time != None:
                           d[video_id]['rev_times'][k] += my_rev_time
                       else:
                           d[video_id]['rev_times'][k] = my_rev_time
                   else:
                       d[video_id]['rev_times'][k] = my_rev_time
       else:
           d[video_id] = { 'external_id':r[2], 'language':r[3], 'vhash':r[4], 'username':r[5], 
                           'session_ids':[r[0]], 'regenerate':r[7], 'evaluated':r[8], 
                           'subs_langs_mod':mods.keys(), 
                           'rev_times':{ k:mods_stats[k].get('rev_time', None) for k in mods.keys() }, 
                           'merged_mods':mods, 'postprocessed':r[9], 'align_id':r[11], 'session_type':[r[13]],
                           'subs_langs_mod_after_align':[] }
           if r[11] == None:
               d[video_id]['subs_langs_mod_after_align'] = mods.keys()
    return d

def get_finished_sessions_text():
    cur.execute("""SELECT s.id, d.id, d.external_id, d.language, d.hash, u.username, s.mods_stats, s.do_regenerate, s.evaluated
                   FROM mod_sessions s
                   INNER JOIN documents d ON d.id = s.object_id
                   INNER JOIN users u ON d.owner_id = u.id
                   WHERE s.object_type = 'documents' AND s.ended_at IS NOT NULL AND s.requires_postprocess = True AND s.postprocessed = False""")
    rows = cur.fetchall()
    d = {}
    for r in rows:
       if r[6] != None:
           mods = json.loads(base64.b64decode(r[6]), encoding="UTF-8")
       else:
           mods = {}
       doc_id = r[1]
       if doc_id in d:
           d[doc_id]['session_ids'].append(r[0])
           d[doc_id]['regenerate'] = d[doc_id]['regenerate'] and r[7]
           d[doc_id]['langs_mod'] = list(set(d[doc_id]['langs_mod'] + mods.keys()))
       else:
           d[doc_id] = { 'external_id':r[2], 'language':r[3], 'hash':r[4], 'username':r[5], 'session_ids':[r[0]], 'regenerate':r[7], 'evaluated':r[8], 'langs_mod':mods.keys() }
    return d


U_PUNCT_CATS = set(['Pc', 'Pd', 'Ps', 'Pe', 'Pi', 'Pf', 'Po'])

def preproc(st):
  aux =  ''.join(x for x in st if unicodedata.category(x) not in U_PUNCT_CATS)
  aux = aux.lower().strip()
  return aux


def get_text_eval(pdfxp):
  doc_ts = pdfxp.timeStamp
  hyps = []
  refs = []
  nwords = 0
  nwords_sup = 0
  for d1 in pdfxp.data:
    if isinstance(d1,TlSegment):
      refseg = d1
      refsegfilt = filterSegment(refseg,filt=1,get_literal_on_subs=True)
      reftext = preproc(unicode(refsegfilt.toText()))
      refs.append(reftext)
      w = len(reftext.split())
      nwords += w
      if refseg.authorType == "human":
        nwords_sup += w
  for d1 in pdfxp.oridata:
    if isinstance(d1,TlSegment):
      hypseg = d1
      hypsegfilt = filterSegment(hypseg,filt=1,get_literal_on_subs=True)
      hyptext = preproc(unicode(hypsegfilt.toText()))
      hyps.append(hyptext)
  if nwords == 0:
    ratio_supervised = 0
  else:
    ratio_supervised = nwords_sup/float(nwords)*100
  hyps = [ ' '.join(hyps) ]
  refs = [ ' '.join(refs) ]
  return hyps, refs, ratio_supervised

def get_text_eval_dtlx(dtlx):
  hyps = []
  refs = []
  nwords = 0
  nwords_sup = 0
  for s in dtlx.sentences:
    # Ref
    s.filter(Document.FILTER_GET_BEST)
    t = s.filteredTranslation
    reftext = preproc(unicode(t.rawText))
    refs.append(reftext)
    w = len(reftext.split())
    nwords += w
    if t.authorType == "human":
      nwords_sup += w
    # Hyps
    s.filter(Document.FILTER_GET_FORMER)
    t = s.filteredTranslation
    hyptext = preproc(unicode(t.rawText))
    hyps.append(hyptext)
  ratio_supervised = nwords_sup/float(nwords)*100
  hyps = [ ' '.join(hyps) ]
  refs = [ ' '.join(refs) ]
  return hyps, refs, ratio_supervised


def my_wer(hyps, refs):
  ac_ops = 0
  ac_ref_words = 0
  for i in xrange(len(hyps)):
    hyptext = hyps[i]
    reftext = refs[i]
    ins, dels, subs = lev_changes_naive(reftext.strip().split(), hyptext.strip().split())
    ac_ops = ac_ops + ins + dels + subs
    ac_ref_words += len(reftext.strip().split())
  try:
    wer = ac_ops/float(ac_ref_words)*100
  except:
    wer = None
  return wer

def my_cer(hyps, refs):
  ac_ops = 0
  ac_ref_words = 0
  for i in xrange(len(hyps)):
    hyptext = hyps[i]
    reftext = refs[i]
    ins, dels, subs = lev_changes_naive(reftext.strip(), hyptext.strip())
    ac_ops = ac_ops + ins + dels + subs
    ac_ref_words += len(reftext.strip())
  try:
    cer = ac_ops/float(ac_ref_words)*100
  except:
    cer = None
  return cer

def my_bleu(hyps, refs):
  bleu = bleu_score(hyps, refs) 
  return bleu

def my_ter(hyps, refs):
  return None


def compute_caption_stats(csub_type, dfxp_fp):
    
    try:
        pdfxp = parseDfxp(dfxp_fp)
    except:
        sys.stderr.write("[EE] Could not parse DFXP file: %s\n" % dfxp_fp)
        traceback.print_exc()
        return False
    try:
        hyps, refs, sup_ratio = get_text_eval(pdfxp)
    except:
        sys.stderr.write("[EE] Could not process DFXP to get text for evaluation: %s\n" % dfxp_fp)
        traceback.print_exc()
        return False
    if sup_ratio == 0:
        sys.stderr.write("[EE] Supervised ratio is 0%%! %s\n" % dfxp_fp)
        return False

    sup_wer = None; sup_cer = None; sup_bleu = None; sup_ter = None
    sup_ratio /= 100.0
   
    if refs[0].strip() != "":
 
        if csub_type == "tx":
    
            info_str = "[II] ---> Sup. Ratio: %.2f" % sup_ratio

            # WER computation
            if CFG['GEN']['CWER']:
                wer_func = my_wer
                if CFG['EXTM']['WER'] != None: 
                    wer_func = CFG['EXTM']['WER'].compute_metric
                try:
                    sup_wer = wer_func(hyps, refs)
                except:
                    sys.stderr.write("[EE] Could not compute WER of DFXP file: %s\n" % dfxp_fp)
                    traceback.print_exc()
                if sup_wer != None: 
                    sup_wer = sup_wer / sup_ratio
                    info_str += ", WER: %.2f" % sup_wer
        
            # CER computation
            if CFG['GEN']['CCER']:
                cer_func = my_cer
                if CFG['EXTM']['CER'] != None: 
                    cer_func = CFG['EXTM']['CER'].compute_metric
                try:
                    sup_cer = cer_func(hyps, refs)
                except:
                    sys.stderr.write("[EE] Could not compute CER of DFXP file: %s\n" % dfxp_fp)
                    traceback.print_exc()
                if sup_cer != None: 
                    sup_cer = sup_cer / sup_ratio
                    info_str += ", CER: %.2f" % sup_cer

            sys.stderr.write("%s\n" % info_str)
        
        else: #tl
    
            # BLEU computation
            if CFG['GEN']['CBLE']:
                bleu_func = my_bleu
                if CFG['EXTM']['BLEU'] != None: 
                    bleu_func = CFG['EXTM']['BLEU'].compute_metric
                try:
                    sup_bleu = bleu_func(hyps, refs)
                except:
                    sys.stderr.write("[EE] Could not compute BLEU of DFXP file: %s\n" % dfxp_fp)
                    traceback.print_exc()
                    #return False
                if sup_bleu != None: 
                    sup_bleu = (sup_bleu - (1-sup_ratio)*100)/sup_ratio
        
            # TER computation
            if CFG['GEN']['CTER']:
                ter_func = my_ter
                if CFG['EXTM']['TER'] != None: 
                    ter_func = CFG['EXTM']['TER'].compute_metric
                try:
                    sup_ter = ter_func(hyps, refs)
                except:
                    sys.stderr.write("[EE] Could not compute TER of DFXP file: %s\n" % dfxp_fp)
                    traceback.print_exc()
                    #return False
                if sup_ter != None: 
                    sup_ter = sup_ter / sup_ratio
        
            info_str = "[II] ---> Sup. Ratio: %.2f" % sup_ratio
            if sup_bleu != None: info_str += ", BLEU: %.2f" % sup_bleu
            if sup_ter != None: info_str += ", TER: %.2f" % sup_ter
            sys.stderr.write("%s\n" % info_str)
    else:
        sys.stderr.write("[WW] Empty reference. Error/quality metrics won't be computed\n") 
        info_str = "[II] Sup. Ratio: %.2f" % sup_ratio
        sys.stderr.write("%s\n" % info_str)
    
    # Store stats
    if sup_ratio == 0:
      cap_status = 0
    elif sup_ratio > 0 and sup_ratio < 1:
      cap_status = 1
    else:
      cap_status = 2

    return cap_status, sup_ratio*100, sup_wer, sup_cer, sup_bleu, sup_ter


def get_transcription(vid):
    cur.execute("""SELECT c.id, f.file_path
                   FROM captions c INNER JOIN files f ON c.id = f.object_id INNER JOIN videos v ON v.id = c.video_id
                   WHERE v.id = %s AND c.language = v.language AND f.object_type_id = 1 
                     AND f.file_type_id = 4 AND f.file_format = 'dfxp'""", (vid,))
    ret = cur.fetchone()
    return ret


def compute_video_captions_stats(vhash):
    cur.execute("""SELECT c.id, c.language, c.sub_type, f.file_path
                   FROM captions c INNER JOIN files f ON c.id = f.object_id INNER JOIN videos v ON v.id = c.video_id
                   WHERE v.hash = %s AND f.object_type_id = 1 
                     AND f.file_type_id = 4 AND f.file_format = 'dfxp'""", (vhash,))
    ret = cur.fetchall()
    if len(ret) == 0:
        sys.stderr.write("[WW] Unknown video hash %s, or existing but has not translations\n" %vhash)
        return
    for c in ret:
        cid, clang, csub_type, rp = c
        sys.stderr.write("- Computing stats for '%s' subtitles (%s) ...\n" % (clang, csub_type))
        dfxp_fp = "%s/%s" % (transmp, rp)
        ret = compute_caption_stats(csub_type, dfxp_fp)
        if ret != False:
            cap_status, sup_ratio, sup_wer, sup_cer, sup_bleu, sup_ter = ret
            try:
                cur.execute("""UPDATE captions 
                                  SET type = %s, sup_ratio = %s, sup_wer = %s, sup_cer = %s, sup_bleu = %s, sup_ter = %s
                                  WHERE id = %s""",(cap_status, sup_ratio, sup_wer, sup_cer, sup_bleu, sup_ter, cid))
                db.commit()
            except:
                sys.stderr.write("[EE] Could not update caption stats\n")
                traceback.print_exc()
        


def process_caption_editions(vid, vs, verb):
    vhash = vs['vhash']
    vlang = vs['language']
    vexternal_id = vs['external_id']
    vowner_username = vs['username']
    for clang in vs['subs_langs_mod']:
        cur.execute("""SELECT c.id, c.language, c.type, c.sub_type, f.file_path, c.rev_time, c.rev_rtf, v.duration
                       FROM captions c INNER JOIN files f ON c.id = f.object_id INNER JOIN videos v ON v.id = c.video_id
                       WHERE c.video_id = %s AND c.language = %s AND f.object_type_id = 1 
                         AND f.file_type_id = 4 AND f.file_format = 'dfxp'""", (vid, clang))
        ret = cur.fetchone()
        if ret == None:
             sys.stderr.write("PANIC")
             sys.exit(1)
        cid, clang, ctype, csub_type, rp, current_rev_time, current_rev_rtf, video_length = ret

        my_rev_time = vs['rev_times'][clang]
        if my_rev_time != None:
            save_rev_time = current_rev_time + my_rev_time
            if video_length > 0:
                save_rev_rtf = save_rev_time / float(video_length)
        else:
            save_rev_time = current_rev_time
            save_rev_rtf = current_rev_rtf
    
        if verb:
            sys.stderr.write("[II] ** Processing Caption ID %d [LANG: %s] **\n" % (cid, clang)) 
    

        dfxp_fp = "%s/%s" % (transmp, rp)
        ret = compute_caption_stats(csub_type, dfxp_fp)
        if ret != False:
            cap_status, sup_ratio, sup_wer, sup_cer, sup_bleu, sup_ter = ret
            try:
                cur.execute("""UPDATE captions 
                                  SET type = %s, sup_ratio = %s, sup_wer = %s, sup_cer = %s, sup_bleu = %s, sup_ter = %s, rev_time = %s, rev_rtf = %s
                                  WHERE id = %s""",(cap_status, sup_ratio, sup_wer, sup_cer, sup_bleu, sup_ter, save_rev_time, save_rev_rtf, cid))
                db.commit()
            except:
                sys.stderr.write("[EE] Could not update caption stats\n")
                traceback.print_exc()
        
  

def compute_translation_stats(dtlx_fp):
    parser = Parser()
    try:
        dtlx = parser.parse(dtlx_fp)
    except:
        sys.stderr.write("[EE] Could not parse DTLX file: %s\n" % dtlx_fp)
        traceback.print_exc()
        return False
    try:
        hyps, refs, sup_ratio = get_text_eval_dtlx(dtlx)
    except:
        sys.stderr.write("[EE] Could not process DTLX to get text for evaluation: %s\n" % dtlx_fp)
        traceback.print_exc()
        return False
    if sup_ratio == 0:
        sys.stderr.write("[EE] Supervised ratio is 0%%! %s\n" % dtlx_fp)
        return False

    sup_bleu = None; sup_ter = None
    sup_ratio /= 100.0
   
    if refs[0].strip() != "":
 
        # BLEU computation
        if CFG['GEN']['CBLE']:
            bleu_func = my_bleu
            if CFG['EXTM']['BLEU'] != None: 
                bleu_func = CFG['EXTM']['BLEU'].compute_metric
            try:
                sup_bleu = bleu_func(hyps, refs)
            except:
                sys.stderr.write("[EE] Could not compute BLEU of DTLX file: %s\n" % dtlx_fp)
                traceback.print_exc()
                #return False
            if sup_bleu != None: 
                sup_bleu = (sup_bleu - (1-sup_ratio)*100)/sup_ratio
    
        # TER computation
        if CFG['GEN']['CTER']:
            ter_func = my_ter
            if CFG['EXTM']['TER'] != None: 
                ter_func = CFG['EXTM']['TER'].compute_metric
            try:
                sup_ter = ter_func(hyps, refs)
            except:
                sys.stderr.write("[EE] Could not compute TER of DTLX file: %s\n" % dtlx_fp)
                traceback.print_exc()
                #return False
            if sup_ter != None: 
                sup_ter = sup_ter / sup_ratio
    
        info_str = "[II] ---> Sup. Ratio: %.2f" % sup_ratio
        if sup_bleu != None: info_str += ", BLEU: %.2f" % sup_bleu
        if sup_ter != None: info_str += ", TER: %.2f" % sup_ter
        sys.stderr.write("%s\n" % info_str)
    else:
        sys.stderr.write("[WW] Empty reference. Error/quality metrics won't be computed\n") 
        info_str = "[II] Sup. Ratio: %.2f" % sup_ratio
        sys.stderr.write("%s\n" % info_str)
    
    # Store stats
    if sup_ratio == 0:
      cap_status = "fully_automatic"
    elif sup_ratio > 0 and sup_ratio < 1:
      cap_status = "partially_human"
    else:
      cap_status = "fully_human"

    return cap_status, sup_ratio*100, sup_bleu, sup_ter

def compute_document_translations_stats(dhash):
    cur.execute("""SELECT t.id, t.language, f.file_path
                   FROM translations t INNER JOIN files f ON t.id = f.object_id INNER JOIN documents d ON d.id = t.document_id
                   WHERE d.hash = %s AND f.object_type_id = 101
                     AND f.file_type_id = 102 AND f.file_format = 'xml'""", (dhash,))
    ret = cur.fetchall()
    if len(ret) == 0:
        sys.stderr.write("[WW] Unknown document hash %s, or existing but has not translations\n" %dhash)
        return
    for t in ret:
        tid, tlang, rp = t
        sys.stderr.write("- Computing stats for '%s' translation...\n" % tlang)
        dtlx_fp = "%s/%s" % (docsmp, rp)
        ret = compute_translation_stats(dtlx_fp)
        if ret != False:
            cap_status, sup_ratio, sup_bleu, sup_ter = ret
            try:
                cur.execute("""UPDATE translations
                                  SET status = %s, sup_ratio = %s, sup_bleu = %s, sup_ter = %s
                                  WHERE id = %s""",(cap_status, sup_ratio, sup_bleu, sup_ter, tid))
                db.commit()
            except:
                sys.stderr.write("[EE] Could not update translation stats\n")
                traceback.print_exc()
        

    
def process_translation_editions_documents(doc_id, vs, verb):
    vhash = vs['hash']
    vlang = vs['language']
    vexternal_id = vs['external_id']
    vowner_username = vs['username']
    for clang in vs['langs_mod']:
        cur.execute("""SELECT t.id, t.language, t.status, f.file_path
                       FROM translations t INNER JOIN files f ON t.id = f.object_id 
                       WHERE t.document_id = %s AND t.language = %s AND f.object_type_id = 101 
                         AND f.file_type_id = 102 AND f.file_format = 'xml'""", (doc_id, clang))
        ret = cur.fetchone()
        if ret == None:
             sys.stderr.write("PANIC")
             sys.exit(1)
        cid, clang, ctype, rp = ret

        if verb:
            sys.stderr.write("[II] ** Processing Translation ID %d [LANG: %s] **\n" % (cid, clang)) 
    
        dtlx_fp = "%s/%s" % (docsmp, rp)
        ret = compute_translation_stats(dtlx_fp)
        if ret != False:
            cap_status, sup_ratio, sup_bleu, sup_ter = ret
            try:
                cur.execute("""UPDATE translations
                                  SET status = %s, sup_ratio = %s, sup_bleu = %s, sup_ter = %s
                                  WHERE id = %s""",(cap_status, sup_ratio, sup_bleu, sup_ter, cid))
                db.commit()
            except:
                sys.stderr.write("[EE] Could not update translation stats\n")
                traceback.print_exc()


def websites_delete_unaccessed_texts(wait_days, del_sup_tl):

  now = datetime.now()
  wait_delete_last_access_td = timedelta(days=wait_days)
 
  cur.execute("""SELECT w.id, w.key, w.language, w.options, w.owner_id
                 FROM websites w""")
  websites = cur.fetchall()

  for web in websites:
      web_id = web[0]
      web_key = web[1]
      web_lang = web[2]
      web_opts = web[3]
      web_owner = web[4]


      if del_sup_tl:
          cur.execute("""SELECT st.id, st.last_access 
                          FROM source_texts st
                          WHERE st.website_id = %s""", (web_id,))
      else:
          cur.execute("""SELECT st.id, st.last_access 
                          FROM source_texts st
                          WHERE st.website_id = %s
                          AND EXISTS 
                            (SELECT * FROM translated_texts tt
                              WHERE tt.source_text_id = st.id
                                AND tt.status = 'automatic')
                          OR NOT EXISTS
                            (SELECT * FROM translated_texts tt
                             WHERE tt.source_text_id = st.id)""", (web_id,))
      rows = cur.fetchall()
      for r in rows:
          if (now - r[1]) > wait_delete_last_access_td:
              if del_sup_tl:
                  cur.execute("""DELETE FROM translated_texts 
                                  WHERE source_text_id = %s""", (r[0],))
                  cur.execute("""DELETE FROM source_texts
                                  WHERE id = %s""", (r[0],))
              else:
                  cur.execute("""DELETE FROM translated_texts 
                                  WHERE source_text_id = %s
                                    AND status = 'automatic'""", (r[0],))
                  cur.execute("""DELETE FROM source_texts st
                                  WHERE st.id = %s
                                    AND NOT EXISTS 
                                      (SELECT * FROM translated_texts tt
                                        WHERE tt.source_text_id = st.id)""", (r[0],))
              db.commit()
              


def websites_check_untranslated_texts(wait_min):
  
  now = datetime.now()
  wait_time_last_added_td = timedelta(minutes=wait_min)

  cur.execute("""SELECT w.id, w.key, w.language, w.options, w.owner_id
                 FROM websites w""")
  websites = cur.fetchall()

  for web in websites:
      web_id = web[0]
      web_key = web[1]
      web_lang = web[2]
      web_opts = web[3]
      web_owner = web[4]

      my_requested_langs = {}
      try:
          trg_langs = web_opts['requested_langs'].keys()
      except:
          sys.stderr.write("[WW] Failed to get 'requested_langs' option from website ID %d\n" % web_id)
          trg_langs = []

      my_hash = "%s" % (web_key)

      cur.execute("""SELECT count(*) FROM uploads
                      WHERE hash = %s
                        AND status_id < 3""", (my_hash,))
      ret = cur.fetchone()[0]
      if ret == 0: # Check no IS processess on this website
          
          texts_list = []
          for trg_lang in trg_langs:
    
              cur.execute("""SELECT hash, added_on 
                             FROM source_texts s 
                             WHERE s.website_id = %s
                               AND NOT EXISTS( SELECT * 
                                               FROM translated_texts tt 
                                               WHERE tt.source_text_id = s.id 
                                                 AND tt.language = %s)""", (web_id, trg_lang))
    
              stexts = cur.fetchall()
              if len(stexts) > 0:
                  max_ts = datetime(year=1970, month=1, day=1)
                  for st in stexts:
                      texts_list.append(st[0])
                      if st[1] > max_ts:
                          max_ts = st[1]
                  if (now - max_ts) > wait_time_last_added_td:
                      my_requested_langs[trg_lang] = web_opts["requested_langs"][trg_lang]
    
          texts_list = list(set(texts_list))
          if len(my_requested_langs.keys()) > 0:
        
              manifest = {
                "website_key":web_key,
                "language":web_lang,
                "texts":texts_list,
                "requested_langs":my_requested_langs
              } 
              upload_id = genUploadID()
              sys.stderr.write("[II] Submitting translation of Website %s:\n" % (web_key))
              sys.stderr.write("%s\n" % (json.dumps(manifest, indent=4, encoding="utf-8")))
              manifest_b64 = base64.b64encode(json.dumps(manifest, encoding="utf-8"))
              cur.execute("""INSERT INTO uploads 
                                    (id,hash,op_code,timestamp,status_id,info,last_status_update,
                                       external_id,user_id,manifest,object_type, debug, api_version) 
                                    VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)""",
                                    (upload_id, my_hash, 0, now, 0, 'Pending to be processed', now, 
                                     my_hash, web_owner, manifest_b64, 'websites', False, API_VERSION))
              db.commit()
          

def genUploadID():
    c = 1 
    while c > 0:
        upload_id = "up-%s"%(uuid4()); 
        cur.execute("SELECT count(*) FROM uploads WHERE id = %s AND object_type = %s", (upload_id, 'websites'));
        c = cur.fetchone()[0]
    return upload_id

def genAlignID():
    """ Generates a new align id. """
    c = 1 
    while c > 0:
        align_id = "al-%s"%(uuid4()); 
        cur.execute("SELECT count(*) FROM video_alignments WHERE id = %s", (align_id,));
        c = cur.fetchone()[0]
    return align_id

def get_available_align_languages():
    cur.execute("""SELECT distinct(lang1)
                     FROM systems
                     WHERE type = 'align' and enabled = True""")
    ret = [ x[0] for x in cur.fetchall() ]
    return ret


def cancelAlignment(al_id, sess_ids):
    cur.execute("UPDATE video_alignments set cancelled = True where id = %s", (al_id,))
    cur.execute(cur.mogrify("""UPDATE mod_sessions SET video_alignments_id = NULL, aligned = False WHERE id IN %s""", (tuple(sess_ids),)))
    db.commit()

def requestAlignment(v_id, segs, sess_ids, resegment):
    al_id = genAlignID()
    now = datetime.now()
    cur.execute("""INSERT INTO video_alignments 
                                    (id, video_id, timestamp, status_id, info, input_data, resegment) 
                                    VALUES (%s,%s,%s,%s,%s,%s,%s)""",
                             (al_id, v_id, now, 0, 'Pending to be processed', psycopg2.extras.Json(segs), resegment))
    cur.execute(cur.mogrify("""UPDATE mod_sessions SET video_alignments_id = %s WHERE id IN %s""", (al_id, tuple(sess_ids))))
    db.commit()
    return al_id

def mark_session_alignment_succeded(sess_ids, al_id):
    cur.execute(cur.mogrify("""UPDATE mod_sessions SET aligned = True, alignment_completed = True WHERE id IN %s""", (tuple(sess_ids),)))
    cur.execute("""UPDATE video_alignments SET committed = True WHERE id = %s""", (al_id,))
    db.commit()

def mark_session_alignment_failed(sess_ids):
    cur.execute(cur.mogrify("""UPDATE mod_sessions SET aligned = False, alignment_completed = True WHERE id IN %s""", (tuple(sess_ids),)))
    db.commit()

def mark_session_alignment_not_processed(sess_ids):
    cur.execute(cur.mogrify("""UPDATE mod_sessions SET aligned = False, video_alignments_id = NULL, alignment_completed = True WHERE id IN %s""", (tuple(sess_ids),)))
    db.commit()

def getAlignment(al_id):
    cur.execute("""SELECT id, timestamp, eta, completed_on, status_id, info, error_code, output_data
                               FROM video_alignments
                               WHERE id = %s""", (al_id,))
    ret = cur.fetchone()
    return ret

def commit_alignment(vid, align_segs):
    res = get_transcription(vid)
    dfxp_fp = "%s/%s" % (transmp, res['file_path'])
    try:
        pdfxp = parseDfxp(dfxp_fp)
    except:
        sys.stderr.write("[EE] Could not parse DFXP file: %s\n" % dfxp_fp)
        traceback.print_exc()
        return False
    try:
        res = updateDfxp_with_alignment(pdfxp, align_segs)
    except:
        sys.stderr.write("[EE] Could not update DFXP file: %s\n" % dfxp_fp)
        traceback.print_exc()
        return False
    if res != 0:
        return False
    # write DFXP file
    try:
        with open(dfxp_fp, 'w') as fd:
            fd.write(pdfxp.toXml())
    except:
        sys.stderr.write("[EE] Could not write DFXP file: %s\n" % dfxp_fp)
        traceback.print_exc()
        return False
    #print dfxp_fp
    #print pdfxp.toXml()
    return True
    

CFG_FILE = "%s/config.ini" % os.path.dirname(os.path.realpath(__file__))
CFG_FILENAME = "config.ini"
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))

if __name__ == "__main__":

    sys.stderr.write("[II] %s\n" % str(datetime.now()))

    usage = "usage: %prog [options]"

    parser = OptionParser(usage)
    parser.add_option("-q", "--quiet", dest="verb", action="store_false", default=True, help="Verbose power on!")
    parser.add_option("-c", "--config-file", dest="cfg_fp", type="string", default=CFG_FILE, help="Config file. Default: %s." % CFG_FILENAME)
    parser.add_option("-H", "--database-host", dest="db_host", type="string", default=None, help="Database hostname. Default: from config file.")
    parser.add_option("-U", "--database-user", dest="db_user", type="string", default=None, help="Database username. Default: from config file.")
    parser.add_option("-P", "--database-password", dest="db_pass", type="string", default=None, help="Database password. Default: from config file.")
    parser.add_option("-n", "--do-not-mark-processed", dest="do_mark_proc", action="store_false", default=True, help="Do not mark caption changes as processed.")
    parser.add_option("-r", "--do-not-regenerate", dest="do_regenerate", action="store_false", default=True, help="Do not launch regeneration processes.")
    parser.add_option("-S", "--regen-stats-video-hash", dest="video_hash", help="Just re-compute stats for a Speech object hash and exit.")
    parser.add_option("-T", "--regen-stats-text-hash", dest="document_hash", help="Just re-compute stats for a Speech object hash and exit.")

    (opts, args) = parser.parse_args()

    # Load Config file
    with open(opts.cfg_fp) as cfg_fd:
        try:
            CFG = parse_config(cfg_fd)
        except:
            sys.stderr.write("[EE] Could not parse config file.\n")
            traceback.print_exc()
            sys.exit(1)

    if opts.db_host != None:
        CFG['DB']['HOST'] = opts.db_host
    if opts.db_user != None:
        CFG['DB']['USER'] = opts.db_user
    if opts.db_pass != None:
        CFG['DB']['PASS'] = opts.db_pass

    try:    
        db, cur = dbconnection(CFG['DB']['NAME'], CFG['DB']['USER'], CFG['DB']['HOST'], CFG['DB']['PASS'])
    except:
        sys.stderr.write("[EE] Could not connect to Database.\n")
        traceback.print_exc()
        sys.exit(1)

    tlpcli_sp = TLPSpeechClient(CFG['API']['WSURL'], "", CFG['API']['USR'], CFG['API']['KEY'])
    if CFG['API']['HTEN']:
        tlpcli_sp.http_auth_enabled = True
        tlpcli_sp.http_user = CFG['HTUS']
        tlpcli_sp.http_pass = CFG['HTPW']

    tlpcli_txt = TLPTextClient(CFG['API']['WSURL'], "", CFG['API']['USR'], CFG['API']['KEY'])
    if CFG['API']['HTEN']:
        tlpcli_txt.http_auth_enabled = True
        tlpcli_txt.http_user = CFG['HTUS']
        tlpcli_txt.http_pass = CFG['HTPW']


    cur.execute("SELECT path FROM mount_points WHERE name = 'transcriptions' and machine_id = 0")
    transmp = cur.fetchone()[0]

    cur.execute("SELECT path FROM mount_points WHERE name = 'text_documents' and machine_id = 0")
    docsmp = cur.fetchone()[0]

    if opts.video_hash != None:
        compute_video_captions_stats(opts.video_hash)
        sys.exit(0)
    if opts.document_hash != None:
        compute_document_translations_stats(opts.document_hash)
        sys.exit(0)

    align_langs = get_available_align_languages()

    websites_check_untranslated_texts(CFG['WEB']['WALA'])
    if CFG['WEB']['DT']:
        now = datetime.now()
        if now.hour == CFG['WEB']['DCA'].hour and now.minute == CFG['WEB']['DCA'].minute:
            sys.stderr.write("[II] Checking unaccessed texts to be removed...\n")
            websites_delete_unaccessed_texts(CFG['WEB']['DLA'], CFG['WEB']['DST'])


    check_ongoing_sessions_speech(CFG['GEN']['ESH'], CFG['GEN']['ESHA'], CFG['GEN']['ESC'], ended_by_id=CFG['API']['AID'], ended_by_name=CFG['API']['ANM'])
    check_ongoing_sessions_text(CFG['GEN']['ESH'], CFG['GEN']['ESHA'], CFG['GEN']['ESC'], ended_by_id=CFG['API']['AID'], ended_by_name=CFG['API']['ANM'])

    vid_sessions = get_finished_sessions_speech()

    r_opts = ["tl", "tts"]

    for vid in vid_sessions.keys():
        vs = vid_sessions[vid]

        if opts.verb:
            sys.stderr.write("[II] +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n")
            sys.stderr.write("[II] **** Processing Video ID %d ****\n[II] Metadata: [VHASH: %s, VLANG: %s, VEXTID: %s, VOWNER: %s, SESSION_IDS: {%s}]\n" % (vid, vs['vhash'], vs['language'], vs['external_id'], vs['username'], ', '.join([ str(x) for x in vs['session_ids'] ]))) 

        if not(vs["postprocessed"]): # There are interactions that require postprocessing
         
            if CFG['GEN']['ATX'] == True and vs["align_id"] != None and vs['language'] in vs['subs_langs_mod_after_align']: 
                # A alignment process was launched before and 
                # Posterior mod session(s) edited the transcription file -> Cancel alignment
                if opts.verb:
                    sys.stderr.write("[II] Cancelling Align ID %s ...\n" % (vs["align_id"]))
                cancelAlignment(vs["align_id"], vs['session_ids'])
                vs["align_id"] = None
 
            if not(vs["evaluated"]):
                process_caption_editions(vid, vs, opts.verb)
                try:
                    cur.execute(cur.mogrify("""UPDATE mod_sessions SET evaluated = True WHERE id IN %s""", (tuple(vs['session_ids']),)))
                    db.commit()        
                except:
                    sys.stderr.write("\n[EE] Could not mark sessions as evaluated\n")
                    traceback.print_exc()
                    db.rollback()
    
            if is_being_edited_speech(vid):
                if opts.verb:
                    sys.stderr.write("[II] Skipping... (it is currently being edited)\n") 
                    sys.stderr.write("[II] ---------------------------------------------------------------------------------------------------------------\n\n")
                continue
     
            if CFG['GEN']['ATX'] == True:
                # Request alignment
                if vs["language"] in vs["merged_mods"] and vs["align_id"] == None: # Transcription file has been edited and no alignment process was launched
                    if len(vs["merged_mods"][vs["language"]]["txt"]) > 0: # TX has been edited
                        if vs["language"] in align_langs: # There exist align systems for such language
                            resegment = True
                            if 'player' in vs['session_type']:
                                resegment = False
                            al_id = requestAlignment(vid, vs["merged_mods"][vs["language"]]["txt"], vs['session_ids'], resegment)
                            if opts.verb:
                                sys.stderr.write("[II] Alignment requested with ID %s...\n" % al_id) 
                        else: # No align systems... 
                            sys.stderr.write("[II] No '%s' alignment system available. Won't request alignment.\n" % vs["language"]) 
                            mark_session_alignment_not_processed(vs['session_ids'])
            else:
                mark_session_alignment_not_processed(vs['session_ids'])

            if is_being_processed(vs["vhash"]):
                if opts.verb:
                    sys.stderr.write("[II] Skipping... (it is currently being processed)\n") 
                    sys.stderr.write("[II] ---------------------------------------------------------------------------------------------------------------\n\n")
                continue
     
            if has_translations(vid) and vs['regenerate'] and opts.do_regenerate:
     
                if opts.verb:
                    sys.stderr.write("[II] Requesting re-generation of translations...\n") 

                vowner_username = vs['username']
                vextid = vs['external_id']
    
                try:
                    tlpcli_sp.manifest_init(vextid)
                    tlpcli_sp.manifest_set_options(regenerate=r_opts)
                    tlpcli_sp.api_ingest_update(su=vowner_username)
                except subprocess.CalledProcessError:
                    sys.stderr.write("[EE] Could not create and ingest media package.\n")
                    traceback.print_exc()
                    db.rollback()
                    continue
                sys.stderr.write("%s\n" % tlpcli_sp.get_printable_response_data())
                if opts.verb: sys.stderr.write("\n")
        
            if opts.do_mark_proc:
                try:
                    cur.execute(cur.mogrify("""UPDATE mod_sessions SET postprocessed = True WHERE id IN %s""", (tuple(vs['session_ids']),)))
                    db.commit()        
                except:
                    sys.stderr.write("\n[EE] Could not mark sessions as postprocessed\n")
                    traceback.print_exc()
                    db.rollback()
                    continue
    
        else:
            if CFG['GEN']['ATX'] == True:
                if vs["align_id"] != None:
                    al = getAlignment(vs["align_id"])
                    if al == None:
                        # Align not found on DB
                        if opts.verb:
                            sys.stderr.write("[EE] Align ID %s not found!\n" % (vs["align_id"]))
                        mark_session_alignment_failed(vs['session_ids'])
                    else:
                        # Align finished
                        if al['status_id'] == 2:
                            if opts.verb:
                                sys.stderr.write("[II] Align ID %s finished! Commiting changes...\n" % (vs["align_id"]))
                            # Commit to DFXP
                            align_segs = al['output_data']
                            res = commit_alignment(vid, align_segs)
                            if res:
                                if opts.verb:
                                    sys.stderr.write("[II] Successfully committed.\n")
                                mark_session_alignment_succeded(vs['session_ids'], vs["align_id"])
                            else:
                                if opts.verb:
                                    sys.stderr.write("[EE] Could not commit Align ID %s.\n" % (vs["align_id"]))
                                mark_session_alignment_failed(vs['session_ids'])
                        # Align Failed
                        elif al['status_id'] > 2:
                            if opts.verb:
                                sys.stderr.write("[EE] Align ID %s failed!\n" % (vs["align_id"]))
                            mark_session_alignment_failed(vs['session_ids'])
                        else:
                            if opts.verb:
                                sys.stderr.write("[II] Waiting for Align ID %s...\n" % (vs["align_id"]))
                else:
                    pass
    
    doc_sessions = get_finished_sessions_text()

    for doc_id in doc_sessions.keys():
        ds = doc_sessions[doc_id]

        if opts.verb:
            sys.stderr.write("[II] +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n")
            sys.stderr.write("[II] **** Processing Document ID %d ****\n[II] Metadata: [HASH: %s, LANG: %s, EXTID: %s, OWNER: %s, SESSION_IDS: {%s}]\n" % (doc_id, ds['hash'], ds['language'], ds['external_id'], ds['username'], ', '.join([ str(x) for x in ds['session_ids'] ]))) 

   

        if not(ds["evaluated"]):
            process_translation_editions_documents(doc_id, ds, opts.verb)
            try:
                cur.execute(cur.mogrify("""UPDATE mod_sessions SET evaluated = True WHERE id IN %s""", (tuple(ds['session_ids']),)))
                db.commit()        
            except:
                sys.stderr.write("\n[EE] Could not mark sessions as evaluated\n")
                traceback.print_exc()
                db.rollback()

        if is_being_edited_text(doc_id):
            if opts.verb:
                sys.stderr.write("[II] Skipping... (it is currently being processed/edited)\n") 
                sys.stderr.write("[II] ---------------------------------------------------------------------------------------------------------------\n\n")
            continue
    
        if opts.do_mark_proc:
            try:
                cur.execute(cur.mogrify("""UPDATE mod_sessions SET postprocessed = True WHERE id IN %s""", (tuple(ds['session_ids']),)))
                db.commit()        
            except:
                sys.stderr.write("\n[EE] Could not mark sessions as postprocessed\n")
                traceback.print_exc()
                db.rollback()
                continue

        if opts.verb:
            sys.stderr.write("[II] ---------------------------------------------------------------------------------------------------------------\n\n")


