# -*- coding: utf-8 -
#
#  Copyright 2012,2013,2014 transLectures-UPV Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#


import psycopg2
from psycopg2.extras import Json
from uuid import uuid4
from datetime import datetime
import shutil
import os
from collections import OrderedDict
import base64
import json
import traceback
import hashlib
import re


class tLWSSData(object):
    """ Class that models the response of the API functions to the Web Service. 
    It returns the requested data and/or error codes & messages.
    Class members:
      + scode -> Status Code. 0 If succeed, greater than 0 if error.
      + results -> Number of results returned (rows).
      + data -> List of results. Each element is a dictionary which 
                contains the requested data. 
                Dict keys vary depending on the API function. 
                Example (ingest API function): [ { 'upload_id':'1234abcd' } ]
      + info -> Text with additional information about the status code.

    """

    def __init__(self, scode, results, data, info = None):
        """ Constructor """
        self.scode = scode
        self.results = results
        self.data = data
        self.info = info
   
 
    def __repr__(self):
        """ Class string representation """
        return "<<<<<< tLWSSData Object >>>>>>\n  scode: %s\n  results: %d\n  info: %s\n  data: %s\n"%(self.scode, self.results, self.info, self.data)



class tLWSSLayer(object):

    OBJ_SPEECH = "videos"
    OBJ_TEXT = "documents"
    API_VERSION = 3

    def __init__(self, db, user, host, object_type, passwd=None):
        if object_type not in [self.OBJ_SPEECH, self.OBJ_TEXT]:
            raise Exception('Bad object type')
        self.host = host
        self.db, self.cur = self.__dbconnection__(db, user, host, passwd)
        self.object_type = object_type
        mps = self.__get_mountpoints__()
        if self.object_type == self.OBJ_SPEECH:
            self.mp_media = mps['media']
            self.mp_trans = mps['transcriptions']
            self.mp_uploads = mps['uploads']
        else:
            self.mp_media = None
            self.mp_trans = mps['text_documents']
            self.mp_uploads = mps['text_uploads']
 
    def __dbconnection__(self, db_name, db_user, db_host, db_password=None):
        """ Creates a connection to a postgreSQL database. """
        if db_password != None:
            db = psycopg2.connect(database=db_name, user=db_user, host=db_host, password=db_password);
        else:
            db = psycopg2.connect(database=db_name, user=db_user, host=db_host);
        cur = db.cursor()
        return db, cur

    def __get_mountpoints__(self):
        self.cur.execute("""SELECT name, path 
                            FROM mount_points p 
                            INNER JOIN machines m ON m.id = p.machine_id 
                            WHERE m.hostname = %s""",(self.host,))
        mps = { x[0]:x[1] for x in self.cur.fetchall() }
        return mps
 
    def __rows2data__(self, rows, keys):
        """ Converts rows returned after a SQL query to a list of Dictionaries using 
            the supplied list of keys. Data columns and their corresponding keys should be in the
            same position on both lists.

        """
        data = []
        for row in rows:
            d = OrderedDict()
            for i in xrange(len(row)):
               d[keys[i]] = row[i]
            data.append(d)
        return data

    def __genUploadID__(self):
        """ Generates a new upload id. """
        c = 1 
        while c > 0:
            upload_id = "up-%s"%(uuid4()); 
            self.cur.execute("SELECT count(*) FROM uploads WHERE id = %s AND object_type = %s", (upload_id, self.object_type));
            c = self.cur.fetchone()[0]
        return upload_id

    def __genAlignID__(self):
        """ Generates a new align id. """
        c = 1 
        while c > 0:
            align_id = "al-%s"%(uuid4()); 
            self.cur.execute("SELECT count(*) FROM video_alignments WHERE id = %s", (align_id,));
            c = self.cur.fetchone()[0]
        return align_id



    ####################################################################
    #####################---- API Functions ----########################
    ####################################################################


 
    ############## Shared between /text and /speech APIs ###############

    USERS_KEYS = [ 'id', 'username', 'name', 'is_admin', 'email', 'password' ]

    def tLint_auth_get_user(self, user_id):
        try:
            self.cur.execute("""SELECT id, username, name, admin, email, password 
                                 FROM users
                                 WHERE id = %s""", (user_id,))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get user')
            return WSSdata
        data = self.__rows2data__(rows, tLWSSLayer.USERS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def tLint_auth_get_user_by_username(self, user_name):
        try:
            self.cur.execute("""SELECT id, username, name, admin, email, password 
                                 FROM users
                                 WHERE username = %s""", (user_name,))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get user')
            return WSSdata
        data = self.__rows2data__(rows, tLWSSLayer.USERS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def tLint_auth_is_owner_object(self, user_id, lec_hash, other_perms=True):
        try:
            self.cur.execute("""SELECT id, owner_id  
                                 FROM """ + self.object_type + """ 
                                 WHERE hash = %s""", (lec_hash,))
            row = self.cur.fetchone()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get object')
            return WSSdata
        if row == None:
            WSSdata = tLWSSData(0, 0, False)
            return WSSdata
        object_id = row[0]
        object_owner_id = row[1]
        if object_owner_id == user_id:
            WSSdata = tLWSSData(0, 1, True)
            return WSSdata
        if other_perms:
            try:
                self.cur.execute("""SELECT count(*) 
                                     FROM users_""" + self.object_type + """ 
                                     WHERE user_id = %s AND """ + self.object_type[:-1] + """_id = %s""", (user_id, object_id))
                row = self.cur.fetchone()
            except:
                WSSdata = tLWSSData(1, 0, None, info='Cannot get users_objects')
                return WSSdata
            if row[0] > 0:
                WSSdata = tLWSSData(0, 1, True)
                return WSSdata
            try:
                self.cur.execute("""SELECT count(*) 
                                     FROM groups_""" + self.object_type + """  gv
                                     INNER JOIN users_groups ug ON ug.group_id = gv.group_id
                                     INNER JOIN groups g ON g.id = gv.group_id
                                     WHERE gv.""" + self.object_type[:-1] + """_id = %s 
                                       AND (ug.user_id = %s or g.admin_id = %s)""", (object_id, user_id, user_id))
                row = self.cur.fetchone()
            except:
                WSSdata = tLWSSData(1, 0, None, info='Cannot get groups_objects')
                return WSSdata
            if row[0] > 0:
                WSSdata = tLWSSData(0, 1, True)
                return WSSdata
            else:
                WSSdata = tLWSSData(0, 1, False)
                return WSSdata
        else:
            WSSdata = tLWSSData(0, 1, False)
            return WSSdata
    

    def tLint_auth_is_owner_upload(self, user_id, up_id):
        try:
            self.cur.execute("""SELECT user_id 
                                 FROM uploads
                                 WHERE object_type = %s
                                       AND id = %s""", (self.object_type, up_id))
            row = self.cur.fetchone()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get upload')
            return WSSdata
        if row == None:
            WSSdata = tLWSSData(0, 0, False)
            return WSSdata
        r_user_id = row[0]
        res = (r_user_id == user_id)
        WSSdata = tLWSSData(0, 1, res)
        return WSSdata
 
    SYSTEMS_KEYS = [ 'id', 'lang1', 'lang2', 'type', 'name', 'description', 'tts_voice_gender', 'features_tadapt', 'rtf', 'const_time' ]

    def tLint_systems_getUserSystems(self, user_id):
        """ Returns systems """
        try:
            self.cur.execute("""SELECT DISTINCT s.id, lang1, lang2, type, name, description, tts_voice_gender, features_tadapt, rtf, const_time 
                                       from systems s 
                                       left outer join users_systems us on us.system_id = s.id 
                                       where 
                                            enabled is true and 
                                            (public is true or (us.user_id = %s))""", (user_id,))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get user systems')
            return WSSdata
        data = self.__rows2data__(rows, tLWSSLayer.SYSTEMS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    UPLOADLIST_KEYS = [ 'id', 'status_id', 'last_update', 'object_id', 'uploaded_on' ]

    def tLint_uploadslist_getUploads(self, user_id, object_id=None):
        if object_id == None:
            try:
                self.cur.execute("""SELECT u.id, u.status_id, u.last_status_update, u.external_id, u.timestamp 
                                    FROM uploads u 
                                    WHERE u.object_type = %s
                                      AND u.user_id = %s""",(self.object_type, user_id,))
                rows = self.cur.fetchall()
            except:
                WSSdata = tLWSSData(1, 0, None, info='Cannot get uploads list')
                return WSSdata
        else:
            try:
                self.cur.execute("""SELECT u.id, u.status_id, u.last_status_update, u.external_id, u.timestamp 
                                    FROM uploads u 
                                    WHERE u.user_id = %s 
                                          AND u.object_type = %s
                                          AND u.external_id = %s""",(user_id, self.object_type, object_id))
                rows = self.cur.fetchall()
            except:
                WSSdata = tLWSSData(1, 0, None, info='Cannot get uploads list')
                return WSSdata
        data = self.__rows2data__(rows, tLWSSLayer.UPLOADLIST_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
 

    STATUS_KEYS = [ 'id', 'up_ts', 'upload_status', 'error_code', 'event_code', 'last_check', 'info', 'user_id', 'object_id', 'eta' ]

    def tLint_status_getStatus(self,upid):
        try:
            self.cur.execute("""SELECT u.id, u.timestamp, u.status_id, u.error_code, u.event_code, u.last_status_update, u.info, u.user_id, u.external_id, u.eta
                                FROM uploads u
                                WHERE u.object_type = %s
                                      AND u.id = %s""",(self.object_type, upid))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get status for upload ID %s'%upid)
            return WSSdata
        data = self.__rows2data__(rows, tLWSSLayer.STATUS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
 
    INGEST_KEYS = [ 'upload_id' ]
 
    def tLint_ingest_storeMP(self, opc, manifest, lec_hash, user_id, pkg_file=None, debug_mode=False):
        mp = self.mp_uploads
        # Generate a new unique upload ID.
        try:
            upload_id = self.__genUploadID__()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot generate upload ID.')
            return WSSdata
        # Create directory for the uploaded media package
        upload_dir = "%s/%s"%(mp,upload_id)       
        try:
            if os.path.isdir(upload_dir):
                shutil.rmtree(upload_dir)
            os.makedirs(upload_dir)
            os.chmod(upload_dir,02775)
        except:
            WSSdata = tLWSSData(2, 0, None, info='Cannot create upload directory.')
            return WSSdata
        if pkg_file != None:
            # Save on disk the received Media Package
            mp_path = "%s/mp.zip"%(upload_dir)
            try:
                fd = open(mp_path,'wb')
                while True:
                    chunk = pkg_file.read(100000)
                    if not(chunk):
                        break
                    fd.write(chunk)
                fd.close()
            except:
                shutil.rmtree(upload_dir)
                WSSdata = tLWSSData(3, 0, None, info='Cannot store media package.')
                return WSSdata
        # Insert a new entry into the uploads table of the DB.
        # XAPUSSA!
        if self.object_type == self.OBJ_SPEECH:
            ext_id = manifest['metadata']['external_id']
        elif self.object_type == self.OBJ_TEXT:
            ext_id = "1"
        email = manifest.get("email", None)
        manifest_b64 = base64.b64encode(json.dumps(manifest, encoding="utf-8"))
        try:
            self.cur.execute("INSERT INTO uploads (id,hash,op_code,timestamp,status_id,info,last_status_update,external_id,user_id,manifest,object_type, debug, api_version) VALUES (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)",(upload_id, lec_hash, opc, datetime.now(), 0, 'Pending to be processed', datetime.now(), ext_id, user_id, manifest_b64, self.object_type, debug_mode, self.API_VERSION))
            self.db.commit()
        except:
            self.db.rollback()
            shutil.rmtree(upload_dir)
            WSSdata = tLWSSData(4, 0, None, info='Cannot store upload data into database.')
            traceback.print_exc()
            return WSSdata
        # Return tLWSSData object with the upload_id
        data = [ { tLWSSLayer.INGEST_KEYS[0]:upload_id } ]
        """ "data" should look like:
            [ { 'upload_id':'1234abcd' } ]
        """
        WSSdata = tLWSSData(0, 1, data)
        return WSSdata
 

    def tLint_ingest_cancelUpload(self, upload_id, debug_mode=False):
        try:
            self.cur.execute("UPDATE uploads SET cancel_request = True, debug = %s WHERE id = %s", (debug_mode, upload_id))
            self.db.commit()
        except:
            traceback.print_exc()
            self.db.rollback()
            WSSdata = tLWSSData(1, 0, None, info='Cannot update upload data.')
            return WSSdata
        WSSdata = tLWSSData(0, 1, True)
        return WSSdata

    SESSION_KEYS = [ 'session_id', 'session_type', 'object_type', 'author_id', 'author_conf', 'author_name', 'author_type', 'started_at', 'last_update', 'last_alive', 'ended_at', 'ended_by_id', 'ended_by_name', 'ended_by_type', 'user_id', 'object_id', 'mods', 'mods_stats', 'requires_revision', 'revised' ]

    def get_object_length(self, ohash):
        if self.object_type == self.OBJ_SPEECH:
            self.cur.execute("""SELECT duration FROM videos where hash = %s""", (ohash,))
        elif self.object_type == self.OBJ_TEXT:
            self.cur.execute("""SELECT words FROM documents where hash = %s""", (ohash,))
        else: 
            raise Exception('Bad object type')
        length = self.cur.fetchone()[0]
        return length

    def tLint_check_open_mod_sessions(self, lec_hash):
        self.cur.execute("""SELECT id, session_type, object_type, author_id, author_conf, author_name, author_type, started_at, last_update, last_alive, ended_at, ended_by_id, ended_by_name, ended_by_type, user_id, object_id, mods, mods_stats, requires_revision, revised
                            FROM mod_sessions 
                            WHERE object_id = (SELECT id FROM """ + self.object_type + """ WHERE hash = %s) 
                                  AND object_type = %s
                                  AND ended_at IS NULL""", (lec_hash, self.object_type))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.SESSION_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
        
    def tLint_check_previous_unrevised_session(self, lec_hash, author_id):
        self.cur.execute("""SELECT id, session_type, object_type, author_id, author_conf, author_name, author_type, started_at, last_update, last_alive, ended_at, ended_by_id, ended_by_name, ended_by_type, user_id, object_id, mods, mods_stats, requires_revision, revised
                            FROM mod_sessions 
                            WHERE object_id = (SELECT id FROM """ + self.object_type + """ WHERE hash = %s)
                                  AND author_id = %s 
                                  AND ended_at IS NOT NULL
                                  AND object_type = %s
                                  AND requires_revision = True AND revised = False""", (lec_hash, author_id, self.object_type))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.SESSION_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
       

    def tLint_restart_session(self, session_id):
        now = datetime.now()
        self.cur.execute("""UPDATE mod_sessions
                            SET last_update = %s, last_alive = %s
                            WHERE id = %s
                            RETURNING id, session_type, object_type, author_id, author_conf, author_name, author_type, started_at, last_update, last_alive, ended_at, ended_by_id, ended_by_name, ended_by_type, user_id, object_id, mods, mods_stats, requires_revision, revised""",
                            (now, now, session_id))
        self.db.commit()
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.SESSION_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def tLint_start_session(self, lec_hash, user_id, session_type, author_id, author_conf, author_name=None, prev_session_id=None, prev_mods=None, prev_mods_stats=None):
        if author_name != None:
            author_name = author_name.encode('utf-8')
        now = datetime.now()
        self.cur.execute("""INSERT INTO mod_sessions 
             (object_id, session_type, object_type, user_id, author_id, author_conf, author_name, started_at, last_update, last_alive, resumes_session_id, mods, mods_stats) 
             VALUES 
             ((select id from """ + self.object_type + """ where hash = %s), %s, %s, %s, %s,%s,%s, %s, %s, %s, %s, %s, %s)
             RETURNING id, session_type, object_type, author_id, author_conf, author_name, author_type, started_at, last_update, last_alive, ended_at, ended_by_id, ended_by_name, ended_by_type, user_id, object_id, mods, mods_stats, requires_revision, revised""",
           (lec_hash, session_type, self.object_type, user_id, author_id, author_conf, author_name, now, now, now, prev_session_id, prev_mods, prev_mods_stats))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.SESSION_KEYS)
        if prev_session_id != None:
            self.cur.execute("""UPDATE mod_sessions 
                SET resumed_by_session_id = %s, requires_revision = False, requires_postprocess = False, revised = False
                WHERE id = %s""",
               (data[0]['session_id'], prev_session_id))
        self.db.commit()
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
    
    def tLint_store_mod(self, session_id, language, mod, mod_stats, session_mods, session_mods_stats, session_alive):
        self.cur.execute("""INSERT INTO mods (mod_session_id, timestamp, language, on_session, mod, mod_stats)
                            VALUES (%s, %s, %s, %s, %s, %s) """, (session_id, datetime.now(), language, session_alive, mod, mod_stats))
        if session_alive:
            now = datetime.now()
            self.cur.execute("""UPDATE mod_sessions SET last_update = %s, last_alive = %s, mods = %s, mods_stats = %s WHERE id = %s""", (now, now, session_mods, session_mods_stats, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_get_mod_session(self, session_id):
        self.cur.execute("""SELECT id, session_type, object_type, author_id, author_conf, author_name, author_type, started_at, last_update, last_alive, ended_at, ended_by_id, ended_by_name, ended_by_type, user_id, (select o.hash from """ + self.object_type + """ o where o.id = object_id), mods, mods_stats, requires_revision, revised
                            FROM mod_sessions 
                            WHERE id = %s""", (session_id,))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.SESSION_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def tLint_update_last_session_update(self, session_id, now):
        self.cur.execute("""UPDATE mod_sessions 
             SET last_update = %s, last_alive = %s 
             WHERE id = %s""",
           (now, now, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_update_last_session_alive(self, session_id, now):
        self.cur.execute("""UPDATE mod_sessions 
             SET last_alive = %s 
             WHERE id = %s""",
           (now, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_end_session(self, session_id, ended_by_id, req_revision, req_postprocess, ended_by_name=None, ended_by_type = 'human', do_regenerate=True):
        now = datetime.now()
        self.cur.execute("""UPDATE mod_sessions 
             SET ended_at = %s, ended_by_id = %s, ended_by_name = %s, ended_by_type = %s, requires_revision = %s, requires_postprocess = %s, do_regenerate = %s
             WHERE id = %s""",
           (now, ended_by_id, ended_by_name, ended_by_type, req_revision, req_postprocess, do_regenerate, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_mark_revised(self, session_id, author_id, author_name=None, rev_session_id=None):
        now = datetime.now()
        self.cur.execute("""UPDATE mod_sessions 
             SET revised = True, revised_at = %s, revised_by_id = %s, revised_by_name = %s, revised_in_session_id = %s, revised_via = 'mark_revised'
             WHERE id = %s""",
           (now, author_id, author_name, rev_session_id, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_unmark_revised(self, session_id):
        now = datetime.now()
        self.cur.execute("""UPDATE mod_sessions 
             SET revised = False, revised_at = NULL, revised_by_id = NULL, revised_by_name = NULL, revised_in_session_id = NULL, postprocessed = False
             WHERE id = %s""",
           (session_id,))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_reject(self, session_id, author_id, author_name=None):
        now = datetime.now()
        self.cur.execute("""UPDATE mod_sessions 
             SET revised = True, revised_at = %s, revised_by_id = %s, revised_by_name = %s, revised_via = 'reject'
             WHERE id = %s""",
           (now, author_id, author_name, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def tLint_accept(self, session_id, author_id, author_name=None):
        now = datetime.now()
        self.cur.execute("""UPDATE mod_sessions 
             SET revised = True, revised_at = %s, revised_by_id = %s, revised_by_name = %s, revised_via = 'accept', requires_postprocess = True, postprocessed = False, do_regenerate = True
             WHERE id = %s""",
           (now, author_id, author_name, session_id))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata


    REVISIONS_KEYS = [ 'media_id', 'session_id', 'session_type', 'author_id', 'author_conf', 'author_name', 'timestamp', 'requires_revision', 'revised', 'revised_at', 'revised_by_id', 'revised_by_name', 'revised_via', 'revised_in_session_id', 'edit_stats' ]

    def tLint_get_all_media_editions(self, lec_hash):
        self.cur.execute("""SELECT (SELECT external_id FROM """ + self.object_type + """ WHERE hash = %s), id, session_type, author_id, author_conf, author_name, ended_at, requires_revision, revised, revised_at, revised_by_id, revised_by_name, revised_via, revised_in_session_id, mods_stats
                           FROM mod_sessions  
                           WHERE object_id = (SELECT id FROM """ + self.object_type + """ WHERE hash = %s)
                                 AND ended_at IS NOT NULL
                                 AND mods IS NOT NULL
                                 AND resumed_by_session_id IS NULL
                                 AND author_type = 'human'
                                 AND object_type = %s
                           ORDER BY ended_at desc""", (lec_hash,lec_hash,self.object_type))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.REVISIONS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
 
    def tLint_get_all_pending_revisions(self, user_id):
        self.cur.execute("""SELECT o.external_id, s.id, s.session_id, author_id, author_conf, author_name, ended_at, requires_revision, revised, revised_at, revised_by_id, revised_by_name, revised_via, revised_in_session_id, mods_stats
                           FROM mod_sessions s
                           INNER JOIN  """ + self.object_type + """ o ON s.object_id = o.id
                           INNER JOIN users u ON u.id = o.owner_id
                           WHERE o.owner_id = %s
                            AND ended_at IS NOT NULL
                            AND mods IS NOT NULL
                            AND requires_revision = True
                            AND revised = False
                            AND author_type = 'human'
                            AND object_type = %s
                           ORDER BY ended_at desc""", (user_id, self.object_type))
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.REVISIONS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata




    ############## Only for /speech API ###############



 
    LIST_KEYS = [ 'id', 'title', 'lang', 'duration' ]

    def tLint_list_getList(self, user_id):
        """ Returns list of media """
        try:
            self.cur.execute("""SELECT external_id, title, language, duration 
                                       FROM videos v 
                                       INNER JOIN videos_metadata vm ON v.id = vm.id
                                       WHERE v.owner_id = %s""", (user_id,))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get user systems')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.LIST_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    MEDIA_KEYS = [ 'type_code', 'media_format', 'location', 'is_url' ]

    def tLint_lecturedata_getMedia(self, extid, lec_hash):
        """ API Function
            Called by the /lecturedata interface, returns a list of media files amd thumbnails given an specific lecture id.
            Input: 
             + extid -> External ID (Lecture ID)
            Output:
             + tLWSSData object with the data specified in the static variable "tLWSSLayer.MEDIA_KEYS"

        """
        try:
            self.cur.execute("""SELECT f.file_type_id, f.file_format, f.file_path, f.is_url 
                              FROM files f
                              INNER JOIN videos v ON f.object_id = v.id 
                              WHERE v.hash = %s
                                AND f.object_type_id = 0
                                AND f.file_type_id IN (select id from file_types where name in ('media', 'thumbnail', 'waveform'))""",(lec_hash,)) 
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get media files for video ID %s'%extid)
            return WSSdata
        # Return tLWSSdata object with the list of languages available properly formatted
        data = self.__rows2data__(rows, tLWSSLayer.MEDIA_KEYS)
        """ "data" should look like:
            [ {'media_format': 'mp4', 'location': '/path/to/media.mp4', 'is_url': 0} , 
              {'media_format': 'ogv', 'location': 'http://path.com/to/media.ogv', 'is_url': 1} ]
        """
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
  
    def tLint_lecturedata_getAttachments(self, extid, lec_hash):
        """ API Function
            Called by the /lecturedata interface, returns a list of media files amd thumbnails given an specific lecture id.
            Input: 
             + extid -> External ID (Lecture ID)
            Output:
             + tLWSSData object with the data specified in the static variable "tLWSSLayer.MEDIA_KEYS"

        """
        try:
            self.cur.execute("""SELECT f.file_type_id, f.file_format, f.file_path, f.is_url 
                              FROM files f
                              INNER JOIN videos v ON f.object_id = v.id 
                              WHERE v.hash = %s
                                AND f.object_type_id = 0
                                AND f.file_type_id IN (select id from file_types where name in ('docs', 'slides'))""",(lec_hash,)) 
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get media files for video ID %s'%extid)
            return WSSdata
        # Return tLWSSdata object with the list of languages available properly formatted
        data = self.__rows2data__(rows, tLWSSLayer.MEDIA_KEYS)
        """ "data" should look like:
            [ {'media_format': 'mp4', 'location': '/path/to/media.mp4', 'is_url': 0} , 
              {'media_format': 'ogv', 'location': 'http://path.com/to/media.ogv', 'is_url': 1} ]
        """
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
 
    AUDIOTRACKS_KEYS = [ 'id', 'lang', 'voice_type', 'voice_gender', 'audio_type', 'desc', 'sub_type', 'media_format', 'location', 'is_url' ]

    def tLint_lecturedata_getAudiotracks(self,extid, lec_hash):
        try:
            self.cur.execute("""SELECT a.id, a.language, a.voice_type, a.voice_gender, a.audio_type, a.description, c.type, f.file_format, f.file_path, f.is_url
                              FROM files f
                              INNER JOIN audiotracks a ON f.object_id = a.id
                              INNER JOIN videos v ON a.video_id = v.id
                              LEFT OUTER JOIN captions c ON a.caption_id = c.id
                              WHERE v.hash = %s
                                AND f.object_type_id = 2
                                AND f.file_type_id = 5""",(lec_hash,)) 
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get media files for video ID %s'%extid)
            return WSSdata
        # Return tLWSSdata object with the list of languages available properly formatted
        data = self.__rows2data__(rows, tLWSSLayer.AUDIOTRACKS_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
  
    AUDIOTRACK_KEYS = [ 'fp' ]
 
    def tLint_audiotrack_getAudiotrack(self, extid, lec_hash, lang, aid):
        mp = self.mp_media
        # Get relative filepath of the DFXP file
        try:
            self.cur.execute("""SELECT f.file_path 
                                FROM files f 
                                INNER JOIN audiotracks a ON f.object_id = a.id 
                                INNER JOIN videos v ON a.video_id = v.id 
                                WHERE v.hash = %s AND a.language = %s
                                AND a.id = %s
                                AND f.object_type_id = 2
                                AND f.file_type_id = 5""",(lec_hash,lang,aid))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get Audiotrack ID %d file of video ID %s and language %s'%(aid,extid,lang))
            return WSSdata
        # Return tLWSSData object with the DFXP filepath
        data = [ { tLWSSLayer.AUDIOTRACK_KEYS[0]:'%s/%s'%(mp,r[0]) } for r in rows ]
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata


    LECTUREINFO_KEYS = [ 'language', 'duration', 'title', 'topic' ]
 
    def tLint_lecturedata_getLectureinfo(self,extid, lec_hash):
        """ API Function
            Called by the /lectureinfo interface, returns metadata of an specific lecture id.
            Input: 
             + extid -> External ID (Lecture ID)
            Output:
             + tLWSSData object with the data specified in the static variable "tLWSSLayer.LECTUREINFO_KEYS"

        """
        # Get list of languages availables
        try:
            self.cur.execute("""SELECT v.language, v.duration, m.title, m.topic 
                              FROM videos v
                              INNER JOIN videos_metadata m ON v.id = m.id
                              WHERE v.hash = %s""",(lec_hash,)) 
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get media files for video ID %s'%extid)
            return WSSdata
        # Return tLWSSdata object with the list of languages available properly formatted
        data = self.__rows2data__(rows, tLWSSLayer.LECTUREINFO_KEYS)
        """ "data" should look like:
            {'language': 'es', 'duration': 520, 'title': 'LECTURE TITLE', 'speaker_name': 'Pepita Greus'} 
        """
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
 
    SPEAKERS_KEYS = [ 'name' ]
 
    def tLint_lecturedata_getSpeakers(self, extid, lec_hash):
        """ API Function
            Called by the /lectureinfo interface, returns speakers of the provided lecture ID.
            Input: 
             + extid -> External ID (Lecture ID)
            Output:
             + tLWSSData object with the data specified in the static variable "tLWSSLayer.LECTUREINFO_KEYS"

        """
        # Get list of languages availables
        try:
            self.cur.execute("""SELECT s.name 
                              FROM videos v
                              INNER JOIN speakers_videos sv ON sv.video_id = v.id
                              INNER JOIN speakers s ON s.id = sv.speaker_id
                              WHERE v.hash = %s""",(lec_hash,)) 
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get media files for video ID %s'%extid)
            return WSSdata
        # Return tLWSSdata object with the list of languages available properly formatted
        data = self.__rows2data__(rows, tLWSSLayer.SPEAKERS_KEYS)
        """ "data" should look like:
            {'name': 'Barack Obama'} 
        """
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata
 

    LANGS_KEYS = [ 'code', 'value', 'type', 'sup_ratio', 'locked', 'rev_time', 'rev_rtf', 'system_time', 'system_rtf', 'media_lang' ]
 
    def tLint_langs_getLangs(self,extid, lec_hash):
        """ API Function
            Called by the /langs interface, returns a list of subtitle languages available for an specific lecture id.
            Input: 
             + extid -> External ID (Lecture ID)
            Output:
             + tLWSSData object with the data specified in the static variable "tLWSSLayer.LANGS_KEYS"

        """
        # Get list of languages availables
        self.cur.execute("""SELECT c.language, l.name, c.type, c.sup_ratio, c.locked, c.rev_time, c.rev_rtf, c.system_time, c.system_rtf, v.language 
                          FROM captions c 
                          INNER JOIN videos v ON c.video_id = v.id 
                          INNER JOIN languages l ON c.language = l.code 
                          WHERE v.hash = %s""",(lec_hash,)) 
        rows = self.cur.fetchall()
        # Return tLWSSdata object with the list of languages available properly formatted
        data = self.__rows2data__(rows, tLWSSLayer.LANGS_KEYS)
        """ "data" should look like:
            [ {'media_lang': 'es', 'code': 'es', 'type': '0', 'value': 'Spanish'} , 
              {'media_lang': 'es', 'code': 'en', 'type': '1', 'value': 'English'} , 
              {'media_lang': 'es', 'code': 'ca', 'type': '1', 'value': 'Catalan'} ]
        """
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    DFXP_KEYS = [ 'fp' ]

    def tLint_lock_subtitles(self, lec_hash, locked): 
        self.cur.execute("""UPDATE captions 
             SET locked = %s 
             WHERE video_id = (select id from videos where hash = %s)""",
           (locked, lec_hash))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata


    def tLint_dfxp_getDFXP(self, lec_hash, lang):
        """ API Function
            Called by the /dfxp and /mod interfaces, returns the DFXP filepath of an specific lecture and language.
            Input: 
             + extid -> External ID (Lecture ID)
             + lang  -> Language code.
            Output:
             + tLWSSData object with the data specified in the static variable "tLWSSLayer.DFXP_KEYS"

        """
        mp = self.mp_trans
        # Get relative filepath of the DFXP file
        try:
            self.cur.execute("""SELECT f.file_path 
                                FROM files f 
                                INNER JOIN captions c ON f.object_id = c.id 
                                INNER JOIN videos v ON c.video_id = v.id 
                                WHERE v.hash = %s AND c.language = %s 
                                AND f.object_type_id = 1 
                                AND f.file_type_id = (select id from file_types where name = 'captions')
                                AND f.file_format = 'dfxp'""",(lec_hash,lang))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get DFXP file')
            return WSSdata
        # Return tLWSSData object with the DFXP filepath
        data = [ { tLWSSLayer.DFXP_KEYS[0]:'%s/%s'%(mp,r[0]) } for r in rows ]
        """ "data" should look like:
            [ {'fp': '/absolute/path/to/file.dfxp'}]
        """
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata


    def tLint_mod_storeDFXP(self, lec_hash, lang, data, status):
        """ API Function
            Called by the /mod interface, it stores an updated (modified) DFXP file.
            Input: 
             + extid -> External ID (Lecture ID).
             + lang -> Language code.
             + data -> DFXP data to be stored.
             + status -> DFXP status attribute.
            Output:
             + Empty tLWSSData object.

        """
        # Set caption type
        if status == "fully_automatic":
            ctype = 0
        elif status == "partially_human":
            ctype = 1
        elif status == "fully_human":
            ctype = 2
        else: # None or unexpected value
            ctype = 1
        # Get dfxp filepath
        dfxpdata = self.tLint_dfxp_getDFXP(lec_hash, lang)
        if dfxpdata.scode != 0 or dfxpdata.results == 0 or dfxpdata.results > 1:
            WSSdata = tLWSSData(1, 0, None, info='Cannot open DFXP file.')
            return WSSdata
        dfxp_fn = dfxpdata.data[0][tLWSSLayer.DFXP_KEYS[0]]
        # Open file and write changes on disk
        try:
            dfxp_fd = open(dfxp_fn,'wb') 
            dfxp_fd.write(data)
            dfxp_fd.close()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot store DFXP file.')
            return WSSdata
        # Update DB, set caption type to supervised.
        try:
            self.cur.execute("""UPDATE captions 
                                SET type = %s
                                WHERE video_id = (select id from videos where hash = %s) AND language = %s""",(ctype, lec_hash, lang))
            self.db.commit()
        except:
            self.db.rollback()
            WSSdata = tLWSSData(2, 0, None, info='Cannot store data in DB.')
            return WSSdata
        # Return empty tLWSSData object
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata


 
    def tLint_align_new(self, lec_hash, segs, session_id=None, debug_mode=False):
        # Generate a new unique align ID.
        now = datetime.now()
        try:
            align_id = self.__genAlignID__()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot generate align ID.')
            traceback.print_exc()
            return WSSdata

        try:
            self.cur.execute("""INSERT INTO video_alignments 
                                    (id, video_id, timestamp, status_id, info, input_data, session_id, debug) 
                                    VALUES (%s, (select id from videos where hash = %s),%s,%s,%s,%s,%s,%s)""",
                             (align_id, lec_hash, now, 0, 'Pending to be processed', psycopg2.extras.Json(segs), session_id, debug_mode))
            self.db.commit()
        except:
            self.db.rollback()
            WSSdata = tLWSSData(2, 0, None, info='Cannot store align data into database.')
            traceback.print_exc()
            return WSSdata
        # Return tLWSSData object with the upload_id
        return tLWSSData(0, 1, align_id)


    ALIGN_KEYS = [ 'id', 'timestamp', 'eta', 'completed_on', 'status_id', 'info', 'error_code', 'output_data' ]

    def tLint_align_get(self, align_id, lec_hash):
        try:
            self.cur.execute("""SELECT id, timestamp, eta, completed_on, status_id, info, error_code, output_data
                                FROM video_alignments
                                WHERE id = %s
                                      AND video_id = (select id from videos where hash = %s)""", (align_id, lec_hash))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get align info for align ID %s'% align_id)
            return WSSdata
        data = self.__rows2data__(rows, tLWSSLayer.ALIGN_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def tLint_align_cancel(self, align_id, lec_hash):
        try:
            self.cur.execute("""UPDATE video_alignments SET cancelled = True
                                WHERE id = %s
                                      AND video_id = (select id from videos where hash = %s)""", (align_id, lec_hash))
            self.db.commit()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get align info for align ID %s'% align_id)
            return WSSdata
        if self.cur.rowcount == 0:
            WSSdata = tLWSSData(2, 0, None, info='Align ID not found')
            return WSSdata
        WSSdata = tLWSSData(0, 0, None)
        return WSSdata
 

    ############## Only for /text API ###############

 
    TEXT_LIST_KEYS = [ 'id', 'title', 'lang', 'format', 'sentences', 'words', 'chars' ]

    def text_list_getList(self, user_id):
        """ Returns list of media """
        try:
            self.cur.execute("""SELECT external_id, title, language, format, sentences, words, chars
                                       FROM documents d
                                       WHERE d.owner_id = %s""", (user_id,))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get user documents')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.TEXT_LIST_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    TEXT_METADATA_KEYS = [ 'title', 'language', 'format', 'sentences', 'words', 'characters' ]

    def text_metadata_getMetadata(self, extid, doc_hash):
        try:
            self.cur.execute("""SELECT title, language, format, sentences, words, chars
                                       FROM documents d
                                       WHERE d.hash = %s""", (doc_hash,))
            rows = self.cur.fetchall()
        except:
            WSSdata = tLWSSData(1, 0, None, info='Cannot get document')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.TEXT_METADATA_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    TEXT_LANGS_KEYS = [ 'code', 'value', 'status', 'sup_ratio', 'locked', 'document_lang' ]
 
    def text_langs_getLangs(self,extid, doc_hash):
        self.cur.execute("""SELECT t.language, l.name, t.status, t.sup_ratio, t.locked, d.language
                          FROM translations t 
                          INNER JOIN documents d ON t.document_id = d.id 
                          INNER JOIN languages l ON t.language = l.code 
                          WHERE d.hash = %s""",(doc_hash,)) 
        rows = self.cur.fetchall()
        data = self.__rows2data__(rows, tLWSSLayer.TEXT_LANGS_KEYS)
        for r in data:
            if r['status'] == 'not_translated':
                 r['status'] = 0
            elif r['status'] == 'fully_automatic':
                 r['status'] = 1
            elif r['status'] == 'partially_human':
                 r['status'] = 2
            elif r['status'] == 'fully_human':
                 r['status'] = 3
            else:
                 raise Exception('Unexpected translations.status value "%s" from DB' % r['status'])
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def text_get_getSrcDocument(self, doc_hash, form):
        mp = self.mp_trans
        if form in (2,3): file_type = "translation"
        elif form == 1: file_type = "parsed_document"
        elif form == 0: file_type = "ori_document"
        else: raise Exception("Unknown format")
        self.cur.execute("""SELECT f.file_path, f.file_format 
                                FROM files f 
                                INNER JOIN documents d ON f.object_id = d.id 
                                WHERE d.hash = %s 
                                AND f.object_type_id = (select id from object_types where table_name = 'documents') 
                                AND f.file_type_id = (select id from file_types where name = %s)""",(doc_hash, file_type))
        rows = self.cur.fetchall()
        if len(rows) == 0:
            return tLWSSData(1, 0, None, info='Cannot find document file')
        elif len(rows) > 1:
            return tLWSSData(2, 0, None, info='Unexpected result from DB')
        rfp = rows[0][0]
        ff = rows[0][1]
        fp = '%s/%s' %(mp,rfp)
        try:
            with open(fp, 'r') as doc_fd:
                data = doc_fd.read()
        except:
            return tLWSSData(3, 0, None, info='Cannot read document file')
        return tLWSSData(0, 1, [{'ff':ff, 'data':data}])


    def text_get_getTrgDocument(self, doc_hash, lang, form):
        mp = self.mp_trans
        if form in (2,3): file_type = "translation"
        elif form == 1: file_type = "parsed_document"
        elif form == 0: file_type = "ori_document"
        else: raise Exception("Unknown format")
        self.cur.execute("""SELECT f.file_path, f.file_format 
                                FROM files f 
                                INNER JOIN translations t ON f.object_id = t.id 
                                INNER JOIN documents d ON t.document_id = d.id 
                                WHERE d.hash = %s AND t.language = %s 
                                AND f.object_type_id = (select id from object_types where table_name = 'translations') 
                                AND f.file_type_id = (select id from file_types where name = %s)""", (doc_hash, lang, file_type))
        rows = self.cur.fetchall()
        if len(rows) == 0:
            return tLWSSData(1, 0, None, info='Cannot find document file')
        elif len(rows) > 1:
            return tLWSSData(2, 0, None, info='Unexpected result from DB')
        rfp = rows[0][0]
        ff = rows[0][1]
        fp = '%s/%s' %(mp,rfp)
        try:
            with open(fp, 'r') as doc_fd:
                data = doc_fd.read()
        except:
            return tLWSSData(3, 0, None, info='Cannot read document file')
        return tLWSSData(0, 1, [{'ff':ff, 'data':data}])


    def text_endsession_storeTranslation(self, doc_hash, lang, data, status, sup_ratio):
        mp = self.mp_trans
        self.cur.execute("""SELECT f.file_path 
                                FROM files f 
                                INNER JOIN translations t ON f.object_id = t.id 
                                INNER JOIN documents d ON t.document_id = d.id 
                                WHERE d.hash = %s AND t.language = %s 
                                AND f.object_type_id = (select id from object_types where table_name = 'translations') 
                                AND f.file_type_id = (select id from file_types where name = 'translation')""", (doc_hash, lang))
        dtlx_fn = self.cur.fetchone()[0]
        dtlx_fp = "%s/%s" % (mp, dtlx_fn)
        dtlx_fd = open(dtlx_fp,'wb') 
        dtlx_fd.write(data)
        dtlx_fd.close()
        self.cur.execute("""UPDATE translations 
                            SET status = %s, sup_ratio = %s
                            WHERE document_id = (select id from documents where hash = %s) AND language = %s""",(status, sup_ratio, doc_hash, lang))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata

    def text_endsession_storeDocument(self, doc_hash, lang, data, form=1):
        mp = self.mp_trans
        if form == 1: file_type = "parsed_document"
        elif form == 0: file_type = "ori_document"
        else: raise Exception("Unknown format")
        self.cur.execute("""SELECT f.file_path 
                                FROM files f 
                                INNER JOIN translations t ON f.object_id = t.id 
                                INNER JOIN documents d ON t.document_id = d.id 
                                WHERE d.hash = %s AND t.language = %s 
                                AND f.object_type_id = (select id from object_types where table_name = 'translations') 
                                AND f.file_type_id = (select id from file_types where name = %s)""", (doc_hash, lang, file_type))
        html_fn = self.cur.fetchone()[0]
        html_fp = "%s/%s" % (mp, html_fn)
        html_fd = open(html_fp, 'wb') 
        html_fd.write(data)
        html_fd.close()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata



    def tLint_lock_translations(self, doc_hash, locked): 
        self.cur.execute("""UPDATE translations 
             SET locked = %s 
             WHERE document_id = (select id from documents where hash = %s)""",
           (locked, doc_hash))
        self.db.commit()
        WSSdata =  tLWSSData(0, 0, None)
        return WSSdata



    ############## Only for /web API ###############

 
    WEB_WEBSITES_KEYS = [ 'id', 'key', 'name', 'language', 'url_pattern', 'owner_id', 'options', 'active' ]

    def __md5(self, data):
        h = hashlib.md5()
        if isinstance(data, unicode):
            data = data.encode('utf-8')
        h.update(data)
        return h.hexdigest()

    def web_get_websites_list(self, user_id):
        try:
            self.cur.execute("""SELECT id, key, name, language, url_pattern, owner_id, options, active
                                       FROM websites w
                                       WHERE w.owner_id = %s""", (user_id,))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get website info')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.WEB_WEBSITES_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata


    def web_get_website(self, key):
        try:
            self.cur.execute("""SELECT id, key, name, language, url_pattern, owner_id, options, active
                                       FROM websites w
                                       WHERE w.key = %s""", (key,))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get website info')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.WEB_WEBSITES_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def web_get_language_names(self, lang_list):
        try:
            self.cur.execute("""SELECT code, name
                                       FROM languages
                                       WHERE code in %s""", (tuple(lang_list),))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get languages')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2dict__(rows)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata


    def web_get_user_website(self, key, user_id):
        try:
            self.cur.execute("""SELECT id, key, name, language, url_pattern, owner_id, options, active
                                       FROM websites w
                                       WHERE w.key = %s AND w.owner_id = %s""", (key, user_id))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get website info')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.WEB_WEBSITES_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata


    def web_get_user_websites_by_url_pattern(self, pattern, user_id):
        pattern = '%%%s%%' % pattern
        try:
            self.cur.execute("""SELECT id, key, name, language, url_pattern, owner_id, options, active
                                       FROM websites w
                                       WHERE w.owner_id = %s AND w.url_pattern LIKE %s""", (user_id, pattern))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get website info')
            return WSSdata
        # Return tLWSSData object with such information
        data = self.__rows2data__(rows, tLWSSLayer.WEB_WEBSITES_KEYS)
        WSSdata = tLWSSData(0, len(rows), data)
        return WSSdata

    def web_get_translations(self, website_id, texts, language):
        now = datetime.now()
        # Get available translation
        hashlist = tuple(texts.keys())
        try:
            self.cur.execute("""SELECT s.hash, t.text
                                       FROM source_texts s
                                       LEFT OUTER JOIN translated_texts t 
                                           ON s.id = t.source_text_id AND t.language = %s
                                       WHERE s.website_id = %s 
                                         AND s.hash IN %s""", (language, website_id, hashlist))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(2, 0, None, info='Cannot get translated texts')
            return WSSdata
        retdict = self.__rows2dict__(rows)
        existing_src_texts = [ r[0] for r in rows ]
        # Update access timestamp to existing source_texts
        if len(existing_src_texts) > 0:
            try:
                self.cur.execute("""UPDATE source_texts 
                                      SET last_access = %s
                                           WHERE website_id = %s 
                                             AND hash IN %s""", (now, website_id, tuple(existing_src_texts)))
                self.db.commit()
            except:
                traceback.print_exc()
                WSSdata = tLWSSData(3, 0, None, info='Cannot update source_texts')
                return WSSdata
        # Add unexisting source texts
        list_tups = [ (website_id, k, now, now, texts[k]) for k in texts.keys() if k not in existing_src_texts ]
        if len(list_tups) > 0:
            arg = ','.join(self.cur.mogrify("%s", (t,)) for t in list_tups)
            try:
                self.cur.execute("""INSERT INTO source_texts 
                                      (website_id, hash, added_on, last_access, text)
                                      VALUES %s""" % arg)
                self.db.commit()
            except:
                traceback.print_exc()
                WSSdata = tLWSSData(4, 0, None, info='Cannot add new source_texts')
                return WSSdata
        WSSdata = tLWSSData(0, len(retdict.keys()), retdict)
        return WSSdata

    def web_get_all_translations(self, website_id, language):
        try:
            self.cur.execute("""SELECT s.hash, t.text, t.status, t.last_update
                                       FROM source_texts s
                                       LEFT OUTER JOIN translated_texts t 
                                           ON s.id = t.source_text_id AND t.language = %s
                                       WHERE s.website_id = %s""", (language, website_id))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get translated texts')
            return WSSdata
        retdict = self.__rows2dict_tuple__(rows)
        WSSdata = tLWSSData(0, len(retdict.keys()), retdict)
        return WSSdata

    def web_get_all_source_texts(self, website_id):
        try:
            self.cur.execute("""SELECT s.hash, s.text, s.added_on, s.last_access
                                       FROM source_texts s
                                       WHERE s.website_id = %s""", (website_id,))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get translated texts')
            return WSSdata
        retdict = self.__rows2dict_tuple__(rows)
        WSSdata = tLWSSData(0, len(retdict.keys()), retdict)
        return WSSdata


    def web_check_passwd(self, website_id, passwd):
        try:
            self.cur.execute("""SELECT u.password
                                       FROM users u
                                       INNER JOIN websites w ON w.owner_id = u.id
                                       WHERE w.id = %s""", (website_id,))
            rows = self.cur.fetchall()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get user info')
            return WSSdata
        if len(rows) == 0:
            WSSdata = tLWSSData(2, 0, None, info='Did not find user info')
            return WSSdata
        thepasswd = rows[0][0]
        equal = passwd == thepasswd
        WSSdata = tLWSSData(0, 0, equal)
        return WSSdata


    def web_amend_translations(self, website_id, texts, language):
        now = datetime.now()
        # Get available translations
        hashlist = tuple(texts.keys())
        try:
            self.cur.execute("""SELECT s.hash, t.id
                                       FROM source_texts s
                                       LEFT OUTER JOIN translated_texts t 
                                           ON s.id = t.source_text_id
                                       WHERE s.website_id = %s 
                                         AND s.hash IN %s
                                         AND (t.language = %s OR t.language IS NULL)""", (website_id, hashlist, language))
            rows = self.cur.fetchall()
        except:
            #traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot get available translations')
            return WSSdata
        d_ex_tl = self.__rows2dict__(rows)
        # Update existing translations
        list_tups = [ (d_ex_tl[k], 'revised', now, texts[k]) for k in texts.keys() if k in d_ex_tl.keys() ]
        if len(list_tups) > 0:
            arg = ','.join(self.cur.mogrify("%s", (t,)) for t in list_tups)
            try:
                self.cur.execute("""UPDATE translated_texts AS tt
                                      SET
                                        status = n.status::t_translated_texts_status,
                                        last_update = n.last_update,
                                        text = n.text
                                      FROM (VALUES %s) 
                                        AS n(id, status, last_update, text)
                                      WHERE tt.id = n.id""" % arg)
                self.db.commit()
            except:
                traceback.print_exc()
                WSSdata = tLWSSData(2, 0, None, info='Cannot update translations')
                return WSSdata
        WSSdata = tLWSSData(0, 0, len(list_tups))
        return WSSdata

    def web_amend_translation(self, website_id, language, src_hash, tl_text):
        now = datetime.now()
        # Get available translations
        try:
            self.cur.execute("""UPDATE translated_texts
                                  SET 
                                      status = %s::t_translated_texts_status,
                                      last_update = %s,
                                      text = %s 
                                 WHERE id = (SELECT tt.id FROM translated_texts tt
                                                  INNER JOIN source_texts st ON st.id = tt.source_text_id
                                                  INNER JOIN websites w ON w.id = st.website_id
                                                  WHERE w.id = %s
                                                    AND st.hash = %s
                                                    AND tt.language = %s)
                                         """, ('revised', now, tl_text, website_id, src_hash, language))
            self.db.commit()
            nrows = self.cur.rowcount
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot update translations')
            return WSSdata
        WSSdata = tLWSSData(0, 0, nrows)
        return WSSdata

    def check_language_code(self, lang_code):
        ret = False
        try:
            self.cur.execute("""SELECT count(*) FROM languages WHERE code = %s""", (str(lang_code),)) 
            ret = bool(self.cur.fetchone()[0])
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot check language code')
            return WSSdata
        return tLWSSData(0, 0, ret)

    def web_add_website(self, name, lang, url_pattern, req_langs, owner_id):
        key = self.__genWebsiteKey__()
        options = Json({"requested_langs":req_langs})
        try:
            self.cur.execute("""INSERT INTO websites
                                 (key, name, language, url_pattern, owner_id, options)
                                  VALUES (%s, %s, %s, %s, %s, %s)""", (key, name, lang, url_pattern, owner_id, options))
            self.db.commit()
        except:
            traceback.print_exc()
            WSSdata = tLWSSData(1, 0, None, info='Cannot insert website')
            return WSSdata
        WSSdata = tLWSSData(0, 0, key)
        return WSSdata

    def web_edit_website(self, web_key, enabled=None, url_pattern=None, req_langs=None):
        args = []
        sets = []
        if enabled != None:
            args.append(enabled)
            sets.append("active = %s")
        if url_pattern != None:
            args.append(url_pattern)
            sets.append("url_pattern = %s")
        if req_langs != None:
            args.append(Json({"requested_langs":req_langs}))
            sets.append("options = %s")
        if len(sets) > 0:
            sets_str = ", ".join(sets)
            args.append(web_key)
            try:
                self.cur.execute("UPDATE websites SET " + sets_str + " WHERE key = %s", tuple(args))
                self.db.commit()
            except:
                traceback.print_exc()
                WSSdata = tLWSSData(1, 0, None, info='Cannot update website')
                return WSSdata
        WSSdata = tLWSSData(0, 0, None)
        return WSSdata
               

    def web_delete_website(self, web_id, purge=False):

        if purge:
            self.cur.execute("""SELECT st.id
                            FROM source_texts st
                            WHERE st.website_id = %s""", (web_id,))
        else:
            self.cur.execute("""SELECT st.id
                            FROM source_texts st
                            WHERE st.website_id = %s
                            AND EXISTS 
                              (SELECT * FROM translated_texts tt
                                WHERE tt.source_text_id = st.id
                                  AND tt.status = 'automatic')
                            OR NOT EXISTS
                              (SELECT * FROM translated_texts tt
                               WHERE tt.source_text_id = st.id)""", (web_id,))
        rows = self.cur.fetchall()
  
        for r in rows:
            if purge:
                self.cur.execute("""DELETE FROM translated_texts 
                                WHERE source_text_id = %s""", (r[0],))
                self.cur.execute("""DELETE FROM source_texts
                                WHERE id = %s""", (r[0],))
            else:
                self.cur.execute("""DELETE FROM translated_texts 
                                WHERE source_text_id = %s
                                  AND status = 'automatic'""", (r[0],))
                self.cur.execute("""DELETE FROM source_texts st
                                WHERE st.id = %s
                                  AND NOT EXISTS 
                                    (SELECT * FROM translated_texts tt
                                      WHERE tt.source_text_id = st.id)""", (r[0],))
  
        if purge:
            self.cur.execute("""DELETE FROM websites WHERE id = %s""", (web_id,))
        else:
            self.cur.execute("""UPDATE websites SET active = False WHERE id = %s""", (web_id,))
  
        self.db.commit()
        WSSdata = tLWSSData(0, 0, None)
        return WSSdata
 

    def web_generate_js(self, js_fp, key, api_url):
        try:
            with open(js_fp, 'r') as fd:
                data = fd.read()
        except:
            traceback.print_exc()
            return tLWSSData(1, 0, None, info='Cannot open JS file')
        data, nsubs = re.subn("var TLP_KEY = '';", "var TLP_KEY = '%s';" % key, data)
        if nsubs != 1: return tLWSSData(2, 0, None, info='Cannot customize JS file')
        data, nsubs = re.subn("var TLP_API = '';", "var TLP_API = '%s';" % api_url, data)
        if nsubs != 1: return tLWSSData(3, 0, None, info='Cannot customize JS file')
        return tLWSSData(0, 0, data)
 
    def __genWebsiteKey__(self):
        """ Generates a new website key. """
        c = 1 
        while c > 0:
            web_key = str(uuid4()); 
            self.cur.execute("SELECT count(*) FROM websites WHERE key = %s", (web_key,));
            c = self.cur.fetchone()[0]
        return web_key

    def __rows2dict__(self, rows, key_pos=0, value_pos=1, null_values=False):
        return { r[key_pos]:r[value_pos] for r in rows if (r[value_pos] != None or null_values) }
    
    def __rows2dict_tuple__(self, rows):
        return { r[0]:tuple(r[1:]) for r in rows }


